------------------------------------------------------------------------------
-- BUSH AdaScript Language Parser                                           --
--                                                                          --
-- Part of BUSH                                                             --
------------------------------------------------------------------------------
--                                                                          --
--                 Copyright (C) 2001 Ken O. Burtch & FSF                   --
--                                                                          --
-- This is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  This is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with this;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- This is maintained at http://www.vaxxine.com/pegasoft                    --
--                                                                          --
------------------------------------------------------------------------------

with system,
    ada.text_io.editing,
    ada.command_line.environment,
    ada.strings.unbounded.text_io,
    ada.numerics.float_random,
    gnat.regexp,
    gnat.os_lib,
    gnat.directory_operations,
    bush_os,
    user_io,
    script_io,
    scanner,
    builtins,
    run_cmd,
    string_util,
    --signals,
    parser_aux,
    parser_strings,
    parser_numerics,
    parser_cmd;
use ada.text_io,
    ada.text_io.editing,
    ada.command_line,
    ada.command_line.environment,
    ada.strings.unbounded,
    ada.strings.unbounded.text_io,
    gnat.regexp,
    gnat.os_lib,
    gnat.directory_operations,
    bush_os,
    user_io,
    script_io,
    scanner,
    builtins,
    run_cmd,
    string_util,
    --signals,
    parser_aux,
    parser_strings,
    parser_numerics,
    parser_cmd;

package body parser is

errno : integer;
pragma import( c, errno );
-- errno is in gnat.os_lib and bush_linux/hp.  Muliple use clauses hide it
-- this won't be a problem when gnat.os_lib is phased out.

---------------------------------------------------------
-- GLOBAL STUFF
---------------------------------------------------------

originalStandardInput  : aFileDescriptor; -- before Set_Input
originalStandardOutput : aFileDescriptor; -- before Set_Output
originalStandardError  : aFileDescriptor; -- before Set_Error

currentStandardInput  : aFileDescriptor; -- as set by Set_Input
currentStandardOutput : aFileDescriptor; -- as set by Set_Output
currentStandardError  : aFileDescriptor; -- as set by Set_Error

-- sockets and files identifier fields

ch_field    : constant natural := 1;
fd_field    : constant natural := 2;
line_field  : constant natural := 3;
eol_field   : constant natural := 4;
name_field  : constant natural := 5;
mode_field  : constant natural := 6; -- files only
doget_field : constant natural := 6; -- sockets only
eof_field   : constant natural := 7;

type decimal_output_type is delta 0.0001 digits 18;
package decimal_io is new Ada.Text_IO.Editing.Decimal_Output( decimal_output_type );
use decimal_io;
-- the decimal output type is arbitrarily chosen

---------------------------------------------------------
-- START OF ADASCRIPT PARSER
---------------------------------------------------------

procedure ParseNewIdentifier( id : out identifier ) is
  -- expect a token that is a new, not previously declared identifier
  -- or one previously declared in a different scope that must be re-declared
  -- in this scope
begin
  id := eof_t; -- dummy
  if identifiers( token ).kind /= new_t then
     if token = number_t then
        err( "type name expected, not a number" );
     elsif token = strlit_t then
        err( "type name expected, not a string literal" );
     elsif is_reserved( token ) then
        err( "type name expected, not a keyword" );
     elsif token = symbol_t then
        err( "type name expected, not a symbol" );
     elsif isLocal( token ) then
        err( "already declared '" & to_string( identifiers( token ).name ) & "'" );
     else
        -- create a new one in this scope
        declareIdent( id, identifiers( token ).name, identifiers( token ).kind,
        identifiers( token ).class);
     end if;
     get_next_token;
  else
     id := token;
     get_next_token;
  end if;
exception when symbol_table_overflow =>
  err( "symbol table overflow" );
end ParseNewIdentifier;

procedure ParseIdentifier( id : out identifier ) is
  -- expect a  previously declared identifier
begin
  if token = number_t then
     err( "identifier expected, not a number" );
     id := token;
  elsif token = strlit_t then
     err( "identifier expected, not a string literal" );
     id := token;
  elsif is_reserved( token ) then
     err( "identifier expected, not a keyword" );
     id := token;
  elsif token = symbol_t then
     err( "identifier expected, not a symbol" );
     id := token;
  elsif identifiers( token ).kind = new_t or identifiers( token ).deleted then
     -- if we're skipping a block, it doesn't matter if the identifier is
     -- declared, but it does if we're executing a block or checking syntax
     if isExecutingCommand or syntax_check then
        for i in identifiers'first..identifiers_top-1 loop
            if i /= token and not identifiers(i).deleted then
               if typoOf( identifiers(i).name, identifiers(token).name ) then
                  err( "'" &
                  to_string( identifiers(token).name ) &
                  "' possible typo of '" &
                  to_string( identifiers(i).name ) &
                  "'" );
               end if;
           end if;
       end loop;
       err( "'" & to_string( identifiers( token ).name ) & "' not declared" );
     end if;
     -- this only appears if err in typo loop didn't occur
     discardUnusedIdentifier( token );
     id := eof_t;
  else
     id := token;
     get_next_token;
  end if;
exception when symbol_table_overflow =>
  err( "symbol table overflow" );
end ParseIdentifier;

procedure ParseIsOpen( b : out identifier ) is
  -- is_open( file )
  -- Ada.Text_IO.Is_Open
  file : identifier;
begin
  b := false_t;
  get_next_token;
  expect( symbol_t, "(" );
  file := token;
  get_next_token;
  expect( symbol_t, ")" );
  if token = standard_output_t then
     b := true_t;
  elsif token = standard_error_t then
     b := true_t;
  elsif uni_types_ok( file, file_type_t ) then
     if length( identifiers( file ).value ) > 0 then
        b := true_t;
     end if;
  end if;
end ParseIsOpen;

procedure ParseEndOfFile( result : out unbounded_string ) is
  -- Syntax: End_of_file( f )
  -- Source: Ada.Text_IO.End_Of_File
  file : identifier;
  kind : identifier;
begin
  result := to_unbounded_string( false'img );
  get_next_token;
  expect( symbol_t, "(" );
  ParseOpenFileOrSocket( file, kind );
  if isExecutingCommand then
     if kind = file_type_t then
        if identifier'value( to_string( stringField( file, mode_field ) ) ) /= in_file_t then
           err( "end_of_file only applies to in_mode files" );
        end if;
     end if;
  end if;
  expect( symbol_t, ")" );
  if isExecutingCommand then
     result := stringField( file, eof_field );
  end if;
end ParseEndOfFile;

procedure ParseEndOfLine( result : out unbounded_string ) is
  -- Syntax: end_of_line( open-file )
  -- Source: Ada.Text_IO.End_Of_Line
  file : identifier;
begin
  result := to_unbounded_string( integer'image( 0 ) );
  get_next_token;
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     result := stringField( file, eol_field );
  end if;
end ParseEndOfLine;

procedure ParseLine( result : out unbounded_string ) is
  -- Syntax: line( open-file )
  -- Source: Ada.Text_IO.Line
  file : identifier;
begin
  result := to_unbounded_string( integer'image( 0 ) );
  get_next_token;
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     result := stringField( file, line_field );
  end if;
end ParseLine;

procedure ParseName( result : out unbounded_string ) is
  -- Syntax: name( open-file )
  -- Source: Ada.Text_IO.Name
  file : identifier;
begin
  result := null_unbounded_string;
  get_next_token;
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     result := stringField( file, name_field );
  end if;
end ParseName;

procedure ParseMode( result : out unbounded_string ) is
  -- Syntax: mode( open-file )
  -- Source: Ada.Text_IO.Mode
  file : identifier;
begin
  result := null_unbounded_string;
  get_next_token;
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     result := stringField( file, mode_field );
     if identifier'value( to_string( result ) ) = in_file_t then
        result := identifiers( in_file_t ).value;
     elsif identifier'value( to_string( result ) ) = out_file_t then
        result := identifiers( out_file_t ).value;
     elsif identifier'value( to_string( result ) ) = append_file_t then
        result := identifiers( append_file_t ).value;
     else
        err( "internal error: unable to determine file mode" );
     end if;
  end if;
end ParseMode;

procedure DoGet( file : identifier );

procedure ParseInkey( str : out unbounded_string ) is
  -- Syntax: inkey
  -- Source: Ada.Text_IO.Inkey
  ch : character;
begin
  expect( inkey_t );
  if isExecutingCommand then
     getKey( ch );
     str := to_unbounded_string( ch & "" );
  end if;
end ParseInkey;

procedure ParseGetLine( str : out unbounded_string ) is
  -- Syntax: get_line [ (open-file) ]
  -- Source: Ada.Text_IO.Get_Line
  -- Note: Gnat get_line can't be used here because it does something
  -- odd with file descriptor 0
  file : identifier := eof_t;
  kind : identifier;
  --fd   : aFileDescriptor;
  ch   : character;
  --result : long_integer;
begin
  str := null_unbounded_string;
  get_next_token;
  --fd := stdin;
  if token = symbol_t and then identifiers( Token ).value = "(" then
      expect( symbol_t, "(" );
      ParseOpenFileOrSocket( file, kind );
      expect( symbol_t, ")" );
      --if isExecutingCommand then
      --   fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
      --end if;
  end if;
  if isExecutingCommand then
     if trace then
        put_trace( "Using file descriptor " & to_string( stringField( file, fd_field ) ) );
     end if;
     if file /= eof_t then
        if kind = socket_type_t and then stringField( file, doget_field ) = "1" then
           DoGet( file );
           replaceField( file, doget_field, false'img );
        end if;
        loop
          ch := Element( identifiers( file ).value, 1 );
          if stringField(file, eof_field ) = "1" then
             err( "end of file" );
             exit;
          end if;
          DoGet( file );
          exit when ch = ASCII.LF or error_found;
          str := str & ch;
        end loop;
     else
        getLine( str );
     end if;
     --else
     --if fd = 0 then
--      getLine( str );
     --else
     --   loop
     --      read( result, fd, ch, 1 );
     --      if result < 0 then -- should handle eagain
--            if errno /= EAGAIN and errno /= EINTR then
 --                err( "unable to read file:" & OSerror( errno ) );
  --               exit;
     --         end if;
     --      elsif result = 0 then
     --         err( "end of file" );
     --         exit;
     --      else
     --         if ch = ASCII.LF then
     --            exit;
     --         else
     --            str := str & ch;
     --         end if;
     --      end if;
     --   end loop;
     -- end if;
     -- if file /= eof_t then -- not stdin (or error)?
--       replaceField( file, line_field,
--          long_integer'image( long_integer'value(
--             to_string( stringField( file, line_field ) ) ) + 1 ) );
 --     end if;
  end if;
end ParseGetLine;

procedure ParseFactor( f : out unbounded_string; kind : out identifier ) is
  -- Syntax: factor = (expr) | "strlit" | numeric-lit | identifier | built-in fn
  -- if the identifier is volatile, reload the value from the environment
  castType : identifier;
  type aUniOp is ( noOp, doPlus, doMinus, doNot );
  uniOp : aUniOp := noOp;
  t : identifier;
begin
  if Token = symbol_t and then identifiers( Token ).value = "(" then
     expect( symbol_t, "(" );
     ParseExpression( f, kind );
     expect( symbol_t, ")" );
  else
     if Token = symbol_t and identifiers( Token ).value = "+" then
        uniOp := doPlus;
        get_next_token;
     elsif Token = symbol_t and identifiers( Token ).value = "-" then
        uniOp := doMinus;
        get_next_token;
     elsif Token = not_t then
        uniOp := doNot;
        get_next_token;
     end if;
     if Token = symbol_t and then identifiers( Token ).value = "$?" then
        f := to_unbounded_string( last_status'img );
        kind := uni_numeric_t;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value = "$$" then
        f := to_unbounded_string( aPID'image( getpid ) );
        kind := uni_numeric_t;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value = "$#" then
        if onlyAda95 then
           err( "$# not allowed with pragma 95 -- use command_line package" );
        end if;
        f := to_unbounded_string( integer'image( Argument_Count-optionOffset) );
        kind := uni_numeric_t;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value >= "$1" and then
        identifiers( Token ).value <= "$9" then
        -- this could be done a little tighter (ie length check)
        if onlyAda95 then
           err( "$1..$9 not allowed with pragma 95 -- use command_line package" );
        end if;
        begin
        f := to_unbounded_string(
           Argument(
             integer'value(
               "" & Element( identifiers( Token ).value, 2 ) )+optionOffset ) );
        kind := uni_string_t;
        exception when program_error =>
           err( "program_error exception raised" );
           kind := eof_t;
        when others =>
           err( "no such argument" );
           kind := eof_t;
        end;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value = "$0" then
        if onlyAda95 then
           err( "$0 not allowed with pragma 95 -- use command_line package" );
        end if;
        f := to_unbounded_string( Ada.Command_Line.Command_Name );
        kind := uni_string_t;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value = "@" then
        if onlyAda95 then
           err( "@ is not allowed with pragma ada_95" );
           f := null_unbounded_string;
           kind := eof_t;
        elsif not itself_defined then
           err( "@ is not defined outside of an assignment statement" );
           f := null_unbounded_string;
           kind := eof_t;
        else
           f := itself;
           kind := itself_type;
        end if;
        get_next_token;
     elsif Token = symbol_t and then identifiers( Token ).value = "%" then
        if onlyAda95 then
           err( "% is not allowed with pragma ada_95" );
           f := null_unbounded_string;
           kind := eof_t;
        elsif syntax_check then             -- % depends on run-time
           f := to_unbounded_string( "0" ); -- so just use a dummy
           kind := universal_t;             -- typeless value
        else
           f := last_output;
           kind := last_output_type;
        end if;
        get_next_token;
     elsif token = number_t then                           -- numeric literal
        f := identifiers( token ).value;
        kind := identifiers( token ).kind;
        get_next_token;
     elsif token = charlit_t then                          -- character literal
        if length( identifiers( charlit_t ).value ) > 1 then
           err( "character literal more than 1 character" );
        end if;
        f := identifiers( token ).value;
        kind := identifiers( token ).kind;
        get_next_token;
     elsif token = strlit_t then                           -- string literal
        f := identifiers( token ).value;
        kind := identifiers( token ).kind;
        get_next_token;
     elsif token = get_line_t then                        -- get line function
        ParseGetLine( f );
        kind := universal_t;
     elsif token = inkey_t then                           -- inkey function
        ParseInkey( f );
        kind := character_t;
     elsif token = end_of_file_t then                     -- end_of_file function
        ParseEndOfFile( f );
        kind := boolean_t;
     elsif token = end_of_line_t then                     -- end_of_line function
        ParseEndOfLine( f );
        kind := boolean_t;
     elsif token = line_t then                            -- line function
        ParseLine( f );
        kind := uni_numeric_t;   -- probably should be something specific
     elsif token = name_t then                            -- name function
        ParseName( f );
        kind := uni_string_t;
     elsif token = mode_t then                            -- mode function
        ParseMode( f );
        kind := file_mode_t;
     elsif token = is_open_t then                         -- is_open function
        ParseIsOpen( t );
        f := identifiers( t ).value;
        kind := boolean_t;
     elsif token = element_t then                         -- element function
        ParseStringsElement( f );
        kind := character_t;
     elsif token = slice_t then                           -- slice function
        ParseStringsSlice( f );
        kind := string_t;
     elsif token = index_t then                           -- index function
        ParseStringsIndex( f );
        kind := natural_t;
     elsif token = index_non_blank_t then                 -- index_non_blank function
        ParseStringsIndexNonBlank( f );
        kind := natural_t;
     elsif token = count_t then                           -- count function
        ParseStringsCount( f );
        kind := natural_t;
     elsif token = replace_slice_t then                   -- replace_slice function
        ParseStringsReplaceSlice( f );
        kind := string_t;
     elsif token = insert_t then                          -- insert function
        ParseStringsInsert( f );
        kind := string_t;
     elsif token = overwrite_t then                       -- overwrite function
        ParseStringsOverwrite( f );
        kind := string_t;
     elsif token = sdelete_t then                         -- delete function
        ParseStringsDelete( f );
        kind := string_t;
     elsif token = trim_t then                            -- trim function
        ParseStringsTrim( f );
        kind := string_t;
     elsif token = length_t then                          -- trim function
        ParseStringsLength( f );
        kind := natural_t;
     elsif token = head_t then                            -- head function
        ParseStringsHead( f );
        kind := string_t;
     elsif token = tail_t then                            -- tail function
        ParseStringsTail( f );
        kind := string_t;
     elsif token = field_t then                           -- field function
        ParseStringsField( f );
        kind := string_t;
     elsif token = mktemp_t then                          -- mktemp function
        ParseStringsMkTemp( f );
        kind := string_t;
     elsif token = val_t then                             -- val function
        ParseStringsVal( f );
        kind := character_t;
     elsif token = image_t then                           -- image function
        ParseStringsImage( f );
        kind := string_t;
     elsif token = glob_t then                            -- glob function
        ParseStringsGlob( f );
        kind := boolean_t;
     elsif token = match_t then                           -- match function
        ParseStringsMatch( f );
        kind := boolean_t;
     elsif token = random_t then                          -- random function
        ParseNumericsRandom( f );
        kind := uni_numeric_t;
     elsif token = shift_left_t then                      -- shift_left function
        ParseNumericsShiftLeft( f );
        kind := uni_numeric_t;
     elsif token = shift_right_t then                     -- shift_right function
        ParseNumericsShiftRight( f );
        kind := uni_numeric_t;
     elsif token = rotate_left_t then                     -- rotate_left function
        ParseNumericsRotateLeft( f );
        kind := uni_numeric_t;
     elsif token = rotate_right_t then                    -- rotate_right function
        ParseNumericsRotateRight( f );
        kind := uni_numeric_t;
     elsif token = shift_right_arith_t then               -- ASR function
        ParseNumericsASR( f );
        kind := uni_numeric_t;
     elsif token = abs_t then                             -- abs function
        ParseNumericsAbs( f );
        kind := uni_numeric_t;
     elsif token = sqrt_t then                            -- sqrt function
        ParseNumericsSqrt( f );
        kind := uni_numeric_t;
     elsif token = log_t then                             -- log function
        ParseNumericsLog( f );
        kind := uni_numeric_t;
     elsif token = exp_t then                             -- exp function
        ParseNumericsExp( f );
        kind := uni_numeric_t;
     elsif token = sin_t then                             -- sin function
        ParseNumericsSin( f );
        kind := uni_numeric_t;
     elsif token = cos_t then                             -- cos function
        ParseNumericsCos( f );
        kind := uni_numeric_t;
     elsif token = tan_t then                             -- tan function
        ParseNumericsTan( f );
        kind := uni_numeric_t;
     elsif token = cot_t then                             -- cot function
        ParseNumericsCot( f );
        kind := uni_numeric_t;
     elsif token = arcsin_t then                          -- arcsin function
        ParseNumericsArcSin( f );
        kind := uni_numeric_t;
     elsif token = arccos_t then                          -- arccos function
        ParseNumericsArcCos( f );
        kind := uni_numeric_t;
     elsif token = arctan_t then                          -- arctan function
        ParseNumericsArcTan( f );
        kind := uni_numeric_t;
     elsif token = arccot_t then                          -- arccot function
        ParseNumericsArcCot( f );
        kind := uni_numeric_t;
     elsif token = sinh_t then                            -- sinh function
        ParseNumericsSinH( f );
        kind := uni_numeric_t;
     elsif token = cosh_t then                            -- cosh function
        ParseNumericsCosH( f );
        kind := uni_numeric_t;
     elsif token = tanh_t then                            -- tanh function
        ParseNumericsTanH( f );
        kind := uni_numeric_t;
     elsif token = coth_t then                            -- coth function
        ParseNumericsCotH( f );
        kind := uni_numeric_t;
     elsif token = arcsinh_t then                         -- arcsinh function
        ParseNumericsArcSinH( f );
        kind := uni_numeric_t;
     elsif token = arccosh_t then                         -- arccosh function
        ParseNumericsArcCosH( f );
        kind := uni_numeric_t;
     elsif token = arctanh_t then                         -- arctanh function
        ParseNumericsArcTanH( f );
        kind := uni_numeric_t;
     elsif token = arccoth_t then                         -- arccoth function
        ParseNumericsArcCotH( f );
        kind := uni_numeric_t;
     elsif token = floor_t then                           -- floor function
        ParseNumericsFloor( f );
        kind := uni_numeric_t;
     elsif token = ceiling_t then                         -- ceiling function
        ParseNumericsCeiling( f );
        kind := uni_numeric_t;
     elsif token = rounding_t then                        -- rounding function
        ParseNumericsRounding( f );
        kind := uni_numeric_t;
     elsif token = unbiased_rounding_t then               -- unbiased rounding function
        ParseNumericsUnbiasedRounding( f );
        kind := uni_numeric_t;
     elsif token = truncation_t then                      -- truncation function
        ParseNumericsTruncation( f );
        kind := uni_numeric_t;
     elsif token = Remainder_t then                       -- remainder function
        ParseNumericsRemainder( f );
        kind := uni_numeric_t;
     elsif token = Exponent_t then                        -- exponent function
        ParseNumericsExponent( f );
        kind := uni_numeric_t;
     elsif token = Fraction_t then                        -- fraction function
        ParseNumericsFraction( f );
        kind := uni_numeric_t;
     elsif token = leading_part_t then                    -- leading part function
        ParseNumericsLeadingPart( f );
        kind := uni_numeric_t;
     elsif token = copy_sign_t then                       -- copy sign function
        ParseNumericsCopySign( f );
        kind := uni_numeric_t;
     elsif token = sturges_t then                         -- sturges function
        ParseNumericsSturges( f );
        kind := uni_numeric_t;
     elsif token = max_t then                             -- max function
        ParseNumericsMax( f );
        kind := uni_numeric_t;
     elsif token = min_t then                             -- min function
        ParseNumericsMin( f );
        kind := uni_numeric_t;
     elsif token = machine_t then                         -- machine function
        ParseNumericsMachine( f );
        kind := uni_numeric_t;
     elsif token = scaling_t then                         -- scaling function
        ParseNumericsScaling( f );
        kind := uni_numeric_t;
     elsif token = value_t then                           -- value function
        ParseNumericsValue( f );
        kind := uni_numeric_t;
     elsif token = pos_t then                             -- pos function
        ParseNumericsPos( f );
        kind := uni_numeric_t;
     elsif token = cmd_argcount_t then
        ParseArgument_Count( f );
        kind := uni_numeric_t;
     elsif token = cmd_argument_t then
        ParseArgument( f );
        kind := uni_string_t;
     elsif token = cmd_commandname_t then
        ParseCommand_Name( f );
        kind := uni_string_t;
     elsif identifiers( token ).kind = keyword_t then      -- no keywords
        f := null_unbounded_string;                        -- (always return something)
        kind := universal_t;
        err( "variable/value expected" );
     else                                                  -- some kind of user ident?
        ParseIdentifier( t );
        if identifiers( t ).volatile then                  -- volatile user identifier
           RefreshVolatile( t );
           f := identifiers( t ).value;
           kind := identifiers( t ).kind;
        elsif identifiers( t ).class = subClass or         -- type cast
           identifiers( t ).class = typeClass then
           castType := t;
           expect( symbol_t, "(" );
           ParseExpression( f, kind );
           expect( symbol_t, ")" );
           if uni_types_ok( castType, kind ) then
              kind := castType;
           end if;
        else                                               -- user identifier
           f := identifiers( t ).value;
           kind := identifiers( t ).kind;
        end if;
     end if;
     case uniOp is
     when noOp => null;
     when doPlus =>
          if base_types_ok( kind, uni_numeric_t ) then
             null;
          end if;
     when doMinus =>
          begin
             if base_types_ok( kind, uni_numeric_t ) then
                if isExecutingCommand then
                   f := to_unbounded_string( -to_numeric( f ) );
                end if;
             end if;
          exception when others =>
             err( "exception raised" );
          end;
     when doNot =>
          begin
             if base_types_ok( kind, boolean_t ) then
                if isExecutingCommand then
                   if to_numeric( f ) = 1.0 then
                      f := to_unbounded_string( "0" );
                   else
                      f := to_unbounded_string( "1" );
                   end if;
                end if;
             end if;
          exception when others =>
             err( "exception raised" );
          end;
     when others =>
         err( "unexpected uniary operation error" );
     end case;
  end if;
end ParseFactor;

procedure ParsePowerTermOperator( op : out unbounded_string ) is
-- Syntax: termop = "**"
begin
  -- token value is checked by parseTerm, but not token name
  if Token /= symbol_t then
     err( "operator expected");
  elsif identifiers( Token ).value /= "**" then
     err( "** operator expected");
  else
     op := identifiers( token ).value;
  end if;
  get_next_token;
end ParsePowerTermOperator;

procedure ParsePowerTerm( term : out unbounded_string; term_type : out identifier ) is
  -- Syntax: term = "factor powerterm-op factor"
  factor1  : unbounded_string;
  factor2  : unbounded_string;
  kind1    : identifier;
  kind2    : identifier;
  operator : unbounded_string;
  operation: identifier;
begin
  ParseFactor( factor1, kind1 );
  term := factor1;
  term_type := kind1;
  while identifiers( Token ).value = "**" loop
     ParsePowerTermOperator( operator );
     ParseFactor( factor2, kind2 );
     if base_types_ok( kind1, kind2 ) then
        term_type := get_base_type( kind1 );
        operation := get_uni_type( kind1 );
        if operation = universal_t then
           operation := get_uni_type( kind2 );
        end if;
        if operation = universal_t then
           operation := uni_string_t;
        end if;
        if operation = uni_numeric_t then
           if operator = "**" then
              begin
                 if isExecutingCommand then
                    term := to_unbounded_string(
                         to_numeric( term ) **
                         natural( to_numeric( factor2 ) ) );
                 end if;
              exception when program_error =>
                 err( "program_error exception raised" );
                 term := null_unbounded_string;
               when others =>
                 err( "exception raised" );
                 term := null_unbounded_string;
              end;
          else
              err( "interal error: unknown power operator" );
          end if;
        else
           err( "operation ** not defined for these types" );
        end if;
     end if;
  end loop;
end ParsePowerTerm;

procedure ParseTermOperator( op : out unbounded_string ) is
  -- Syntax: termop = '*' | '/' | '&'
begin
  -- token value is checked by parseTerm, but not token name
  if Token = mod_t or Token = rem_t then
     op := identifiers( token ).name;
  elsif Token /= symbol_t then
     err( "operator expected");
  elsif identifiers( Token ).value /= "*" and identifiers( Token ).value /= "/" and identifiers( Token ).value /= "&" then
     err( "term operator expected");
  else
     op := identifiers( token ).value;
  end if;
  get_next_token;
end ParseTermOperator;

procedure ParseTerm( term : out unbounded_string; term_type : out identifier ) is
  -- Syntax: term = "powerterm term-op powerterm"
  pterm1   : unbounded_string;
  pterm2   : unbounded_string;
  kind1    : identifier;
  kind2    : identifier;
  operator : unbounded_string;
  operation: identifier;
begin
  ParsePowerTerm( pterm1, kind1 );
  term := pterm1;
  term_type := kind1;
  while identifiers( Token ).value = "*" or
        identifiers( Token ).value = "/" or
        identifiers( Token ).value = "&" or
        Token = mod_t or Token = rem_t loop
     ParseTermOperator( operator );
     ParsePowerTerm( pterm2, kind2 );
     term_type := get_base_type( kind2 );
     operation := get_uni_type( kind2 );
     if operation = universal_t then
        operation := get_uni_type( kind1 );
     end if;
     if operation = universal_t then
        operation := uni_string_t;
     end if;
     if operation = uni_numeric_t then
        if base_types_ok( kind1, kind2 ) then
             if operator = "*" then
                begin
                  if isExecutingCommand then
                     term := to_unbounded_string(
                        to_numeric( term ) *
                        to_numeric( pterm2 ) );
                  end if;
                 exception when program_error =>
                    err( "program_error exception raised" );
                    term := null_unbounded_string;
                 when others =>
                    err( "exception raised" );
                    term := null_unbounded_string;
                 end;
             elsif operator = "/" then
                begin
                  if isExecutingCommand then
                     term := to_unbounded_string(
                        to_numeric( term ) /
                        to_numeric( pterm2 ) );
                  end if;
                 exception when program_error =>
                    err( "program_error exception raised" );
                    term := null_unbounded_string;
                 when others =>
                    err( "exception raised" );
                    term := null_unbounded_string;
                 end;
             elsif operator = "mod" then
                begin
                  if isExecutingCommand then
                     term := to_unbounded_string(
                        long_long_integer'image(
                        long_long_integer( to_numeric( term ) ) mod
                        long_long_integer( to_numeric( pterm2 ) ) ) );
                  end if;
                exception when program_error =>
                   err( "program_error exception raised" );
                   term := null_unbounded_string;
                when others =>
                   err( "exception raised" );
                   term := null_unbounded_string;
                end;
             elsif operator = "rem" then
                begin
                  if isExecutingCommand then
                     term := to_unbounded_string(
                        long_long_integer'image(
                        long_long_integer( to_numeric( term ) ) rem
                        long_long_integer( to_numeric( pterm2 ) ) ) );
                  end if;
                exception when program_error =>
                   err( "program_error exception raised" );
                   term := null_unbounded_string;
                when others =>
                   err( "exception raised" );
                   term := null_unbounded_string;
                end;
             else
                err( "Internal error: unable to handle term operator" );
             end if;
           end if;
        elsif operation = uni_string_t then
           begin
              if operator = "&" then
                 if get_base_type( kind1 ) = character_t and get_base_type( kind2 ) = string_t then
                    kind1 := string_t;
                 elsif get_base_type( kind2 ) = character_t and get_base_type( kind1 ) = string_t then
                    kind2 := string_t;
                    term_type := kind2;
                 elsif get_base_type( kind1 ) = character_t and get_base_type( kind1 ) = character_t then
                    kind1 := string_t;
                    kind2 := string_t;
                    term_type := string_t;
                 end if;
                 if base_types_ok( kind1, kind2 ) then
                    if isExecutingCommand then
                       term := term & pterm2;
                    end if;
                 end if;
              elsif operator = "*" then
                 if base_types_ok( kind1, natural_t ) then
                    if base_types_ok( kind2, uni_string_t ) then
                       if isExecutingCommand then
                          term := natural( to_numeric( term ) ) * pterm2;
                       end if;
                    end if;
                 end if;
              else
                 err( "operation not defined for string types" );
              end if;
           exception when program_error =>
              err( "program_error exception raised" );
              term := null_unbounded_string;
           when others =>
              err( "exception raised" );
              term := null_unbounded_string;
           end;
        else
           if operator = "*" then
              err( "operation * not defined for these types" );
           elsif operator = "/" then
              err( "operation / not defined for these types" );
           elsif operator = "&" then
              err( "operation & not defined for these types" );
           end if;
        end if;
  end loop;
end ParseTerm;

procedure ParseSimpleExpressionOperator( op : out unbounded_string ) is
  -- Syntax: simple-expr-op = '+' | '-'
begin
  -- token value is checked by parseTerm, but not token name
  if Token /= symbol_t then
     err( "operator expected");
  elsif identifiers( Token ).value /= "+" and identifiers( Token ).value /= "-" then
     err( "simple expression operator expected");
  end if;
  op := identifiers( token ).value;
  get_next_token;
end ParseSimpleExpressionOperator;

procedure ParseSimpleExpression( se : out unbounded_string; expr_type : out identifier ) is
  -- Syntax: term = "term expr-op term"
  term1    : unbounded_string;
  term2    : unbounded_string;
  kind1    : identifier;
  kind2    : identifier;
  operator : unbounded_string;
  operation: identifier;
begin
  ParseTerm( term1, kind1 );
  se := term1;
  expr_type := kind1;
  while identifiers( Token ).value = "+" or identifiers( Token ).value = "-" loop
     ParseSimpleExpressionOperator( operator );
     ParseTerm( term2, kind2 );
     if base_types_ok( kind1, kind2 ) then
        expr_type := get_base_type( kind1 );
        operation := get_uni_type( kind1 );
        if operation = universal_t then
           operation := get_uni_type( kind2 );
        end if;
        if operation = universal_t then
           operation := uni_string_t;
        end if;
        if operation = uni_numeric_t then
             if operator = "+" then
                begin
                  if isExecutingCommand then
                     se := to_unbounded_string(
                        to_numeric( se ) +
                        to_numeric( term2 ) );
                  end if;
                exception when program_error =>
                   err( "program_error exception raised" );
                   se := null_unbounded_string;
                when others =>
                   err( "exception raised" );
                   se := null_unbounded_string;
                end;
             elsif operator = "-" then
                begin
                  if isExecutingCommand then
                     se := to_unbounded_string(
                        to_numeric( se ) -
                        to_numeric( term2 ) );
                  end if;
                exception when program_error =>
                   err( "program_error exception raised" );
                   se := null_unbounded_string;
                when others =>
                   err( "exception raised" );
                   se := null_unbounded_string;
                end;
             end if;
        else
             if operator = "+" then
                err( "operation + not defined for these types" );
             elsif operator = "-" then
                err( "operation - not defined for these types" );
             end if;
        end if;
     end if;
  end loop;
  --put_line( "Simple Expression value = " & to_string( se ) );
end ParseSimpleExpression;

procedure ParseRelationalOperator( op : out unbounded_string ) is
  -- Syntax: rel-op = >, >=, etc.
begin
  -- token value is checked by parseTerm, but not token name
  if Token /= symbol_t then
     err( "operator expected");
  elsif identifiers( Token ).value /= ">=" and
        identifiers( Token ).value /= ">" and
        identifiers( Token ).value /= "<" and
        identifiers( Token ).value /= "<=" and
        identifiers( Token ).value /= "=" and
        identifiers( Token ).value /= "/=" then
     err( "relational operator expected");
  end if;
  op := identifiers( token ).value;
  get_next_token;
end ParseRelationalOperator;

procedure ParseRelation( re : out unbounded_string; rel_type : out identifier ) is
  -- Syntax: relation = "simple-expr" =|>|<|... "simple-expr"
  se1      : unbounded_string;
  se2      : unbounded_string;
  kind1    : identifier;
  kind2    : identifier;
  operator : unbounded_string;
  operation: identifier;
  b        : boolean;
begin
  ParseSimpleExpression( se1, kind1 );
  re := se1;
  rel_type := kind1;
  if identifiers( Token ).value = ">=" or
        identifiers( Token ).value = ">" or
        identifiers( Token ).value = "<" or
        identifiers( Token ).value = "<=" or
        identifiers( Token ).value = "=" or
        identifiers( Token ).value = "/=" then
     rel_type := boolean_t; -- always
     ParseRelationalOperator( operator );
     ParseSimpleExpression( se2, kind2 );
     if base_types_ok( kind1, kind2 ) then
        operation := get_uni_type( kind1 );
        if operation = universal_t then
           operation := get_uni_type( kind2 );
        end if;
        if operation = universal_t then
           operation := uni_string_t;
        end if;
        if operation = uni_numeric_t or operation = root_enumerated_t then
             begin
               if operator = ">=" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) >= to_numeric( se2 );
                  end if;
               elsif operator = ">" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) > to_numeric( se2 );
                  end if;
               elsif operator = "<" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) < to_numeric( se2 );
                  end if;
               elsif operator = "<=" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) <= to_numeric( se2 );
                  end if;
               elsif operator = "=" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) = to_numeric( se2 );
                  end if;
               elsif operator = "/=" then
                  if isExecutingCommand then
                     b := to_numeric( se1 ) /= to_numeric( se2 );
                  end if;
               else
                  err( "Internal error: couldn't handle relational operator" );
               end if;
               if b then
                  re := to_unbounded_string( "1" );
               else
                  re := to_unbounded_string( "0" );
               end if;
             exception when others =>
               err( "exception raised" );
             end;
        elsif operation = uni_string_t then
             if operator = ">=" then
                if isExecutingCommand then
                   b := se1 >= se2;
                end if;
             elsif operator = ">" then
                if isExecutingCommand then
                   b := se1 > se2;
                end if;
             elsif operator = "<" then
                if isExecutingCommand then
                   b := se1 < se2;
                end if;
             elsif operator = "<=" then
                if isExecutingCommand then
                   b := se1 <= se2;
                end if;
             elsif operator = "=" then
                if isExecutingCommand then
                   b := se1 = se2;
                end if;
             elsif operator = "/=" then
                if isExecutingCommand then
                   b := se1 /= se2;
                end if;
             else
                err( "Internal error: couldn't handle relational operator" );
             end if;
             if b then
                re := to_unbounded_string( "1" );
             else
                re := to_unbounded_string( "0" );
             end if;
        else
             err( "relational operation not defined for these types" );
        end if;
     end if;
  end if;
end ParseRelation;

procedure ParseExpressionOperator( op : out identifier ) is
  -- Syntax: expr-op = "and" | "or" | "xor"
begin
  if Token /= and_t and
     Token /= or_t and
     Token /= xor_t then
     err( "boolean operator expected");
  end if;
  op := Token;
  get_next_token;
end ParseExpressionOperator;

procedure ParseExpression( ex : out unbounded_string; expr_type : out identifier ) is
  -- Syntax: expr = "relation" and|or|xor "relation"
  re1      : unbounded_string;
  re2      : unbounded_string;
  kind1    : identifier;
  kind2    : identifier;
  operator : identifier;
  b        : boolean;
  type bitwise_number is mod 2**64;
begin
  ParseRelation( re1, kind1 );
  ex := re1;
  expr_type := kind1;
  while Token = and_t or Token = or_t or Token = xor_t loop
     ParseExpressionOperator( operator );
     ParseRelation( re2, kind2 );
     if base_types_ok( kind1, kind2 ) then
        if get_base_type( kind1 ) = uni_numeric_t then
           expr_type := get_base_type( kind1 );
           if operator = and_t then
              begin
                if isExecutingCommand then
                   ex := to_unbounded_string(
                      long_float(
                      bitwise_number( to_numeric( re1 ) ) and 
                      bitwise_number( to_numeric( re2 ) ) ) );
                end if;
              exception when program_error =>
                 err( "program_error exception raised" );
                 ex := null_unbounded_string;
              when others =>
                 err( "exception raised" );
                 ex := null_unbounded_string;
              end;
           elsif operator = or_t then
              begin
                if isExecutingCommand then
                   ex := to_unbounded_string(
                      long_float(
                      bitwise_number( to_numeric( re1 ) ) or 
                      bitwise_number( to_numeric( re2 ) ) ) );
                end if;
              exception when program_error =>
                 err( "program_error exception raised" );
                 ex := null_unbounded_string;
              when others =>
                 err( "exception raised" );
                 ex := null_unbounded_string;
              end;
           elsif operator = xor_t then
              begin
                if isExecutingCommand then
                   ex := to_unbounded_string(
                      long_float(
                      bitwise_number( to_numeric( re1 ) ) xor
                      bitwise_number( to_numeric( re2 ) ) ) );
                end if;
              exception when program_error =>
                 err( "program_error exception raised" );
                 ex := null_unbounded_string;
              when others =>
                 err( "exception raised" );
                 ex := null_unbounded_string;
              end;
           end if;
        elsif get_base_type( kind1 ) = boolean_t then
           expr_type := get_base_type( kind1 );
           if operator = and_t then
              if isExecutingCommand then
                 b := re1 = "1" and re2 = "1";
              end if;
           elsif operator= or_t then
              if isExecutingCommand then
                 b := re1 = "1" or re2 = "1";
              end if;
           elsif operator = xor_t then
              if isExecutingCommand then
                 b := re1 = "1" xor re2 = "1";
              end if;
           else
              err( "Internal error: unable to handle boolean operator" );
           end if;
           if b then
              ex := to_unbounded_string( "1" );
           else
              ex := to_unbounded_string( "0" );
           end if;
        else
           err( "boolean or number expected" );
        end if;
     end if;
  end loop;
  --put_line( "Expression value = " & to_string( ex ) );
end ParseExpression;

procedure ParseAssignPart( expr_value : out unbounded_string; expr_type : out identifier ) is
  -- Syntax: assign-part = " := default_value_expression"
  -- return value and type for expression
begin
  expect( symbol_t, ":=" );
  ParseExpression( expr_value, expr_type );
end ParseAssignPart;

procedure ParseDeclarationPart( id : identifier ) is
  -- Syntax: declaration = " : [aliased | constant] ident assign-part"
  -- assigns type of identifier and value (if assignment part)
  type_token : identifier;
  expr_value : unbounded_string;
  expr_type  : identifier := eof_t;
  right_type : identifier;
  expr_expected : boolean := false;
begin
  expect( symbol_t, ":" );
  if token = aliased_t then       -- skip aliased (not supported)
     err( "aliased isn't supported" );
  elsif token = constant_t then   -- if constant, set flags accordingly
     expect( constant_t );
     identifiers( id ).class := constClass;
     expr_expected := true;
  end if;
  ParseIdentifier( type_token );                              -- identify type
  if not class_ok( type_token, typeClass, subClass ) then
     null;
  elsif onlyAda95 and (type_token = uni_string_t or type_token =
     uni_numeric_t or type_token = universal_t) then
     err( "universal/typeless types not allowed with pragam ada_95" );
  elsif get_base_type( type_token ) = command_t then
     if onlyAda95 then
        err( "command types not allowed with pragma ada_95" );
     elsif not expr_expected then
        err( "command variables must be constant" );
     else
        identifiers( id ).kind := type_token;                -- and return it
     end if;
  else
     identifiers( id ).kind := type_token;                   -- and return it
  end if;
  if to_string( identifiers( token ).value ) = ":=" or        -- assign part?
     expr_expected then
     if identifiers( type_token ).limit then
        err( "limited type variables cannot be assigned a value" );
     end if;
     ParseAssignPart( expr_value, right_type );               -- evaluate assignment
     -- commands have special limitations
     if get_base_type( type_token ) = command_t then
        if base_types_ok( uni_string_t, right_type ) then
           type_token := uni_string_t; -- pretend it's a string
           if not C_is_executable_file( to_string( expr_value ) & ASCII.NUL ) then
              err( '"' & to_string( expr_value) & '"' &
                 " is not an executable command" );
           end if;
        end if;
     elsif base_types_ok( type_token, right_type ) then
        null;
     end if;
     if isExecutingCommand then
        if get_uni_type( type_token ) = uni_numeric_t then
           -- numeric test.  universal typelesses could result
           -- in a non-numeric expression that base_types_ok
           -- doesn't catch.
           declare
              lf : long_float;
           begin
              lf := to_numeric( expr_value );
           exception when program_error =>
        err( "program_error exception raised" );
           when others =>
              err( "exception raised" );
           end;
        end if;
     identifiers( id ).value := expr_value;
         if trace then
            put_trace(
               to_string( identifiers( id ).name ) & " := """ &
               to_string( escape_string( expr_value ) ) & """" );
         end if;
      end if;
  end if;
  -- failed somewhere to set a real type?
  -- blow away half-declared variable
  if error_found then
     identifiers( id ).kind := new_t;
     discardUnusedIdentifier( id );
  end if;
end ParseDeclarationPart;

procedure ParsePragma is
  import      : boolean := false;
  ada_95      : boolean := false;
  volatile    : boolean := false;
  asserting   : boolean := false;
  restriction : boolean := false;
  expr_val    : unbounded_string;
  var_id      : identifier;
begin
  expect( pragma_t );
  if identifiers( token ).name = "import" then
     import := true;
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "export" then
     import := false;
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "volatile" then
     volatile := true;
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "assert" then
     asserting := true;
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "ada_95" then
     ada_95 := true;
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "restriction" then
     restriction := true;
     discardUnusedIdentifier( token );
  else
     discardUnusedIdentifier( token );
     err( "expecting ada_95, assert, export, import, restriction or volatile" );
     return;
  end if;

  get_next_token;
  if not ada_95 then
     expect( symbol_t, "(" );
  end if;

  -- pragma assert

  if asserting then
     ParseExpression( expr_val, var_id );

  -- pragma volatile

  elsif volatile then
     ParseIdentifier( var_id );
  
  -- pragma ada_95

  elsif ada_95 then
     null;

  -- pragma restriction

  elsif restriction then
     if identifiers( token ).name /= "no_auto_declarations" then
        discardUnusedIdentifier( token );
        err( "only 'no_auto_declarations' restriction supported" );
        return;
     else
        discardUnusedIdentifier( token );
        get_next_token;
     end if;

  -- pragma import/export

  elsif identifiers( token ).name /= "shell" then
     discardUnusedIdentifier( token );
     err( "only 'shell' convention supported" );
     return;
  else
     discardUnusedIdentifier( token );
     get_next_token;
     expect( symbol_t, "," );
     ParseIdentifier( var_id );
  end if;

  if not ada_95 then
     expect( symbol_t, ")" );
  end if;

  if isExecutingCommand then
     if asserting and not syntax_check then        -- has no meaning during syntax check
        if base_types_ok( boolean_t, var_id ) then
           if expr_val = "0" then
              err( "assertion failed" );
           end if;
        end if;
     elsif volatile then
        identifiers( var_id ).volatile := true;
     elsif restriction then
        restriction_no_auto_declarations := true;
     elsif ada_95 then
        onlyAda95 := true;
     elsif import then
        identifiers( var_id ).import := true;
     else
        identifiers( var_id ).export := true;
     end if;
  end if;
end ParsePragma;

procedure ParseType is
   -- Syntax: type = "type newtype is new oldtype"
   -- NOTE: enumerateds aren't overloadable (yet)
   newtype_id : identifier;
   parent_id  : identifier;
   enum_index : integer := 0;
   b : boolean;
begin

   expect( type_t );                                       -- "type"
   ParseNewIdentifier( newtype_id );                       -- typename
   expect( is_t );                                         -- is

   if Token = symbol_t and identifiers( token ).value = "(" then

      -- enumerated
      --
      -- If an error happens during the parsing, some enumerated items
      -- may be left declared.  Should use recursion for parsing the
      -- items so they can be properly "rolled back".

      identifiers( newtype_id ).kind := root_enumerated_t; -- the parent is
      identifiers( newtype_id ).class := typeClass;        -- type based on
      parent_id := newtype_id;                             -- root enumerated
      expect( symbol_t, "(" );                             -- "("
      while token /= eof_t loop                            -- name [,name]
         ParseNewIdentifier( newtype_id );                 -- enumerated item
         -- always execute declarations when syntax checking
         -- because they are needed to test types and interpret
         -- other statements
         if isExecutingCommand or syntax_check then        -- OK to do it?
            identifiers( newtype_id ).class := constClass; -- it's a type
            identifiers( newtype_id ).value := to_unbounded_string( enum_index'img );
            identifiers( newtype_id ).kind := parent_id;   -- based on parent
         else                                              -- otherwise
            b := deleteIdent( newtype_id );                -- discard item
         end if;
         enum_index := enum_index + 1;                     -- next item number
         exit when identifiers( token ).value /= ",";      -- quit when no ","
         expect( symbol_t, "," );                          -- ","
      end loop;
      expect( symbol_t, ")" );                             -- closing ")"
      if error_found or exit_block then                    -- problems?
         b := deleteIdent( parent_id );                    -- discard parent
     end if;
   else

     -- new type

     expect( new_t );                                      -- "new"
     ParseIdentifier( parent_id );                         -- parent type name
     if class_ok( parent_id, typeClass, subClass ) then    -- not a type?
        if isExecutingCommand or syntax_check then         -- OK to do it?
          identifiers( newtype_id ).kind := parent_id;     -- define the type
          identifiers( newtype_id ).class := typeClass;
        else                                               -- otherwise
          b := deleteIdent( newtype_id );                  -- discard new type
        end if;
     end if;

   end if;
end ParseType;

procedure ParseSubtype is
   -- Syntax: type = "subtype newtype is oldtype"
   newtype_id : identifier;
   parent_id : identifier;
   b : boolean;
begin
   expect( subtype_t );                                    -- "subtype"
   ParseNewIdentifier( newtype_id );                       -- type name
   expect( is_t );                                         -- "is"
   ParseIdentifier( parent_id );                           -- old type
   if class_ok( parent_id, typeClass, subClass ) then      -- not a type?
      if isExecutingCommand or syntax_check then              -- OK to execute?
         identifiers( newtype_id ).kind := parent_id;         -- assign subtype
         identifiers( newtype_id ).class := subClass;         -- subtype class
      else                                                    -- otherwise
         b := deleteIdent( newtype_id );                      -- discard subtype
      end if;
   end if;
end ParseSubtype;

procedure ParseIfBlock is
-- Syntax: if-block = "if"... "elsif"..."else"..."end if"
  expr_val  : unbounded_string;
  expr_type : identifier;
  b : boolean := false;
  handled : boolean := false;
begin

  -- if expr then statements

  expect( if_t );                                          -- "if"
  ParseExpression( expr_val, expr_type );                  -- expression
  if not base_types_ok( boolean_t, expr_type ) then        -- not boolean result?
     err( "boolean expression expected" );
  else
     b := expr_val = "1";                                  -- convert to boolean
  end if;
  expect( then_t );                                        -- "then"

  -- elsif expr then statements

  if b then                                                -- was true?
     ParseBlock( elsif_t, else_t );                        -- handle if block
     handled := true;                                      -- remember we did it
  else                                                     -- otherwise
     SkipBlock( elsif_t, else_t );                         -- skip if block
  end if;

  while token = elsif_t loop                               -- a(nother) elsif?
     expect( elsif_t );                                    -- "elsif"
     ParseExpression( expr_val, expr_type );               -- expression
     if not base_types_ok( boolean_t, expr_type ) then     -- not boolean result?
        err( "boolean expression expected" );
     else
        b := expr_val = "1";                               -- convert to boolean
     end if;
     expect( then_t );                                     -- "then"
     if b and not handled then                             -- true (and not previously done)
        ParseBlock( elsif_t, else_t );                     -- handle the elsif block
        handled := true;                                   -- remember we did it
     else                                                  -- otherwise
        SkipBlock( elsif_t, else_t );                      -- skip elsif block
     end if;
  end loop;

  -- else statements

  if token = else_t then                                   -- else part?
     expect( else_t );                                     -- "else"
     if not handled then                                   -- nothing handled yet?
        ParseBlock;                                        -- handle else block
     else                                                  -- otherwise
        SkipBlock;                                         -- skip else block
     end if;
  end if;

  -- end if

  expect( end_t );                                         -- "end if"
  expect( if_t );

end ParseIfBlock;

procedure ParseCaseBlock is
-- Syntax: case-block = "case" ident "is" "when" const-ident "=>" ...
-- "when others =>" ..."end case"
  test_id : identifier;
  case_id : identifier;
  handled : boolean := false;
  b       : boolean := false;
begin

  -- case id is

  expect( case_t );                                        -- "case"
  ParseIdentifier( test_id );                              -- identifier to test
  if class_ok( test_id, otherClass ) then
     expect( is_t );                                       -- "is"
  end if;

  -- when const-id =>

  if token /= when_t then                                 -- first when missing?
     expect( when_t );                                    -- force error
  end if;
  while token = when_t loop
     expect( when_t );                                     -- "when"
     exit when token = others_t;
     -- this should be ParseConstantIdentifier
     if token = strlit_t then
        if base_types_ok( identifiers( test_id ).kind, string_t ) then
           case_id := token;
           get_next_token;
        end if;
     elsif token = charlit_t then
        if base_types_ok( identifiers( test_id ).kind, character_t ) then
           case_id := token;
           get_next_token;
        end if;
     elsif token = number_t then
        if uni_types_ok( identifiers( test_id ).kind, uni_numeric_t ) then
           case_id := token;
           get_next_token;
        end if;
     else
        ParseIdentifier( case_id );                           -- get case
        if identifiers( case_id ).class /= constClass then    -- is constant?
           err( "variable not allowed as a case" );           -- error if not
        elsif base_types_ok( identifiers( test_id ).kind,
              identifiers( case_id ).kind ) then             -- types good?
           null;
        end if;
     end if;
     --
     if not error_found then
        b := identifiers( test_id ).value = identifiers( case_id ).value;
     end if;
     expect( symbol_t, "=>" );                             -- "=>"
     if b and not handled then                             -- handled yet?
        ParseBlock( when_t );                              -- if not, handle
        handled := true;                                   -- and remember done
     else
        SkipBlock( when_t );                               -- else skip case
     end if;
  end loop;

  -- others part

  if token /= others_t then                                -- a little clearer
     err( "when others expected" );                        -- if pointing at
  end if;                                                  -- end case
  expect( others_t );                                      -- "others"
  expect( symbol_t, "=>" );                                -- "=>"
  if not handled then                                      -- not handled yet?
     ParseBlock;                                           -- handle now
  else                                                     -- else just
     SkipBlock;                                            -- skip
  end if;

  -- end case

  expect( end_t );                                         -- "end case"
  expect( case_t );

end ParseCaseBlock;

procedure ParseLoopBlock is
  -- Syntax: loop-block = "loop" ... "end loop"
  exit_on_entry : boolean := exit_block;
begin

  pushBlock( newScope => false, newName => "loop loop" );  -- start new scope

  if syntax_check or exit_block then
     expect( loop_t );                                     -- "loop"
     ParseBlock;                                           -- check loop block
     goto loop_done;
  end if;

  loop
     expect( loop_t );                                     -- "loop"
     ParseBlock;                                           -- handle loop block
     exit when exit_block or error_found or token = eof_t;
     topOfBlock;                                           -- jump to top of block
  end loop;

<<loop_done>>
  pullBlock;                                               -- end of while scope
  if not syntax_check and not exit_on_entry then           -- ignore exit when checking
     if exit_block and not done then                          -- exiting and not returning?
        if trace then
           Put_trace( "exited loop" );
        end if;
        exit_block := false;                                  -- we handled exit_block
     end if;
  end if;

  expect( end_t );                                         -- "end loop"
  expect( loop_t );
end ParseLoopBlock;

procedure ParseWhileBlock is
  -- Syntax: while-block = "while" bool-expr "loop" ... "end loop"
  expr_val  : unbounded_string;
  expr_type : identifier;
  b : boolean := false;
  exiting : boolean := false;
  exit_on_entry : boolean := exit_block;
begin
  pushBlock( newScope => false, newName => "while loop" ); -- start new scope

  if syntax_check or exit_block then
     expect( while_t );                                    -- "while"
     ParseExpression( expr_val, expr_type );               -- expression
     if not base_types_ok( boolean_t, expr_type ) then     -- not boolean?
        err( "boolean expression expected" );
     end if;
     expect( loop_t );                                     --- "loop"
     ParseBlock;                                           -- check while block
     goto loop_done;
  end if;

  loop
     expect( while_t );                                    -- "while"
     ParseExpression( expr_val, expr_type );               -- expression
     if not base_types_ok( boolean_t, expr_type ) then     -- not boolean?
        err( "boolean expression expected" );
        exit;
     elsif expr_val /= "1" or error_found or exit_block then -- skipping?
        expect( loop_t );                                  -- "loop"
        SkipBlock;                                         -- skip while block
        exit;                                              -- and quit
     end if;                                               -- otherwise do loop
     if trace then
        put_trace( "expression is true" );
     end if;
     expect( loop_t );                                     -- "loop"
     ParseBlock;                                           -- handle while block
     exit when exit_block or error_found or token = eof_t;
     topOfBlock;                                           -- jump to top of block
  end loop;

<<loop_done>>
  pullBlock;                                               -- end of while scope
  if not syntax_check and not exit_on_entry then           -- ignore exit when checking
     if exit_block and not done then                          -- exiting and not returning?
        if trace then
           Put_trace( "exited while loop" );
        end if;
        exit_block := false;                                  -- we handled exit_block
     end if;
  end if;

  expect( end_t );                                         -- "end loop"
  expect( loop_t );
end ParseWhileBlock;

procedure ParseForBlock is
  -- Syntax: for-block = "for" local-var "in" expr ".." expr "loop" ... "end loop"
  expr1_val  : unbounded_string;
  expr1_type : identifier;
  expr2_val  : unbounded_string;
  expr2_type : identifier;
  expr2_num  : long_float;
  b : boolean := false;
  test1 : boolean := false;
  test2 : boolean := false;
  exiting : boolean := false;
  for_var   : identifier;
  firstTime : boolean := true;
  isReverse   : boolean := false;
  exit_on_entry : boolean := exit_block;
begin

  pushBlock( newScope => true, newName => "for loop" );  -- start new scope
  -- well, not strictly a new scope, but we'll need to do this
  -- to implement automatic declaration of the index variable

  if syntax_check or exit_block then
     -- this is complicated enough it should be in it's own nested procedure
     expect( for_t );                                      -- "for"
     if identifiers( token ).kind = new_t then          -- for var
        for_var := token;                               -- brand new? ok
     else                                               -- else declare locally
        declareIdent( for_var, identifiers( token ).name, uni_numeric_t);
     end if;
     get_next_token; 
     expect( in_t );                                    -- "in"
     if token = reverse_t then                          -- "reverse"?
        isReverse := true;
        expect( reverse_t );
     end if;
     ParseExpression( expr1_val, expr1_type );          -- low range
     expect( symbol_t, ".." );                          -- ".."
     ParseExpression( expr2_val, expr2_type );          -- high range
     if base_types_ok( expr1_type, expr2_type ) then      -- check types
        if get_uni_type( expr1_type ) = uni_numeric_t then
           null;
       elsif get_uni_type( expr1_type ) = root_enumerated_t then
           null;
       end if;
       if not error_found then
          if isReverse then
             identifiers( for_var ).kind := expr2_type;     -- this type
          else
             identifiers( for_var ).kind := expr1_type;     -- this type
           end if;
        end if;
     end if;
     expect( loop_t );
     ParseBlock;                                           -- check for block
     goto abort_loop;
  end if;

  loop
     expect( for_t );                                      -- "for"
     if firstTime then
        if identifiers( token ).kind = new_t then          -- for var
           for_var := token;                               -- brand new? ok
        else                                               -- else declare locally
           declareIdent( for_var, identifiers( token ).name, uni_numeric_t);
        end if;
        get_next_token; 
        expect( in_t );                                    -- "in"
        if token = reverse_t then                          -- "reverse"?
           isReverse := true;
           expect( reverse_t );
        end if;
        ParseExpression( expr1_val, expr1_type );          -- low range
        expect( symbol_t, ".." );                          -- ".."
        ParseExpression( expr2_val, expr2_type );          -- high range
        if error_found then                                -- errors?
            goto abort_loop;                               -- go no further
        end if;
        if base_types_ok( expr1_type, expr2_type ) then      -- check types
           if get_uni_type( expr1_type ) = uni_numeric_t then
              null;
           elsif get_uni_type( expr1_type ) = root_enumerated_t then
              null;
           else
              err( "numeric or enumerated type expected" );
              -- should be err_previous but haven't exported it yet
           end if;
           if not error_found then
              if isReverse then
                 identifiers( for_var ).value := expr2_val;     -- for var is
                 identifiers( for_var ).kind := expr2_type;     -- this type
                 if isExecutingCommand then
                    expr2_num := to_numeric( expr1_val );
                 end if;
              else
                 identifiers( for_var ).value := expr1_val;     -- for var is
                 identifiers( for_var ).kind := expr1_type;     -- this type
                 if isExecutingCommand then
                    expr2_num := to_numeric( expr2_val );
                 end if;
              end if;
           end if;
        end if;
        expect( loop_t );                                  -- "loop"
        firstTime := false;                                -- don't do this again
     else 
        -- don't interpret for line after first time
-- is this necessary any more?
        while token /= loop_t loop                         -- skip to
           get_next_token;                                 -- "loop"
        end loop;
        expect( loop_t );
        if isReverse then
           if isExecutingCommand then
              identifiers( for_var ).value := to_unbounded_string(
                  long_float( to_numeric( identifiers( for_var ).value ) - 1.0 ) );
           end if;
        else
           if isExecutingCommand then
              identifiers( for_var ).value := to_unbounded_string(
                  long_float( to_numeric( identifiers( for_var ).value ) + 1.0 ) );
           end if;
        end if;
     end if;
     if not isExecutingCommand then -- includes errors or exiting
put_line( "SKIPPING FOR BLOCK" );
        skipBlock;
        exit;
     elsif isReverse then
        if to_numeric( identifiers( for_var ).value ) < expr2_num then
           skipBlock;
           exit;
        end if;
     elsif to_numeric( identifiers( for_var ).value ) > expr2_num then
         skipBlock;
         exit;
     end if;
     if trace then
        put_trace(
            to_string( identifiers( for_var ).name ) & " := '" &
            to_string( identifiers( for_var ).value ) & "'" );
     end if;
     ParseBlock;                                           -- handle for block
     exit when exit_block or error_found or token = eof_t;
     topOfBlock;                                           -- jump to top of block
  end loop;

<<abort_loop>>
  pullBlock;                                               -- end of while scope
  if not syntax_check and not exit_on_entry then           -- ignore exit when checking
     if exit_block and not done then                          -- exiting and not returning?
        if trace then
           Put_trace( "exited for loop" );
        end if;
        exit_block := false;                                  -- we handled exit_block
     end if;
  end if;

  expect( end_t );                                         -- "end loop"
  expect( loop_t );

end ParseForBlock;

procedure ParseOpenFile( f : out identifier ) is
  -- standard output, standard error or a user file
  -- the file must be closed (ie value of variable is null)
  id : identifier := eof_t;
begin
  f := eof_t;
  if token = standard_output_t then
     f := token;
     get_next_token;
  elsif token = standard_error_t then
     f := token;
     get_next_token;
  elsif token = current_input_t then                   -- current input token
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardInput'img );
     get_next_token;
  elsif token = current_output_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardOutput'img );
     get_next_token;
  elsif token = current_error_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardError'img );
     get_next_token;
  else
     ParseIdentifier( id );
     -- if uni_types_ok( identifiers( id ).kind, file_type_t ) then
     if get_uni_type( identifiers( id ).kind ) /= file_type_t then
            err( "expected file_type" );
     else
        if not syntax_check then -- file is never open on a syntax check
           if length( identifiers( id ).value ) = 0 then
              err( "file not open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseOpenFile;

procedure ParseOpenSocket( f : out identifier ) is
  -- standard output, standard error or a user file
  -- the file must be closed (ie value of variable is null)
  id : identifier := eof_t;
begin
  f := eof_t;
  if token = standard_output_t then
     f := token;
     get_next_token;
  elsif token = standard_error_t then
     f := token;
     get_next_token;
  elsif token = current_input_t then                   -- current input token
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardInput'img );
     get_next_token;
  elsif token = current_output_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardOutput'img );
     get_next_token;
  elsif token = current_error_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardError'img );
     get_next_token;
  else
     ParseIdentifier( id );
     -- if uni_types_ok( identifiers( id ).kind, file_type_t ) then
     if get_uni_type( identifiers( id ).kind ) /= socket_type_t then
            err( "expected file_type or socket_type variable" );
     else
        if not syntax_check then -- file is never open on a syntax check
           if length( identifiers( id ).value ) = 0 then
              err( "file not open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseOpenSocket;

procedure ParseOpenFileOrSocket( f, kind : out identifier ) is
  -- standard output, standard error or a user file
  -- the file must be closed (ie value of variable is null)
  id : identifier := eof_t;
begin
  f := eof_t;
  if token = standard_output_t then
     f := token;
     get_next_token;
  elsif token = standard_error_t then
     f := token;
     get_next_token;
  elsif token = current_input_t then                   -- current input token
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardInput'img );
     get_next_token;
  elsif token = current_output_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardOutput'img );
     get_next_token;
  elsif token = current_error_t then
     f := token;
     identifiers( token ).value := to_unbounded_string( currentStandardError'img );
     get_next_token;
  else
     ParseIdentifier( id );
     kind := get_uni_type( identifiers( id ).kind );
     if kind /= file_type_t and kind /= socket_type_t then
            err( "expected file_type or socket_type variable" );
     else
        if not syntax_check then -- file is never open on a syntax check
           if length( identifiers( id ).value ) = 0 then
              err( "file not open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseOpenFileOrSocket;

procedure ParseClosedFile( f : out identifier ) is
  -- user file, must be closed (ie. value of variable not null)
  id : identifier;
begin
  f := eof_t;
  if token = standard_output_t then
     err( "file already open" );
  elsif token = standard_error_t then
     err( "file already open" );
  elsif token = current_input_t then
     err( "file already open" );
  elsif token = current_output_t then
     err( "file already open" );
  elsif token = current_error_t then
     err( "file already open" );
  else
     ParseIdentifier( id );
     if get_uni_type( identifiers( id ).kind ) /= file_type_t then
            err( "expected file_type variable" );
     else
        if not syntax_check then
           if length( identifiers( id ).value ) > 0 then
              err( "file already open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseClosedFile;

procedure ParseClosedSocket( f : out identifier ) is
  -- user file, must be closed (ie. value of variable not null)
  id : identifier;
begin
  f := eof_t;
  if token = standard_output_t then
     err( "file already open" );
  elsif token = standard_error_t then
     err( "file already open" );
  elsif token = current_input_t then
     err( "file already open" );
  elsif token = current_output_t then
     err( "file already open" );
  elsif token = current_error_t then
     err( "file already open" );
  else
     ParseIdentifier( id );
     if get_uni_type( identifiers( id ).kind ) /= socket_type_t then
            err( "expected socket_type variable" );
     else
        if not syntax_check then
           if length( identifiers( id ).value ) > 0 then
              err( "file already open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseClosedSocket;

procedure ParseClosedFileOrSocket( f, kind : out identifier ) is
  -- user file, must be closed (ie. value of variable not null)
  id : identifier;
begin
  f := eof_t;
  if token = standard_output_t then
     err( "file already open" );
  elsif token = standard_error_t then
     err( "file already open" );
  elsif token = current_input_t then
     err( "file already open" );
  elsif token = current_output_t then
     err( "file already open" );
  elsif token = current_error_t then
     err( "file already open" );
  else
     ParseIdentifier( id );
     kind := identifiers( id ).kind;
     if kind /= file_type_t and kind /= socket_type_t then
            err( "expected file_type or socket_type variable" );
     else
        if not syntax_check then
           if length( identifiers( id ).value ) > 0 then
              err( "file already open" );
           elsif not error_found then
              f := id;
           end if;
        end if;
     end if;
  end if;
end ParseClosedFileOrSocket;

-- "DO" procedures should be moved to parser_aux.

procedure DoGet( file : identifier ) is
-- Get the next character from a file or socket.  Save the character
-- in the ch_field field of the file record.  If there is no next
-- character, set the eof_field to true.  The caller is assumed to
-- check that the file is open.  There is no eof_field check.
--
-- Reasoning: UNIX/Linux has a terrible way to handle end-of-file:
-- you have to read one character too many and check to see if no
-- character was read.  As a result, Text_IO routines must always
-- be "double-buffered": they must read the character into a buffer,
-- and then the application must read the character from the buffer
-- to its final destination.  The end-of-file cannot be checked
-- without a read, and reading will cause characters to be lost if
-- they are not double-buffered.  But I didn't design it, did I?
  fd     : aFileDescriptor;    -- file's file descriptor
  ch     : character;          -- a buffer to read the character into
  eof    : boolean := false;   -- true if a character was read
  result : long_integer;       -- bytes read by read
begin
   fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
<<reread>> read( result, fd, ch, 1 );
   if result < 0 then                   -- a problem?
      if errno = EAGAIN  or errno = EINTR then
         goto reread;                   -- interrupted? try again
      end if;                           -- error? report it
      err( "unable to read file:" & OSerror( errno ) );
      return;                           -- and bail out
   elsif result = 0 then                -- nothing read?
      eof := true;                      -- then it's the end of file
   end if;
   if file /= standard_output_t and     -- not a special file?
      file /= standard_input_t and
      file /= standard_error_t and
      file /= current_output_t and
      file /= current_input_t and
      file /= current_error_t then
      if eof then                       -- eof? set eof_field
         replaceField( file, eof_field, "1" );
         replaceField( file, line_field, -- Ada counts EOF as a line!
            long_integer'image( long_integer'value(
            to_string( stringField( file, line_field ) ) ) + 1 ) );
      else                              -- else replace the character
         replace_Element( identifiers( file ).value, 1, ch ); -- save character in ch_field
         if ch = ASCII.LF then          -- a line? increment line_field
            replaceField( file, line_field,
               long_integer'image( long_integer'value(
               to_string( stringField( file, line_field ) ) ) + 1 ) );
            replaceField( file, eol_field, "1" ); -- and set eol_field
         else
            replaceField( file, eol_field, "0" ); -- else not
         end if;                        -- the end of the line
      end if;
   end if;
end DoGet;

procedure DoFileOpen( file : in out identifier;  mode : identifier; create : boolean;
  name : string ) is
  result : aFileDescriptor;
  flags  : anOpenFlag;
begin
  if create then
     flags := O_CREAT;
  else
     flags := 0;
  end if;
  if mode = in_file_t then
     result := open( name & ASCII.NUL, flags+O_RDONLY, 8#644# );
     -- OS-specific
     --if result < 0 then
     --   err( "Unable to open file: " & OSerror( errno ) );
     --end if;
  elsif mode = out_file_t then
     result := open( name & ASCII.NUL, flags+O_WRONLY+O_TRUNC, 8#644# );
     --if result < 0 and errno = 2 then
     --   err( "File must already exist to be opened for write" );
     --end if;
  elsif mode = append_file_t then
     result := open( name & ASCII.NUL, flags+O_WRONLY+O_APPEND, 8#644# );
     --if result < 0 and errno = 2 then
     --   err( "File must already exist to be opened for append" );
     --end if;
  else
     err( "internal error: unexpected file mode" );
  end if;
  if result < 0 then
     err( "Unable to open file: " & OSerror( errno ) );
  elsif not error_found then
     -- construct the file variable's value, a series of nul delimited fields
     identifiers( file ).value := to_unbounded_string( "." & ASCII.NUL );
     -- 1. character buffer
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string( result'img ) & ASCII.NUL;
     -- 2. file descriptor
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string( " 0" ) & ASCII.NUL;
     -- 3. lines
     --if mode = in_file_t then
     --   identifiers( file ).value := identifiers( file ).value & to_unbounded_string(
     --      isEOF( result )'img ) & ASCII.NUL;
     --else
        identifiers( file ).value := identifiers( file ).value & to_unbounded_string( "0" ) & ASCII.NUL;
     --end if;
     -- 4. eol flag
     identifiers( file ).value := identifiers( file ).value & name & ASCII.NUL;
     -- 5. name
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string(
        mode'img ) & ASCII.NUL;
     -- 6. mode
        identifiers( file ).value := identifiers( file ).value & to_unbounded_string( "0" ) & ASCII.NUL;
     -- 7. eof
     --end if;
     if mode = in_file_t then
        DoGet( file ); -- buffer first character, set eof if none
     end if;
     if trace then
        put_trace( to_string( identifiers( file ).name ) &
          " is file descriptor" & result'img );
     end if;
  end if;
end DoFileOpen;

procedure DoSocketOpen( file : in out identifier; name : unbounded_string ) is
  result : aSocketFD;
  --flags  : anOpenFlag;
  host   : unbounded_string;
  port   : integer;
  pos    : natural;
begin
  pos := index( name, ":" );
  if pos = 0 then
     host := name;
     port := 80;
  else
     begin
       host := to_unbounded_string( slice( name, 1, pos-1 ) );
     exception when others =>
       err( "unable to interpret TCP/IP host" );
     end;
     begin
       port := integer'value( " " & slice( name, pos+1, length( name ) ) );
     exception when others =>
       err( "unable to interpret TCP/IP port" );
     end;
     if port = 19 or port = 25 or port > 32767 then
        err( "access to this TCP/IP port is prohibited" );
     end if;
  end if;
  if error_found then
     return;
  end if;
  result := openSocket( host, port );
  if result < 0 then
     err( "Unable to socket file: " & OSerror( errno ) );
  elsif not error_found then
     -- construct the file variable's value, a series of nul delimited fields
     identifiers( file ).value := to_unbounded_string( " " & ASCII.NUL );
     -- 1. character buffer
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string( result'img ) & ASCII.NUL;
     -- 2. file descriptor
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string( " 0" ) & ASCII.NUL;
     -- 3. lines
     --if mode = in_file_t then
     --   identifiers( file ).value := identifiers( file ).value & to_unbounded_string(
     --      isEOF( result )'img ) & ASCII.NUL;
     --else
        identifiers( file ).value := identifiers( file ).value & to_unbounded_string( "0" ) & ASCII.NUL;
     --end if;
     -- 4. eol flag
     identifiers( file ).value := identifiers( file ).value & name & ASCII.NUL;
     -- 5. name
        identifiers( file ).value := identifiers( file ).value & to_unbounded_string( "1" ) & ASCII.NUL;
     --end if;
     -- 6. doGet flag
     identifiers( file ).value := identifiers( file ).value & to_unbounded_string( "0" ) & ASCII.NUL;
     --end if;
     -- 7. eof

     -- a socket cannot do an initial "DoGet" because we don't know if the user
     -- will be reading or writing first.  DoGet could cause a hang as the server
     -- is waiting for an instruction and the script is waiting for a response
     -- from the server.  As a result, we use a "DoGet" flag.  If DoGet is true,
     -- eof_field and ch_field are not valid until an initial DoGet is done.
     if trace then
        put_trace( to_string( identifiers( file ).name ) &
          " is file descriptor" & result'img );
     end if;
  end if;
end DoSocketOpen;

procedure ParseOpen( create : boolean := false ) is
  -- Syntax: open( closed-file, mode, name );
  -- Syntax: create( closed-file [,mode] [,name] );
  -- Source: Ada.Text_IO.Open
  -- Source: Ada.Text_IO.Create
  file : identifier;
  mode : identifier;
  name : unbounded_string;
  kind : identifier;
begin
  if create then
     if rshOpt then
        err( "create not allowed in a restricted shell" );
     end if;
     expect( create_t );
     expect( symbol_t, "(" );
     ParseClosedFile( file );
     kind := file_type_t;
     -- the mode is optional, default to out_file
     if token = symbol_t and identifiers( token ).value = ")" then
        mode := out_file_t;
     else
        expect( symbol_t, "," );
        ParseIdentifier( mode );
        if base_types_ok( identifiers( mode ).kind, file_mode_t ) then
           if create and mode = in_file_t then
              err( "Cannot create an in_file" );
           end if;
        end if;
     end if;
     -- the name is optional, default to a temp file name
     if token = symbol_t and identifiers( token ).value = ")" then
        makeTempFile( name );
     else
        expect( symbol_t, "," );
        if uni_types_ok( identifiers( token ).kind, uni_string_t ) then
           name := identifiers( token ).value;
           if length( name ) = 0 then
              err( "pathname should not be null" );
           end if;
           get_next_token;
        end if;
     end if;
     expect( symbol_t, ")" );
  else
     expect( open_t );
     expect( symbol_t, "(" );
     ParseClosedFileOrSocket( file, kind );
     if kind = file_type_t then
        expect( symbol_t, "," );
        ParseIdentifier( mode );
        if base_types_ok( identifiers( mode ).kind, file_mode_t ) then
           if rshOpt and mode = out_file_t then
              err( "out_file mode not allowed in a restricted shell" );
           end if;
        end if;
     end if;
     if not error_found then
        -- not error_found because file must be legit in here
        expect( symbol_t, "," );
        if kind = socket_type_t and get_base_type( identifiers( token ).kind ) = file_mode_t then
           err( "sockets don't have a mode" );
        elsif uni_types_ok( identifiers( token ).kind, uni_string_t ) then
           name := identifiers( token ).value;
           if length( name ) = 0 then
              err( "pathname should not be null" );
           end if;
           get_next_token;
           expect( symbol_t, ")" );
        end if;
     end if; -- if mode and file OK
  end if; -- is open
  -- do it
  if isExecutingCommand then -- should use umask for permissions
     if kind = file_type_t then
        DoFileOpen( file, mode, create, to_string( name ) );
     else
        DoSocketOpen( file, name );
     end if;
  end if;
end ParseOpen;

procedure ParseReset is
  -- Syntax: reset( open-file [,mode] )
  -- Source: Ada.Text_IO.Reset
  file : identifier;
  mode : identifier := eof_t;
  name : unbounded_string;
  modestr : unbounded_string;
  fd   : aFileDescriptor;
begin
  expect( reset_t );
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  if token = symbol_t and identifiers( token ).value = "," then
     expect( symbol_t, "," );
     if base_types_ok( identifiers( token ).kind, file_mode_t ) then
        mode := token;
        get_next_token;
     end if;
  end if;
  expect( symbol_t, ")" );
  if isExecutingCommand then
     fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
     name := stringField( file, name_field );
     if mode = eof_t then
        modestr := stringField( file, mode_field );
        if to_string( modestr ) = in_file_t'img then
           mode := in_file_t;
        elsif to_string( modestr ) = out_file_t'img then
           mode := out_file_t;
        elsif to_string( modestr ) = append_file_t'img then
           mode := append_file_t;
        else
           err( "internal error: unable to determine file mode " &
             to_string( modestr ) );
        end if;
     end if;
     close( fd );
     DoFileOpen( file, mode, false, to_string( name ) );
  end if;
end ParseReset;

procedure ParseClose is
  -- Syntax: close( open-file )
  -- Source: Ada.Text_IO.Close
  file : identifier;
  fd   : aFileDescriptor;
  kind : identifier;
begin
  expect( close_t );
  expect( symbol_t, "(" );
  ParseOpenFileOrSocket( file, kind );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
     if fd = currentStandardInput then
        err( "this file is the current input file" );
     elsif fd = currentStandardInput then
        err( "this file is the current output file" );
     elsif fd = currentStandardInput then
        err( "this file is the current error file" );
     else
        close( fd );
        if trace then
           put_trace( "Closed file descriptor" & to_string( stringField( file, fd_field ) ) );
        end if;
        identifiers( file ).value := null_unbounded_string;
     end if;
  end if;
end ParseClose;

procedure ParseDelete is
  -- Syntax: delete( open-file )
  -- Source: Ada.Text_IO.Delete
  file : identifier;
  name : unbounded_string;
  fd   : aFileDescriptor;
  result : integer;
begin
  expect( delete_t );
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
     if fd = currentStandardInput then
        err( "this file is the current input file" );
     elsif fd = currentStandardInput then
        err( "this file is the current output file" );
     elsif fd = currentStandardInput then
        err( "this file is the current error file" );
     else
        name := stringField( file, name_field );
        close( fd );
        identifiers( file ).value := null_unbounded_string;
        result := unlink( to_string( name ) & ASCII.NUL );
        if result /= 0 then
           err( "unable to delete file: " & OSerror( errno ) );
        end if;
        if trace then
           put_trace( "delete file " & to_string( name ) );
        end if;
     end if;
  end if;
end ParseDelete;

procedure ParseSkipLine is
  -- Syntax: skip_line [ (open-file) ]
  -- Source: Ada.Text_IO.Skip_Line
  file   : identifier := eof_t;
  --fd     : aFileDescriptor;
  ch     : character;
  result : long_integer;
  kind   : identifier;
  str    : unbounded_string;
begin
  expect( skip_line_t );
  --fd := stdin;
  if token = symbol_t and then identifiers( Token ).value = "(" then
      get_next_token;
      ParseOpenFileOrSocket( file, kind );
      expect( symbol_t, ")" );
  end if;
  if isExecutingCommand then
     --if file /= eof_t then
     --   fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
     --end if;

---
     if trace then
        put_trace( "Using file descriptor " & to_string( stringField( file, fd_field ) ) );
     end if;
     if file /= eof_t then
        if kind = socket_type_t and then stringField( file, doget_field ) = "1" then
           put_line( "Initial DoGet..." );
           DoGet( file );
           put_line( "Clearing flag..." );
           replaceField( file, doget_field, "0" );
           put_line( "done" );
        end if;
       loop
         ch := Element( identifiers( file ).value, 1 );
         if stringField(file, eof_field ) = "1" then
             err( "end of file" );
             exit;
         end if;
         DoGet( file );
         exit when ch = ASCII.LF or error_found;
         str := str & ch;
       end loop;
    else
      -- stdin (I don't like this)
     loop
        read( result, stdin, ch, 1 );
        if result < 0 then
           if errno /= EAGAIN and errno /= EINTR then
              err( "unable to read file:" & OSerror( errno ) );
              exit;
           end if;
        elsif result = 0 then
           err( "skipped past the end of the file" );
           exit;
        else
           if ch = ASCII.LF then
              exit;
           end if;
        end if;
        str := str & ch;
      end loop;
    end if;
    if trace then
       Put_Trace( "Skipped '" & to_string( str ) & "'" );
    end if;
  end if;
end ParseSkipLine;

procedure ParseGet is
  -- Syntax: get [ (open-file), ch ]
  -- Source: Ada.Text_IO.Get
  -- Note: Gnat get can't be used here because it does something
  -- odd with file descriptor 0
  file      : identifier := eof_t;
  kind      : identifier;
  fd        : aFileDescriptor;
  ch        : character;
  id        : identifier;
  result    : long_integer;
begin
  expect( get_t );
  fd := stdin;
  expect( symbol_t, "(" );
  if identifiers( token ).kind /= keyword_t then
     ParseOpenFileOrSocket( file, kind );
     expect( symbol_t, "," );
  else
     file := standard_input_t;
  end if;
  ParseIdentifier( id );
  if base_types_ok( identifiers( id ).kind, character_t ) then
     expect( symbol_t, ")" );
  end if;
  if isExecutingCommand then
     if trace then
        put_trace( "Using file descriptor " & to_string( stringField( file, fd_field ) ) );
     end if;
     if file /= eof_t then
        if kind = socket_type_t and then stringField( file, doget_field ) = "1" then
           DoGet( file );
           replaceField( file, doget_field, false'img );
        end if;
        if stringField( file, eof_field ) = "1" then
           err( "end of file" );
        else
           ch := Element( identifiers( file ).value, 1 );
           identifiers( id ).value := to_unbounded_string( "" & ch );
           DoGet( file );
        end if;
     else
        fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
   <<reread>> read( result, fd, ch, 1 );
         if result < 0 then
              if errno = EAGAIN  or errno = EINTR then
                 goto reread;
              end if;
              err( "unable to read file:" & OSerror( errno ) );
         elsif result = 0 then
            err( "end of file" );
         else
            identifiers( id ).value := to_unbounded_string( "" & ch );
         end if;
      end if;
      if file /= standard_output_t and
         file /= standard_input_t and
         file /= standard_error_t and
         file /= current_output_t and
         file /= current_input_t and
         file /= current_error_t then
         if ch = ASCII.LF then -- not stdin (or error)?
            replaceField( file, line_field,
               long_integer'image( long_integer'value(
               to_string( stringField( file, line_field ) ) ) + 1 ) );
            replaceField( file, eol_field, "1" );
         else
            replaceField( file, eol_field, "0" );
         end if;
      end if;
  end if;
end ParseGet;

procedure ParsePutLine is
  -- Syntax: put_line( [file,] expression )
  -- Source: Ada.Text_IO.Put_Line
  target    : Identifier;
  kind      : identifier := file_type_t;
  stderr    : boolean := false;
  expr_val  : unbounded_string;
  expr_type : identifier;
  result    : long_integer;
  ch        : character;
  fd        : aFileDescriptor;
begin
  expect( put_line_t );
  expect( symbol_t, "(" );
  if identifiers( token ).kind /= keyword_t then
     kind := get_uni_type( token );
     if kind = file_type_t then
        ParseOpenFile( target );
        expect( symbol_t, "," );
     elsif kind = socket_type_t then
        ParseOpenSocket( target );
        expect( symbol_t, "," );
     else
        target := standard_output_t;
     end if;
     --ParseOpenFileOrSocket( target, kind );
  else
     target := standard_output_t;
  end if;
  ParseExpression( expr_val, expr_type );
  -- this sould be moved to an image function
  if get_uni_type( expr_type ) = root_enumerated_t then
     for i in identifiers'first..identifiers_top-1 loop
         if identifiers( i ).kind = expr_type then
            if identifiers( i ).value = expr_val then
               expr_val := identifiers( i ).name;
               exit; -- first occurrence should be original enumerated value
            end if;
         end if;
     end loop;
  end if;
  expect( symbol_t, ")" );
  if isExecutingCommand then
     if target = standard_error_t then
        Put_Line( standard_error, expr_val );
     elsif target = standard_output_t then
        Put_Line( expr_val );
        last_output := expr_val;
        last_output_type := expr_type;
     else
        fd := aFileDescriptor'value( to_string( stringField( target, fd_field ) ) );
        for i in 1..length( expr_val ) loop
            ch := Element( expr_val, i );
<<rewrite>> write( result, fd, ch, 1 );
            if result < 0 then
               if errno = EAGAIN or errno = EINTR then
                  goto rewrite;
               end if;
               err( "unable to write: " & OSerror( errno ) );
               exit;
            end if;
        end loop;
        ch := ASCII.LF;
        write( result, fd, ch, 1 ); -- add a line feed
<<rewrite2>> if result < 0 then
           if errno = EAGAIN or errno = EINTR then
              goto rewrite2;
           end if;
           err( "unable to write: " & OSerror( errno ) );
         else
           -- kludge: need a real file variable to record line count
           if target /= standard_output_t and
              target /= standard_input_t and
              target /= standard_error_t and
              target /= current_output_t and
              target /= current_input_t and
              target /= current_error_t then
              replaceField( target, line_field,
                 long_integer'image( long_integer'value(
                    to_string( stringField( target, line_field ) ) ) + 1 ) );
           end if;
        end if;
     end if;
  end if;
end ParsePutLine;

procedure ParseQuestion is
  -- Syntax: "?" expression
  -- Source: BUSH built-in
  expr_val  : unbounded_string;
  expr_type : identifier;
begin
  expect( symbol_t );
  if onlyAda95 then
      err( "pragma ada_95 doesn't allow ?" );
      return;
  end if;
  ParseExpression( expr_val, expr_type );
  -- this sould be moved to an image function
  if get_uni_type( expr_type ) = root_enumerated_t then
     for i in identifiers'first..identifiers_top-1 loop
         if identifiers( i ).kind = expr_type then
            if identifiers( i ).value = expr_val then
               expr_val := identifiers( i ).name;
               exit; -- first occurrence should be original enumerated value
            end if;
         end if;
     end loop;
  end if;
  if isExecutingCommand then -- fix this for no output on error!
     Put_Line( expr_val );
     last_output := expr_val;
     last_output_type := expr_type;
  end if;
end ParseQuestion;

procedure ParsePut is
  -- Syntax: put( [open-file,] expression [, picture] )
  -- Source: Ada.Text_IO.Editing.Put
  target    : Identifier;
  kind      : identifier;
  stderr    : boolean := false;
  expr_val  : unbounded_string;
  expr_type : identifier;
  result    : long_integer;
  ch        : character;
  fd        : aFileDescriptor;
  pic       : Picture;
  pic_val   : unbounded_string;
  pic_type  : identifier;
begin
  expect( put_t );
  expect( symbol_t, "(" );
  if identifiers( token ).kind /= keyword_t then
     kind := get_uni_type( token );
     if kind = file_type_t then
        ParseOpenFile( target );
        expect( symbol_t, "," );
     elsif kind = socket_type_t then
        ParseOpenSocket( target );
        expect( symbol_t, "," );
     else
        target := standard_output_t;
     end if;
  else
     target := standard_output_t;
  end if;
  ParseExpression( expr_val, expr_type );
  -- this sould be moved to an image function
  if get_uni_type( expr_type ) = root_enumerated_t then
     for i in identifiers'first..identifiers_top-1 loop
         if identifiers( i ).kind = expr_type then
            if identifiers( i ).value = expr_val then
               expr_val := identifiers( i ).name;
               exit; -- first occurrence should be original enumerated value
            end if;
         end if;
     end loop;
  end if;
  -- apply optional numeric formatting
  if token = symbol_t and identifiers( token ).value = "," then
     expect( symbol_t, "," );
     ParseExpression( pic_val, pic_type );
     if get_uni_type( pic_type ) /= uni_string_t then
        err( "number format picture string expected" );
     elsif not valid( to_string( pic_val ) ) then
        err( "number not a valid format picture" );
     elsif get_uni_type( expr_type ) /= uni_numeric_t then
        err( "only numeric types can use a format picture" );
     else
        pic := to_picture( to_string( pic_val ) );
        expr_val := to_unbounded_string( image( decimal_output_type( to_numeric( expr_val ) ), pic ) );
        last_output_type := uni_string_t;
     end if;
  else
     last_output_type := expr_type;
  end if;
  expect( symbol_t, ")" );
  if isExecutingCommand then
     if target = standard_error_t then
        Put( standard_error, expr_val );
     elsif target = standard_output_t then
        Put( expr_val );
        last_output := expr_val;
     else
        fd := aFileDescriptor'value( to_string( stringField( target, fd_field ) ) );
        for i in 1..length( expr_val ) loop
            ch := Element( expr_val, i );
            write( result, fd, ch, 1 );
<<rewrite>> if result < 0 then
               if errno = EAGAIN or errno = EINTR then
                  goto rewrite;
               end if;
               err( "unable to write: " & OSerror( errno ) );
               exit;
            end if;
        end loop;
     end if;
  end if;
end ParsePut;

procedure ParseNewLine is
  -- Syntax: new_line
  -- Source: Ada.Text_IO.New_Line
  target : identifier;
  kind   : identifier;
  fd     : aFileDescriptor;         -- Linux file descriptor of output file
  ch     : character;
  result : long_integer;
begin
  expect( new_line_t );
  if token = symbol_t and identifiers( token ).value = "(" then
     expect( symbol_t, "(" );
     ParseOpenFileOrSocket( target, kind );
     expect( symbol_t, ")" );
  else
     target := standard_output_t;
  end if;
  if isExecutingCommand then
     if target = standard_error_t then
        New_Line( standard_error );
     elsif target = standard_output_t then
        New_Line;
     else
        fd := aFileDescriptor'value( to_string( stringField( target, fd_field ) ) );
        ch := ASCII.LF;
<<rewrite>> write( result, fd, ch, 1 );
        if result < 0 then
           if errno = EAGAIN or errno = EINTR then
              goto rewrite;
           end if;
           err( "unable to write: " & OSerror( errno ) );
        end if;
     end if;
  end if;
end ParseNewLine;

procedure ParseSetInput is
  -- Syntax: set_input( open-file )
  -- Source: Ada.Text_IO.Set_Input
  file   : identifier;              -- open file to assign output to
  fd     : aFileDescriptor;         -- Linux file descriptor of output file
  result : aFileDescriptor := 0;    -- result of dup2
begin
  expect( set_input_t );
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     if file = standard_output_t then
        err( "standard_output cannot be assigned for input" );
     elsif file = standard_error_t then
        err( "standard_error cannot be assigned for input" );
     else
        fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
        result := dup2( fd, stdin );
     end if;
     if result < 0 then
        err( "unable to set input: " & OSerror( errno ) );
     elsif not error_found then
        currentStandardInput := fd;
        if trace then
           put_trace( "input is currently from file descriptor" &
             currentStandardInput'img );
        end if;
     end if;
  end if;
end ParseSetInput;

procedure ParseSetOutput is
  -- Syntax: set_output( open-file )
  -- Source: Ada.Text_IO.Set_Output
  file   : identifier;              -- open file to assign output to
  fd     : aFileDescriptor;         -- Linux file descriptor of output file
  result : aFileDescriptor := 0;    -- result of dup2
begin
  expect( set_output_t );
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     if file = standard_input_t then
        err( "standard_input cannot be assigned for output" );
     else
        fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
        result := dup2( fd, stdout );
     end if;
     if result < 0 then
        err( "unable to set output: " & OSerror( errno ) );
     elsif not error_found then
        currentStandardOutput := fd;
        if trace then
           put_trace( "output is currently to file descriptor" &
             currentStandardOutput'img );
        end if;
     end if;
  end if;
end ParseSetOutput;

procedure ParseSetError is
  -- Syntax: set_error( open-file )
  -- Source: Ada.Text_IO.Set_Error
  file   : identifier;
  result : aFileDescriptor := 0;
  fd     : aFileDescriptor;
begin
  expect( set_error_t );
  expect( symbol_t, "(" );
  ParseOpenFile( file );
  expect( symbol_t, ")" );
  if isExecutingCommand then
     if file = standard_input_t then
        err( "standard_input cannot be assigned for standard error" );
     else
        fd := aFileDescriptor'value( to_string( stringField( file, fd_field ) ) );
        result := dup2( fd, stderr );
     end if;
     if result < 0 then
        err( "unable to set error: " & OSerror( errno ) );
     elsif not error_found then
        currentStandardError := fd;
        if trace then
           put_trace( "error output is currently to file descriptor" &
             currentStandardError'img );
        end if;
     end if;
  end if;
end ParseSetError;

procedure ParseSystem is
  -- Syntax: system( string )
  -- Source: BUSH built-in
  expr_val  : unbounded_string;
  expr_type : identifier;
begin
  expect( system_t );
  expect( symbol_t, "(" );
  ParseExpression( expr_val, expr_type );
  expect( symbol_t, ")" );
  if uni_types_ok( uni_string_t,  expr_type ) then
     if isExecutingCommand then
        last_status:= linux_system( to_string( expr_val ) & ascii.nul );
     end if;
  end if;
end ParseSystem;

procedure ParseJobs is
  -- Syntax: jobs
  -- Source: BUSH built-in
begin
  expect( jobs_t );
  if isExecutingCommand then
     -- not actually in built-ins package
     putJobList;
  end if;
end ParseJobs;

procedure ParseDelay is
  -- Syntax: delay expression
  -- Source: Ada built-in
  expr_val  : unbounded_string;
  expr_type : identifier;
begin
  expect( delay_t );
  ParseExpression( expr_val, expr_type );
  if base_types_ok( expr_type, duration_t ) then
     if isExecutingCommand then
        begin
          delay duration( to_numeric( expr_val ) );
        exception when others =>
          err( "exception raised" );
        end;
        if trace then
           put_trace( "duration := " & to_string( expr_val ) );
        end if;
     end if;
  end if;
end ParseDelay;

procedure ParseUnset is
  -- Syntax: unset identifier
  -- Source: BUSH built-in
  id : identifier;
begin
   expect( unset_t );
   if onlyAda95 then
      discardUnusedIdentifier( token );
      err( "pragma ada_95 doesn't allow unset" );
      return;
   elsif inputMode /= interactive then
      discardUnusedIdentifier( token );
      err( "unset only allowed in an interactive session" );
      return;
   end if;
   ParseIdentifier( id );
   if rshOpt and then identifiers( id ).name = "PATH" then
      err( "unsetting PATH not allowed in restricted shell" );
   elsif isExecutingCommand then
      if not deleteIdent( id ) then
         err( "unable to unset identifier" );
       end if;
   end if;
end ParseUnset;

procedure ParseTypeset is
  -- Syntax: set identifier
  -- Source: BUSH built-in
  id     : identifier;
  typeid : identifier := eof_t;
  b      : boolean;
begin
   expect( typeset_t );
   if onlyAda95 then
      discardUnusedIdentifier( token );
      err( "pragma ada_95 doesn't allow typeset" );
      return;
   elsif inputMode /= interactive then
      discardUnusedIdentifier( token );
      err( "typeset only allowed in an interactive session" );
      return;
   end if;
   if identifiers( token ).kind = new_t then
      ParseNewIdentifier( id );
   else
      ParseIdentifier( id );
   end if;
   if token = is_t then
      expect( is_t );
      ParseIdentifier( typeid );
   end if;
   if isExecutingCommand then
      if typeid = eof_t then
         identifiers( id ).kind := universal_t;
      else
         identifiers( id ).kind := typeid;
      end if;
   else
      b := deleteIdent( id );
   end if;
end ParseTypeset;

procedure ParseEnvironment is
  -- Syntax: env
  -- Source: BUSH built-in
  ident : identifier := eof_t;
begin
  expect( env_t );
  if not( token = symbol_t and identifiers( token ).value = ";" ) then
     -- ParseIdentifier( ident );
     -- Don't use ParseIdent: take anything from symbol table
     if identifiers( token ).kind = new_t then
        ParseIdentifier( ident ); -- force error/typo suggestion
     else
        ident := token;
        get_next_token;
     end if;
  end if;
  if isExecutingCommand then
     builtins.env( ident );
  end if;
end ParseEnvironment;

procedure ParseWait is
  -- Syntax: wait
  -- Source: BUSH built-in
begin
  expect( wait_t );
  if isExecutingCommand then
     wait4children;
  end if;
end ParseWait;

procedure ParseClear is
  -- Syntax: clear
  -- Source: BUSH built-in
begin
  expect( clear_t );
  if isExecutingCommand then
     put( term_reset );
     put( term_clear );
  end if;
end ParseClear;

procedure ParseTrace is
  -- Syntax: trace [true | false]
  -- Source: BUSH built-in
begin
  expect( trace_t );
  if token = symbol_t and identifiers( token ).value = ";" then
     if isExecutingCommand then
        if trace then
           put_line( "Trace is currently on" );
        else
           put_line( "Trace is currently off" );
        end if;
     end if;
  elsif base_types_ok( boolean_t, identifiers( token ).kind ) then
     if isExecutingCommand then -- or syntax_check then -- when debugging
        if token = true_t then
           trace := true;
           put_line( "Trace is on" );
        else
           trace := false;
           put_line( "Trace is off" );
        end if;
      end if;
      get_next_token;
  end if;
end ParseTrace;

procedure ParseHelp is
  -- Syntax: help = "help or help ident"
  -- Source: BUSH built-in
begin
  expect( help_t );
  --if not error_found and not exit_block then
  if token = symbol_t and identifiers( token ).value = ";" then
     Put_Line( "The following topics are available:" );
     New_Line;
     Put_Line( "  case              expressions     new_line            strings.match    " );
     Put_Line( "  cd                for             null                subtype          " );
     Put_Line( "  clear             get             numerics            system           " );
     Put_Line( "  close             get_line        open                trace            " );
     Put_Line( "  command           if              pragma              type             " );
     Put_Line( "  create            inkey           put                 typeset          " );
     Put_Line( "  declare           is_open         put_line            unset            " );
     Put_Line( "  delay             jobs            pwd                 variables        " );
     Put_Line( "  delete            line            reset               wait             " );
     Put_Line( "  end_of_file       logout          return              while            " );
     Put_Line( "  end_of_line       loop            set_input           ?                " );
     Put_Line( "  env               mode            skip_line           keys             " );
     Put_Line( "  exit              name            strings             # / -- a comment " );
     return;
  end if;
  if token = case_t then
     Put_Line( "case - test multiple cases" );
     Put_Line( "  case var is when literal|const => ... when others => ... end case" );
  elsif token = cd_t then
     Put_Line( "cd - change directory" );
     Put_Line( "  - - the previous directory" );
  elsif token = clear_t then
     Put_Line( "clear - reset tty device and clear the screen" );
  elsif token = close_t then
     Put_Line( "close - close an open file" );
     Put_Line( "  close( file )" );
  elsif token = command_t then
     Put_Line( "command - run a Linux command (instead of a built-in command)" );
     Put_Line( "  command cmd" );
  elsif token = create_t then
     Put_Line( "create - create a new file or overwrite an existing file" );
     Put_Line( "  create( file [, out_file | append_file] [, path ] )" );
     Put_Line( "The default type is out_file" );
     Put_Line( "The default path a temporary file name" );
  elsif token = declare_t then
     Put_Line( "declare - begin a new block" );
     Put_Line( "  [declare declarations] begin ... end" );
  elsif token = delay_t then
     Put_Line( "delay - wait (sleep) for a specific time" );
     Put_Line( "  delay secs" );
  elsif token = delete_t then
     Put_Line( "delete - close and delete a file" );
     Put_Line( "  delete( file )" );
  elsif token = end_of_file_t then
     Put_Line( "end_of_file - true if an in_file file has no more data" );
     Put_Line( "  end_of_file( file )" );
  elsif token = end_of_line_t then
     Put_Line( "end_of_line - true if an in_file file has reached the end of a line with get" );
     Put_Line( "  end_of_line( file )" );
  elsif token = env_t then
     Put_Line( "env - show an identifier/all identifiers" );
     Put_Line( "  env [identifier]" );
  elsif token = exit_t then
     Put_Line( "exit - break out of a loop" );
     Put_Line( "  exit or exit when condition" );
  elsif token = for_t then
     Put_Line( "for - for loop" );
     Put_Line( "  for var in [reverse] first..last loop ... end loop" );
  elsif token = get_t then
     Put_Line( "get - read a character from a line of text" );
     Put_Line( "  get ([file,] ch)" );
  elsif token = get_line_t then
     Put_Line( "get_line - read a line of text" );
     Put_Line( "  var := get_line [ (file) ]" );
  elsif identifiers( token ).name = "keys" then
     Put_Line( "Keyboard Keys:" );
     New_Line;
     Put_Line( "Emacs Mode:                          vi Mode:" );
     Put_Line( "  control-b - move backwards           J - move backwards" );
     Put_Line( "  control-f - move forwards            K - move forwards" );
     Put_Line( "  control-p - move up                  I - move up" );
     Put_Line( "  control-n - move down                M - move down" );
     Put_Line( "  control-x - erase line" );
     Put_Line( "  control-a - move to start            ^ - move to start" );
     Put_Line( "  control-e - move to end              $ - move to end" );
     Put_Line( "  control-r - search history" );
     Put_Line( "  control-] - character search" );
     discardUnusedIdentifier( token );
  elsif token = if_t then
     Put_Line( "if - conditional execution" );
     Put_Line( "  if expression then ... [elsif expr then ...] [ else ...] " );
  elsif token = inkey_t then
     Put_Line( "inkey - read a character from standard input without echoing" );
     Put_Line( "  c := inkey" );
  elsif token = is_open_t then
     Put_Line( "is_open - true if file is open" );
     Put_Line( "  is_open( file )" );
  elsif token = jobs_t then
     Put_Line( "jobs - list current status of commands running in the background" );
  elsif token = line_t then
     Put_Line( "line - the number of read/written lines" );
     Put_Line( "  line( file )" );
  elsif token = logout_t then
     Put_Line( "logout - exit shell interactive session" );
  elsif token = loop_t then
     Put_Line( "loop - general loop" );
     Put_Line( "  loop ... end loop" );
  elsif token = new_line_t then
     Put_Line( "new_line - start a new line" );
     Put_Line( "  new_line [(file)]" );
  elsif token = match_t then
     Put_Line( "strings.match - pattern matching with UNIX V7 regular expressions and PERL extensions" );
     Put_Line( "  bool := match( expression, string )" );
     New_Line;
     Put_Line( "^    - at beginning            .   - any character" );
     Put_Line( "$    - at end                  ?   - zero or one character" );
     Put_Line( "[s]  - any in set s            +   - one or more characters" );
     Put_Line( "[^s] - any not in set s        *   - zero or more characters" );
     Put_Line( "\    - escape character        (e) - nested expression" );
     Put_Line( "|    - alternative                                    " );
     New_Line;
     Put_Line( "Regular expressions are described in ""man 5 regexp""" );
  elsif token = mode_t then
     Put_Line( "mode - the file mode (in_file, out_file, append_file)" );
     Put_Line( "  mode( file )" );
  elsif token = name_t then
     Put_Line( "name - name of an open file" );
     Put_Line( "  name( file )" );
  elsif token = null_t then
     Put_Line( "null - do nothing" );
  elsif token = open_t then
     Put_Line( "open - open an existing file or open a socket" );
     Put_Line( "  open( file, in_file | out_file | append_file, path )" );
  elsif token = pragma_t then
     Put_Line( "pragma - interpreter directive" );
     New_Line;
     Put_Line( "  pragma ada_95 - enforce Ada 95 restrictions" );
     Put_Line( "  pragma assert( condition ) - terminate program on condition" );
     Put_Line( "  pragma export( shell, var ) - export an environment variable" );
     Put_Line( "  pragma import( shell, var ) - import an environment variable" );
     Put_Line( "  pragma volatile( var ) - load value from environment on every access" );
     Put_Line( "  pragma restriction( no_auto_declarations ) - disable automatic command line declarations" );
  elsif token = put_t then
     Put_Line( "put - write to output, no new line" );
     Put_Line( "  put ( [file], expression [,picture] )" );
  elsif token = put_line_t then
     Put_Line( "put_line - write to output and start new line" );
     Put_Line( "  put_line ( [file], expression )" );
  elsif token = pwd_t then
     Put_Line( "pwd - present working directory" );
  elsif token = reset_t then
     Put_Line( "reset - reopen a file" );
     Put_Line( "  reset( file [,mode]" );
  elsif token = return_t then
     Put_Line( "return - exit script and return status code" );
     Put_Line( "  return [status code]" );
  elsif token = typeset_t then
     Put_Line( "typeset - change the type of a variable, declaring it if necessary" );
     Put_Line( "  typeset var [is type]" );
     Put_Line( "The default type is universal_typeless" );
  elsif token = set_input_t or token = set_output_t or token = set_error_t then
     Put_Line( "set_input/output/error - input/output redirection" );
     Put_Line( "  set_input( file )" );
     Put_Line( "  set_output( file )" );
     Put_Line( "  set_error( file )" );
     Put_Line( "The default files are standard_input, standard_output and standard_error" );
  elsif token = skip_line_t then
     Put_Line( "skip_line - discard the next line of input" );
     Put_Line( "  skip_line [(file)]" );
  elsif token = subtype_t then
     Put_Line( "subtype - create an alias for a type" );
     Put_Line( "  subtype newtype is oldtype" );
  elsif token = system_t then
     Put_Line( "system (command) - run a BASH shell command" );
     Put_Line( "  system( commandstring )" );
     New_Line;
     Put_Line( "System (package) - System package constants" );
     New_Line;
     Put_Line( "  System.System_Name            System.Fine_Delta" );
     Put_Line( "  System.Max_Int                System.Tick" );
     Put_Line( "  System.Min_Int                System.Storage_Unit" );
     Put_Line( "  System.Max_Binary_Modulus     System.Word_Size" );
     Put_Line( "  System.Max_Nonbinary_Modulus  System.Memory_Size" );
     Put_Line( "  System.Max_Base_Digits        System.Default_Bit_Order" );
     Put_Line( "  System.Max_Mantissa" );
  elsif token = trace_t then
     Put_Line( "trace - show verbose debugging information" );
     Put_Line( "  trace true or false" );
     Put_Line( "  trace - show current trace setting" );
  elsif token = type_t then
     Put_Line( "type - declare a new type or enumerated type" );
     Put_Line( "  type newtype is new oldtype" );
     Put_Line( "  type newenum is ( enum1 [,enum2...])" );
     New_Line;
     Put_Line( "Built-in types:" );
     Put_Line( " boolean     integer            natural            short_short_integer     " );
     Put_Line( " character   long_float         positive           socket_type" );
     Put_Line( " duration    long_integer       short_float        string" );
     Put_Line( " file_type   long_long_float    short_integer      universal_numeric" );
     Put_Line( " file_mode   long_long_integer  short_short_float  universal_string" );
     Put_Line( " float                                             universal_typeless" );
  elsif token = unset_t then
     Put_Line( "unset - delete an identifier" );
     Put_Line( "  unset ident" );
  elsif token = wait_t then
     Put_Line( "wait - wait for all background commands to finish" );
     Put_Line( "  wait" );
  elsif token = while_t then
     Put_Line( "while - while loop" );
     Put_Line( "  while expression loop ... end loop" );
  elsif identifiers( token ).name = "expressions" then
     Put_Line( "expressions" );
     New_Line;
     Put_Line( "+ - not - uniary operations" );
     New_Line;
     Put_Line( "** - exponentiation" );
     New_Line;
     Put_Line( "* - multiplication             and - bitwise and" );
     Put_Line( "/ - division                   or  - bitwise or" );
     Put_line( "& - string concatenation       xor - bitwise xor" );
     New_Line;
     Put_Line( "+ - addition                   - - subtraction" );
     New_Line;
     Put_Line( "= /= > >= < <= - relational operators" );
     New_Line;
     Put_Line( "and or xor - boolean operators" );
     New_Line;
     Put_Line( "@ - itself                    % - last put" );
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "numerics" then
     Put_Line( "numerics - numerics package constants/functions" );
     New_Line;
     Put_Line( "  e                    log2_e                  log10_e             ln10" );
     Put_Line( "  ln2                  pi                      pi_by_2             pi_by_4" );
     Put_Line( "  pi_under_1                                   pi_under_2" );
     Put_Line( "  sqrt_pi_under_2                              sqrt_2" );
     Put_Line( "  sqrt_2_under_1" );
     New_Line;
     Put_Line( "  sqrt( x )            log( x [,base] )         exp( x )           random" );
     Put_Line( "  shift_left( x, b )   rotate_left( x, b )      shift_right_arithmeric( x, b )" );
     Put_Line( "  shift_right( x, b )  rotate_right( x, b )     floor( x )         ceiling( x )" );
     Put_Line( "  rounding( x )        truncation( x )          unbiased_rounding( x )         " );
     Put_Line( "  remainder( x, y )    copy_sign( x, y )        leading_part( x, y )           " );
     Put_Line( "  exponent( x )        fraction( x )            max( m, n )         min( m, n )" );
     Put_Line( "  machine( x )         scaling( x, y )          value( s )          pos( c )   " );
     Put_Line( "  sturges( l, h, t )                                                           " );
     New_Line;
     Put_Line( "  sin( x [,cycle] )    arcsin( x [,cycle] )     sinh( x )          arcsinh( x )" );
     Put_Line( "  cos( x [,cycle] )    arccos( x [,cycle] )     cosh( x )          arccosh( x )" );
     Put_Line( "  tan( x [,cycle] )    arctan( x, y [,cycle] )  tanh( x )          arctanh( x )" );
     Put_Line( "  cot( x [,cycle] )    arccot( x, y [,cycle] )  coth( x )          arccoth( x )" );
     New_Line;
     Put_Line( "  abs( x ) is not ""numerics.abs( x )"" because it's not in the numerics package" );
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "strings" then
     Put_Line( "strings - strings package functions" );
     New_Line;
     Put_Line( "  element( s, i )           slice( s, l, h )         index( s, p )" );
     Put_Line( "  index_non_blank( s )      count( s, p )            replace_slice( s, l, h, b )");
     Put_Line( "  insert( s, b, n )         overwrite( s, p, n )     delete( s, l, h )" );
     Put_Line( "  trim( s )                 head( s, c )             tail( s, c )" );
     Put_Line( "  length( s )               match( e, s )            glob( e, s ) " );
     Put_Line( "  image( s )                val( n )                 field( s, d, c ) " );
     Put_Line( "  mktemp( p )                                                     " );
     discardUnusedIdentifier( token );
  elsif identifiers( token ).name = "variables" then
     Put_Line( "Variables" );
     Put_Line( "  var1 [, var2...] : [constant ] type [ := expression ] - declaration" );
     Put_Line( "  var := expression - assignment" );
     discardUnusedIdentifier( token );
  elsif identifiers( token ).value = "?" then
     Put_Line( "? - put_line to standard output" );
     Put_Line( "  ? expression" );
  else
     Put_Line( "Not a BUSH command...trying manual entries" );
     delay 2.0;
     declare
       status : integer;
     begin
       status := linux_system( "man "& to_string( identifiers( token ).name ) );
     end;
     discardUnusedIdentifier( token );
  end if;
  get_next_token;
end ParseHelp;

procedure ParseDeclarations is
  -- Syntax: declaration = "new-ident decl-part"
  var_id : identifier;
begin
  while token /= begin_t and token /= eof_t loop
     if token = type_t then
        ParseType;
        expect( symbol_t, ";" );
     elsif token = subtype_t then
        ParseSubtype;
        expect( symbol_t, ";" );
     else
        ParseNewIdentifier( var_id );
        ParseDeclarationPart( var_id );
        expect( symbol_t, ";" );
     end if;
  end loop;
end ParseDeclarations;

procedure SkipBlock( termid1, termid2 : identifier := keyword_t ) is
  old_error : boolean;
begin
  if token = end_t or token = eof_t or token = termid1 or token = termid2 then
     err( "missing statement" );
  end if;
  if syntax_check then               -- if we're checking syntax
     ParseBlock( termid1, termid2 ); -- must process the block to look
     return;                         -- for syntax errors
  end if;
  --old_error := error_found;          -- save error code
  --error_found := true;               -- skip by setting error flag
  old_error := syntax_check;
  syntax_check := true;
  -- if an error happens in the block, we were skipping it anyway...
  while token /= end_t and token /= eof_t and token /= termid1 and token /= termid2 loop
      ParseGeneralStatement;         -- step through context
  end loop;
  --error_found := old_error;          -- ignore any error while skipping
  syntax_check := old_error;
end SkipBlock;

procedure ParseBlock( termid1, termid2 : identifier := keyword_t ) is
  -- Syntax: block = "general-stmt [general-stmt...] termid1 | termid2"
begin
  if token = end_t or token = eof_t or token = termid1 or token = termid2 then
     err( "missing statement" );
  end if;
  while token /= end_t and token /= eof_t and token /= termid1 and token /= termid2 loop
     ParseGeneralStatement;
  end loop;
end ParseBlock;

procedure ParseDeclareBlock is
begin
  expect( declare_t );
  pushBlock( newScope => true, newName => "declare block" );
  ParseDeclarations;
  expect( begin_t );
  ParseBlock;
  expect( end_t );
  pullBlock;
end ParseDeclareBlock;

procedure ParseBeginBlock is
begin
  pushBlock( newScope => true, newName => "begin block" );
  expect( begin_t );
  ParseBlock;
  expect( end_t );
  pullBlock;
end ParseBeginBlock;

procedure ParseEndOfStatement is
  -- NOT YET IMPLEMENTED
  fd               : aFileDescriptor := stderr;
  destroyFileFirst : boolean := false;
  file             : identifier;
begin
  if token = symbol_t and then identifiers( Token ).value = ";" then
     return;
  end if;
  if identifiers( token ).value = "<" then
     fd := stdout;
     destroyFileFirst := true;
  elsif identifiers( token ).value = ">" then
     fd := stdout;
     destroyFileFirst := true;
  elsif identifiers( token ).value = ">>" then
     fd := stdout;
  end if;
  if fd = stderr then -- kludge
     err( "';', '<', '>' or '>>' expected" );
  end if;
  ParseOpenFile( file );
  -- more checking here
  if isExecutingCommand then
     if destroyFileFirst then
        err( "not implemented" );
     end if;
  end if;
end ParseEndOfStatement;

procedure ParseCompletedStatement is
  -- NOT YET IMPLEMENTED
begin
  null;
end ParseCompletedStatement;

procedure ParseShellCommand is
  -- Syntax: command-word ( expr [,expr...] )
  -- Syntax: command-word param-word [param-word...]
  cmdNameToken : identifier;
  cmdName    : unbounded_string;
  paramCnt   : natural;
  expr_val   : unbounded_string;
  expr_type  : identifier;
  ap         : Argument_List_Access;
  firstParam : aScannerState;
  Success    : boolean;

  function countGlobbed( s : unbounded_string ) return natural is
    globCriteria : regexp;
    results      : natural := 0;
    currentDir   : Dir_Type;
    fileName     : string(1..256);
    fileNameLen  : natural;
    noPWD        : boolean := false;
  begin
    globCriteria := Compile( to_string( s ), Glob => true,
       Case_Sensitive => true );
    begin
      open( currentDir, "." );
    exception when others =>
      noPWD := true;
    end;
    if noPWD then
       err( "current directory not accessible / doesn't exist" );
       return 0;
    end if;
    read( currentDir, fileName, fileNameLen ); -- skip "."
    read( currentDir, fileName, fileNameLen ); -- skip ".."
    loop
      read( currentDir, fileName, fileNameLen );
      exit when fileNameLen = 0;
      if Match( fileName(1..fileNameLen ) , globCriteria ) then
         results := results + 1;
      end if;
    end loop;
    if results = 0 then
       results := results + 1;
    end if;
    close( currentDir );
    return results;
  end countGlobbed;

  procedure expandGlobbed( ap : in out argument_list_access;
    paramNum : in out positive;
    word : string; s : string ) is
  -- add glob-expanded arguments to ap.  paramNum is next free
  -- position in ap
    globCriteria : regexp;
    currentDir   : Dir_Type;
    fileName     : string(1..256);
    fileNameLen  : natural;
    found        : boolean := false;
    noPWD        : boolean := false;
  begin
    globCriteria := Compile( s, Glob => true, Case_Sensitive => true );
    begin
      open( currentDir, "." );
    exception when others =>
      noPWD := true;
    end;
    if noPWD then
       err( "current directory not accessible / doesn't exist" );
       return;
    end if;
    read( currentDir, fileName, fileNameLen ); -- skip "."
    read( currentDir, fileName, fileNameLen ); -- skip ".."
    loop
      read( currentDir, fileName, fileNameLen );
      exit when fileNameLen = 0;
      if Match( fileName(1..fileNameLen ) , globCriteria ) then
         ap( paramNum ) := new string( 1..fileNameLen );
         ap( paramNum ).all := fileName( 1..fileNameLen );
         found := true;
         paramNum := paramNum + 1;
      end if;
    end loop;
    -- no matches?  add it as a parameter "as is"
    if not found then
       ap( paramNum ) := new string( 1..word'length );
       ap( paramNum ).all := word;
       paramNum := paramNum + 1;
    end if;
    close( currentDir );
  end expandGlobbed;

  function isParanthesis return boolean is
  begin
     skipWhiteSpace;
     return script( cmdpos ) = '(';
  end isParanthesis;

  word         : unbounded_string;
  pattern      : unbounded_string;
  inBackground : boolean;
  wordCnt      : natural;
  wordType     : aShellWordType;

  pipe2Next    : boolean := false;
  pipeFromLast : boolean := false;

begin

  -- the main loop just did a resumeScanning.  The token should
  -- still be set to the value of the first word.

<<next_in_pipeline>>
  inBackground := false;
  wordCnt  := 0;
  paramCnt := 0;

  if identifiers( token ).kind = command_t then
     -- perform alias substitution here
     cmdName := identifiers( token ).value;
  else
     cmdpos := firstpos;    -- backup to the start of the first token
     cmdNameToken := token;
     getShellWord( wordType, cmdName, pattern, first => true );
     -- the the first "word" of the command
  end if;

  -- AdaScript Syntax: count the number of parameters, generate an argument
  -- list of the correct length, interpret the parameters "for real".

  if isParanthesis then                                  -- parenthesis?
     get_next_token;                                     -- use AdaScript syntax
     expect( symbol_t, "(" );                            -- skip paraenthesis
     markScanner( firstParam );                          -- save position
     while not error_found and not exit_block loop       -- count parameters
        ParseExpression( expr_val, expr_type );
        paramCnt := paramCnt + 1;
        if Token = symbol_t and then identifiers( Token ).value = "," then
           get_next_token;
        else
           exit;
        end if;
     end loop;
     expect( symbol_t, ")" );
     pipe2Next := token = symbol_t and identifiers( token ).value = "|";
     if pipe2Next and onlyAda95 then
        err( "pipelines not allowed with pragma ada_95" );
     end if;

     -- don't consume the close ")" or another line will be read
     -- by the scanner.  OK to consume it if there's an error
     --if token /= symbol_t or identifiers( token ).value /= ")" then
     -- expect( symbol_t, ")" );
     --end if;

     if not error_found and not exit_block then          -- no syntax errors?
          resumeScanning( firstParam );                  -- restore position
          ap := new Argument_List( 1..positive( paramCnt ) );
          paramCnt := 0;                                 -- start with nothing
          loop
             ParseExpression( expr_val, expr_type );     -- do the expression
             paramCnt := paramCnt + 1;                   -- count parameter
             ap( positive( paramCnt ) ) := new string( 1..length( expr_val ) );
             ap( positive( paramCnt ) ).all := To_String( expr_val );
             if Token = symbol_t and then identifiers( Token ).value = "," then
                get_next_token;
             else
                exit;
             end if;
          end loop;
          expect( symbol_t, ")" );
          if token = symbol_t and identifiers(token).value = "&" then
             inbackground := true;
             expect( symbol_t, "&" );
             if pipe2Next then
                err( "no & - piped commands are automatically run in the background" );
             elsif pipeFromLast then
                err( "no & - final piped command always runs in the foreground" );
             end if;
          end if;
          if ap = null then
             err( "Unable to allocate memory to call command" );
          end if;
     end if;

    -- Shell-style parameters.  Read a series of "words", counting the parameters.
    -- Generate an argument list of the correct length and do it again "for real".

  else

     -- count loop
     markScanner( firstParam );
     word := null_unbounded_string;
     loop
        getShellWord( wordType, word, pattern );
        exit when wordType = semicolonWord or wordType = pipeWord  or error_found;
        if word /= "&" then
           paramCnt := paramCnt + countGlobbed( pattern );
        end if;
        wordCnt := wordCnt + 1;
     end loop;
     pipe2Next := word = "|";
     if pipe2Next and onlyAda95 then
        err( "pipelines not allowed with pragma ada_95" );
     end if;
     if paramCnt > 0 and onlyAda95 then
        err( "Bourne shell-style parameters not allowed with pragma ada_95" );
     end if;
     -- create loop
     resumeScanning( firstParam );
     if paramCnt > 0 then
        ap := new Argument_List( 1..positive( paramCnt ) );
     end if;
     paramCnt := 1;
     for i in 1..wordCnt loop
        getShellWord( wordType, word, pattern );
        if word = "&" then
           inbackground := true;
           if pipe2Next then
              err( "no & - piped commands are automatically run in the background" );
           elsif pipeFromLast then
              err( "no & - final piped command always runs in the foreground" );
           end if;
        else
           expandGlobbed( ap, paramCnt, to_string( word ), to_string( pattern ) );
        end if;
     end loop;
     if not error_found then
        -- the scanner position should be the semicolon at the end of the command
        -- read the semicolon into the token variable
        get_next_token;
     end if;
  end if;

  -- If we got here, we've got an argument list

  if isExecutingCommand then                                  -- no problems?
     if ap = null then                                        -- no arguments?
        ap := new Argument_List( 1..0 );                      -- use a null list
     end if;
     if rshOpt and then Element( cmdName, 1 ) = '/' then
        err( "absolute paths to commands not allowed in restricted shells" );
     elsif not pipeFromLast and pipe2next then
        run_inpipe( cmdName, ap, Success,
           background => true,
           cache => inputMode /= interactive );
     elsif pipeFromLast and not pipe2next then
        run_frompipe( cmdName, ap, Success,
           background => false,
           cache => inputMode /= interactive );
        closePipeline;
        -- certain cmds (like "less") need to be cleaned up
        -- with wait4children.  Others are OK.
        wait4children;
     elsif pipeFromLast and pipe2next then
        run_bothpipe( cmdName, ap, Success,
           background => true,
           cache => inputMode /= interactive );
     else
        run( cmdName, ap, Success,
           background => inBackground,
           cache => inputMode /= interactive ); -- run the command
     end if;
     discardUnusedIdentifier( cmdNameToken );
  end if;

-- restore the semi-colon we threw away at the beginning
-- by this point, Token is eof_t, so we'll have to force it to a ';'
-- since once Token is eof_t, it's always eof_t in the scanner

  if ap /= null then
     free( ap ); -- deallocate memory
  end if;

  -- shift down pipeline
  -- done is true for a command in the midst of the pipeline

  pipeFromLast := pipe2Next;
  if pipeFromLast and not error_found and not done then
     expect( symbol_t, "|" );
     if not error_found then
        goto next_in_pipeline;
     end if;
  end if;
end ParseShellCommand;

procedure ParseReturn is
  expr_val   : unbounded_string;
  expr_type  : identifier;
begin
  if inputMode = interactive then
     err( "Warning: This is an interactive shell.  Use logout." );
  else
     done := true;
  end if;
  get_next_token;
  if token /= symbol_t or identifiers( token ).value = ";" or
     identifiers( token ).value = "|" then
     if onlyAda95 then
        err( "return cannot return a status code with pragma ada_95.  use command_line package" );
     else
        ParseExpression( expr_val, expr_type );
        if uni_types_ok( expr_type, uni_numeric_t ) then
           last_status := integer( to_numeric( expr_val ) );
        end if;
     end if;
  end if;
  expect( symbol_t, ";" );
  syntax_check := true;
  if trace then
     put_trace( "Terminating script" );
  end if;
end ParseReturn;

procedure ParseGeneralStatement is
  -- Syntax: env-cmd | clear-cmd | ...
  var_id : identifier;
  expr_value : unbounded_string;
  expr_type  : identifier;
  right_type : identifier;
  cmdStart   : aScannerState;
  --secondToken : aScannerState;
  must_exit  : boolean;
begin

  markScanner( cmdStart );
--put_token;

  -- built-in

  if Token = command_t then
     get_next_token;
     ParseShellCommand;
  elsif Token = env_t then
     ParseEnvironment;
  elsif Token = clear_t then
     ParseClear;
  elsif Token = trace_t then
     ParseTrace;
  elsif Token = typeset_t then
     ParseTypeSet;
  elsif Token = unset_t then
     ParseUnset;
  elsif Token = wait_t then
     ParseWait;
  elsif Token = help_t then
     ParseHelp;
  elsif Token = pragma_t then
     ParsePragma;
  elsif Token = type_t then
     ParseType;
  elsif Token = null_t then
     get_next_token;
  elsif Token = subtype_t then
     ParseSubtype;
  elsif Token = if_t then
     ParseIfBlock;
  elsif Token = case_t then
     ParseCaseBlock;
  elsif Token = while_t then
     ParseWhileBlock;
  elsif Token = for_t then
     ParseForBlock;
  elsif Token = loop_t then
     ParseLoopBlock;
  elsif Token = return_t then
     ParseReturn;
     return;
  elsif token = logout_t then
     if inputMode /= interactive then
        err( "This is not an interactive shell.  Use return." );
     end if;
     get_next_token;
     expect( symbol_t, ";" );
     if not error_found then
        done := true;
        exit_block := true;
        if trace then
           put_trace( "Terminating session" );
        end if;
     end if;
     return;
  elsif Token = create_t then
     ParseOpen( create => true );
  elsif Token = open_t then
     ParseOpen;
  elsif Token = close_t then
     ParseClose;
  elsif Token = put_line_t then
     ParsePutLine;
  elsif token = symbol_t and identifiers( token ).value = "?" then
     ParseQuestion;
  elsif Token = put_t then
     ParsePut;
  elsif Token = new_line_t then
     ParseNewLine;
  elsif Token = skip_line_t then
     ParseSkipLine;
  elsif Token = set_input_t then
     ParseSetInput;
  elsif Token = set_output_t then
     ParseSetOutput;
  elsif Token = set_error_t then
     ParseSetError;
  elsif Token = cmd_setexit_t then
     ParseSetExitStatus;
  elsif Token = reset_t then
     ParseReset;
  elsif Token = delete_t then
     ParseDelete;
  elsif Token = get_t then
     ParseGet;
  elsif Token = system_t then
     ParseSystem;
  elsif Token = jobs_t then
     ParseJobs;
  elsif Token = delay_t then
     ParseDelay;
  elsif Token = else_t then
     err( "else without if" );
  elsif Token = elsif_t then
     err( "elsif without if" );
  elsif Token = with_t then
     err( "with not implemented" );
  elsif Token = use_t then
     err( "use not implemented" );
  elsif Token = case_t then
     err( "case not implemented" );
  elsif Token = task_t then
     err( "tasks not implemented" );
  elsif Token = protected_t then
     err( "protected types not implemented" );
  elsif Token = package_t then
     err( "packages not implemented" );
  elsif Token = function_t then
     err( "functions not implemented" );
  elsif Token = procedure_t then
     err( "procedures not implemented" );
  elsif Token = exit_t then
     if blocks_top = block'first then           -- not complete. should check
         err( "no enclosing loop to exit" );    -- not just for no blocks
     end if;                                    -- but the block type isn't easily checked
     expect( exit_t );
     if token = when_t then
        expect( when_t );
        ParseExpression( expr_value, expr_type );
        if base_types_ok( boolean_t, expr_type ) then
           must_exit :=  expr_value = "1";
        end if;
     else
        must_exit := true;
     end if;
     if isExecutingCommand and must_exit then
        exit_block := true;
        if trace then
           put_trace( "exiting" );
        end if;
     end if;
  elsif Token = declare_t then
     ParseDeclareBlock;
  elsif Token = begin_t then
     ParseBeginBlock;
  elsif Token = eof_t then
     return; -- don't bother with ; -- could be a comment
  else

     -- we need to check the next token then back up
     -- should really change scanner to double symbol look ahead?

     get_next_token;

     -- declarations

     if Token = symbol_t and to_string( identifiers( token ).value ) = ":" then
        resumeScanning( cmdStart );
        ParseNewIdentifier( var_id );
        ParseDeclarationPart( var_id );
     else

        -- assignments
        --
        -- for =, will be treated as a command if we don't force an error
        -- here for missing :=, since it was probably intended as an assignment

        if Token = symbol_t and to_string( identifiers( token ).value ) = "=" then
           expect( symbol_t, ":=" );
        elsif Token = symbol_t and to_string( identifiers( token ).value ) = ":=" then
           resumeScanning( cmdStart );
           if inputMode = interactive then
              if identifiers( token ).kind = new_t and not onlyAda95 and not restriction_no_auto_declarations then
                 ParseNewIdentifier( var_id );
              else
                 ParseIdentifier( var_id );
              end if;
           else
                 ParseIdentifier( var_id );
           end if;
           itself_defined := true;
           itself := identifiers( var_id ).value;
           itself_type := identifiers( var_id ).kind;
           if identifiers( itself_type ).limit then
              err( "limited variables cannot be assigned a value" );
           end if;
           ParseAssignPart( expr_value, right_type );
           if inputMode = interactive then
              if identifiers( var_id ).kind = new_t and not onlyAda95 and not restriction_no_auto_declarations and not error_found then
                 identifiers( var_id ).kind := right_type;
                 put_trace( "Assuming " & to_string( identifiers( var_id ).name ) &
                    " is a new " & to_string( identifiers( right_type ).name ) &
                    " variable" );
              end if;
           end if;
           if base_types_ok( identifiers( var_id ).kind, right_type ) then
               if class_ok( var_id, otherClass ) then
                  if get_uni_type( var_id ) = uni_numeric_t then
                     -- numeric test.  universal typelesses could result
                     -- in a non-numeric expression that base_types_ok
                     -- doesn't catch.
                     declare
                       lf : long_float;
                     begin
                       if isExecutingCommand then
                          lf := to_numeric( expr_value );
                       end if;
                     exception when others =>
                       err( "'" & to_string( expr_value) & "' is not numeric" );
                     end;
                  end if;
                  if isExecutingCommand then
                     identifiers( var_id ).value := expr_value;
                     if trace then
                        -- builtins.env( ident ) would be better if a value is
                        -- returned
                        put_trace(
                           to_string( identifiers( var_id ).name ) & " := '" &
                           to_string( Escape_String( expr_value ) ) & "'" );
                     end if;
                  end if;
               end if;
           end if;
           itself_defined := false;
        elsif cmdStart.token = pwd_t then
           put_line( builtins.pwd );
        elsif cmdStart.token = cd_t then
           if rshOpt then
              err( "cd not allowed in restricted shell" );
           elsif token = symbol_t and identifiers( token ).value = ";" then
              builtins.cd( null_unbounded_string );
           else
              declare
                dir : unbounded_string;
              begin
                for i in cmdpos..script'length loop
                    if script( i ) = ASCII.NUL then
                       builtins.cd(
                          to_unbounded_string(
                             script( firstpos..i-2 ) -- -2 for NUL & ';'
                          )
                       );
                       cmdpos := i-1;
                       exit;
                    end if;
                end loop;
              end;
              get_next_token;
           end if;
        else

           -- assume it's a shell command and run it
           -- current token is first "token" of parameter.  Blow it away
           -- if able (ie. "ls file", current token is file but we don't
           -- need that in our identifier list.)
           resumeScanning( cmdStart );
           ParseShellCommand;
        end if;
     end if;
  end if;
  expect( symbol_t, ";" );
exception when symbol_table_overflow =>
  err( "too many identifiers (symbol table overflow)" );
when block_table_overflow =>
  err( "too many nested statements/blocks (block table overflow)" );
end ParseGeneralStatement;

procedure ParseMainProgram is
  program_id : identifier;
begin
  expect( procedure_t );
  ParseNewIdentifier( program_id );
  expect( is_t );
  pushBlock( newScope => true,
    newName => to_string (identifiers( program_id ).name ) );
  ParseDeclarations;
  expect( begin_t );
  ParseBlock;
  expect( end_t );
  expect( program_id );
  pullBlock;
end ParseMainProgram;

---------------------------------------------------------
-- END OF ADASCRIPT PARSER
---------------------------------------------------------

---------------------------------------------------------
-- START OF HOUSEKEEPING and SETUP
---------------------------------------------------------

procedure interpret is
  -- interpret a command line typed in by the user
  command : unbounded_string;
  result : aFileDescriptor;
begin
  loop                                             -- repeatedly
    updateJobStatus;                               -- cleanup any background jobs
    put( term_bold );                              -- show prompt in bold
    put( prompt );
    put( term_normal );
    getLine( command, keepHistory => true );       -- read command from keyboard
--put_line( "stdin is fd" & stdin'img );
--put_line( "currentStandardInput is fd" & currentStandardInput'img );
--put_line( "originalStandardInput is fd" & originalStandardInput'img );
    if currentStandardInput /= originalStandardInput then
--put_line( "...Redirecting standard input / overwriting stdin" );
       result := dup2( currentStandardInput, stdin );   -- restore stdin
    end if;
    if currentStandardOutput /= originalStandardOutput then
       result := dup2( currentStandardOutput, stdout ); -- restore stdout
    end if;
    if currentStandardError /= originalStandardError then
       result := dup2( currentStandardError, stderr );  -- restore stderr
    end if;
    error_found := false;                               -- no error found
    exit_block := false;                                -- not exit-ing a block
    fixSpacing( command );                              -- strip leading/trailing spaces
    if length( command ) > 0 then                       -- something there?
       if Element( command, length( command ) ) /= ';' then -- missing ending ;?
          command := command & ";";                     -- it's implicit so add it
       elsif length( command ) > 1 then
          if Element( command, length( command )-1) = '\' then
             command := command & ";";
          end if;
       end if;                                          -- w/space in case shell cmd
    end if;
    compileCommand( command );
    if not error_found then
       cmdpos := firstScriptCommandOffset;              -- start at first char
       token := identifiers'first;                      -- dummy, replaced by g_n_t
       lineno := 1;
       get_next_token;                                  -- load first token
       while token /= eof_t and not error_found loop
          ParseGeneralStatement;                        -- do the command
       end loop;
    end if;
    result := dup2( originalStandardInput, stdin );   -- restore standard input
    result := dup2( originalStandardOutput, stdout ); -- restore standard output
    result := dup2( originalStandardError, stderr );  -- restore standard error
    exit when done;                                   -- and stop when done
  end loop;
end interpret;

procedure interpretScript( scriptPath : string ) is
  firstLine : aliased unbounded_string;
begin
  if syntax_check then
     if verboseOpt then
        Put_Trace( "Checking Syntax" );
     end if;
  else
     if verboseOpt then
        Put_Trace( "Executing Script" );
     end if;
  end if;
  scriptFilePath := to_unbounded_string( scriptPath ); -- script name
  inputMode := fromScriptFile;                     -- running a script
  scriptFile := open( scriptPath & ASCII.NUL, 0, 660 ); -- open script
  if scriptFile < 1 then                           -- error?
     scriptFilePath := scriptFilePath & ".bush";   -- try name with ".bush"
     scriptFile := open( scriptPath & ".bush" & ASCII.NUL, 0, 660 );
  end if;
  if scriptFile > 0 then                           -- good?
     error_found := false;                         -- no error found
     exit_block := false;                          -- not exit-ing a block
     if not LineRead( firstLine'access ) then        -- read first line
        put_line( standard_error, "unable to read first line of script" );
        error_found := true;
        goto error;                                -- and interpreting
     end if;
     if script = null then
        if verboseOpt then
           Put_Trace( "Compiling Into Byte Code" );
        end if;
        compileScript( firstline );
     end if;
     lineno := 1;
     if not error_found then
        cmdpos := firstScriptCommandOffset;
        token := identifiers'first;                -- dummy, replaced by g_n_t
        get_next_token;                            -- load first token
        while (not error_found and not done) and (token = procedure_t or
          token = pragma_t or token = trace_t) loop
           if token = pragma_t then
              ParsePragma;
              expect( symbol_t, ";" );
           elsif token = trace_t then
              ParseTrace;
              expect( symbol_t, ";" );
           elsif token = procedure_t then
              ParseMainProgram;
              expect( symbol_t, ";" );
              expect( eof_t );                        -- should be nothing else
              exit;
           end if;
        end loop;
        if not done or token = eof_t then
           loop
             ParseGeneralStatement;                   -- process the first statement
             exit when done or token = eof_t;         -- continue until done
           end loop;                                  --  or eof hit
           if not done then                           -- not exiting?
              expect( eof_t );                        -- should be nothing else
           end if;
        end if;
     end if;
  elsif errno = 2 then                             -- file not found?
     put_line( standard_error, "unable to open script file '" &
       scriptPath & "' : file not found" );
       error_found := true;
  else                                             -- some other error?
     put_line( standard_error, "unable to open script file '" &
       scriptPath & "' : " & OSerror( errno ) );
       error_found := true;
  end if;
<<error>>
  if error_found then                              -- was there an error?
     if last_status = 0 then                       -- no last command status?
        set_exit_status( Failure );                -- just set to 1
        if trace then
           put_trace("Script exit status is" & failure'img );
        end if;
     else                                          -- otherwise
        set_exit_status( exit_status( last_status ) ); -- return last status
        if trace then
           put_trace("Script exit status is" & last_status'img );
        end if;
     end if;
  else                                             -- no error?
     set_exit_status( 0 );                         -- return no error
     if trace then
        put_trace("Script exit status is 0" );
     end if;
  end if;
  close( scriptFile );                             -- close the script file
end interpretScript;

-- Housekeeping

procedure startParser is
begin
  -- Setup standard input/output/error redirection
  --
  -- The original stdin/out/err must be copied to prevent them from being
  -- closed and lost.  These are probably fd 3,4,5, but will record what dup
  -- returns to be safe.  For example, if fd 0 is closed (by dup2) the file
  -- actually will remain open until fd 3 is also closed.
  --
  -- currentStandardInput/Output/Error represent the file stdin/out/err has
  -- been redirected to by Set_Input/Output/Error.  If none, the value is
  -- 0,1,2 respectively.
  --
  -- The standard files standard_input, standard_output and standard_error
  -- represent the original stdin/out/err...ie the copied fd 4,5,6.

  originalStandardInput := dup( stdin );              -- record stdin
  originalStandardOutput := dup( stdout );            -- record stdout
  originalStandardError := dup( stderr );             -- record stderr
  currentStandardInput := originalStandardInput;      -- no redirection, stdin
  currentStandardOutput := originalStandardOutput;    -- no redirection, stdout
  currentStandardError := originalStandardError;      -- no redirection, stderr
  identifiers( standard_input_t ).value :=            -- copy of stdin
     to_unbounded_string( originalStandardInput'img );
  identifiers( standard_output_t ).value :=           -- copy of stdout
     to_unbounded_string( originalStandardOutput'img );
  identifiers( standard_error_t ).value :=            -- copy of stderr
     to_unbounded_string( originalStandardError'img );

  Ada.Numerics.Float_Random.Reset( random_generator );
end startParser;

procedure shutdownParser is
begin
  -- clean up any completed child processes
  wait4children;
end shutdownParser;

procedure parse is
  save_rshOpt : boolean;
  home_id : identifier;
begin
  if optionOffset = 0 then                         -- no arguments or '-'?
     if isLoginShell then                        -- check for startup script
        save_rshOpt := rshOpt;
        rshOpt := true;
        -- we're using scriptFile and scriptFilePath as temp variables
        -- here.  They'll be overwritten when the script is interpreted.
        scriptFile := open( "/etc/bush_profile" & ASCII.NUL, 0, 660 ); -- open script
        if scriptFile > 0 then                       -- good?
           close( scriptFile );                      -- close test fd
           interpretScript( "/etc/bush_profile" );   -- run login script
        end if;
        findIdent( to_unbounded_string( "HOME" ), home_id );
        if home_id /= eof_t then                     -- HOME defined?
           scriptFilePath := identifiers( home_id ).value;
           if Element( scriptFilePath, length( scriptFilePath ) ) /= '/' then
              scriptFilePath := scriptFilePath & "/";
           end if;
           scriptFilePath := scriptFilePath & ".bush_profile";
           scriptFile := open( to_string( scriptFilePath ) & ASCII.NUL, 0, 660 ); -- open script
           if scriptFile > 0 then                       -- good?
              close( scriptFile );                      -- close test fd
              interpretScript( to_string( scriptFilePath ) ); -- run login script
           end if;
        end if;
     end if;
     rshOpt := save_rshOpt;
     inputMode := interactive;                       -- we're interactive
     interpret;
  else
     if syntaxOpt then
        syntax_check := true;
        interpretScript( Argument( OptionOffset ) );
     elsif nosyntaxOpt then
        interpretScript( Argument( OptionOffset ) );
     else
        syntax_check := true;
        interpretScript( Argument( optionOffset ) );
        if not error_found then
           resetScanner;
           interpretScript( Argument( optionOffset ) );
        end if;
     end if;
  end if;
end Parse;

end parser;
