------------------------------------------------------------------------------
-- BUSH Run UNIX/Linux Commands                                             --
--                                                                          --
-- Part of BUSH                                                             --
------------------------------------------------------------------------------
--                                                                          --
--                 Copyright (C) 2001 Ken O. Burtch & FSF                   --
--                                                                          --
-- This is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  This is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with this;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- This is maintained at http://www.vaxxine.com/pegasoft                    --
--                                                                          --
------------------------------------------------------------------------------

with system,
    ada.unchecked_deallocation,
    ada.text_io,
    ada.strings.unbounded.text_io,
    ada.characters.handling,
    bush_os,
    user_io,
    script_io,
    scanner,
    builtins,
    string_util,
    gen_list;
use ada.text_io,
    ada.strings.unbounded,
    ada.strings.unbounded.text_io,
    ada.characters.handling,
    bush_os,
    user_io,
    script_io,
    scanner,
    builtins,
    string_util;

package body run_cmd is

type aPipeStack is array(1..10) of aPipe;
pipeStack    : aPipeStack;
pipeStackTop : natural := 0;

errno :integer;
pragma import( C, errno );
-- multiple declarations caused by gnat.os_lib;

-- Jobs
--
-- Background tasks (jobs started with '&' or in a pipeline)
-- are stored in a linked list called "jobs".  When a job
-- finishes, it is removed from the list.

type aJobStatus is ( running, stopped, unknown );
type aJob is record
   pid    : aPID;               -- process id of the background task
   cmd    : unbounded_string;   -- the command name
   status : aJobStatus;         -- status as of last check
   quiet  : boolean := false;   -- true => no job control messages
end record;

function ">=" ( left, right : aJob ) return boolean is
  -- sorting criteria (not actually used by BUSH but needed
  -- for instantiation of the generic gen_list package).
begin
  return left.pid >= right.pid;
end ">=";

package jobList is new gen_list( aJob, "=", ">=" );
jobs : jobList.List;            -- our list of jobs

-----------------------------------

procedure findCmdPath( cmd : unbounded_string; fullpath : out unbounded_string ) is
   -- Determine the location of the command "cmd".  If a path
   -- was specified in cmd, check for and expand a leading tilda.
   -- If no path, then we need to do a search of the PATH variable.

   slashPos : integer;          -- slashes in the command path
   colonPos : integer;          -- colons in the PATH variable
   pathId   : identifier;       -- PATH in identifiers

   cmdpath  : unbounded_string; -- working copy of the cmd path
   prefixes : unbounded_string; -- PATH prefix list
   prefix   : unbounded_string; -- prefix being tested
begin
   cmdPath := cmd;
   slashPos := index( cmd, "/" );
   if slashPos > 0 then
      if Element( cmdPath, 1 ) = '~' then
         Delete( cmdPath, 1, 1 );
         Insert( cmdPath, 1, to_string( pwd ) );
      end if;
      fullPath := cmdPath;
      if not C_is_executable_file( to_string( fullPath ) & ASCII.NUL ) then
	 fullPath := null_unbounded_string;
      end if;
   else
      findIdent( to_unbounded_string( "PATH" ), pathId );
      if pathId /= eof_t then
         prefixes := identifiers( pathId ).value;
         while length( prefixes ) > 0 loop
            colonPos := Index( prefixes, ":" );
            if colonPos = 0 then
               prefix := prefixes;
               prefixes := null_unbounded_string;
            else
               prefix := to_unbounded_string(
                  slice( prefixes, 1, colonPos-1 ) );
               Delete( prefixes, 1, colonPos );
            end if;
            if Element( prefix, length( prefix ) ) = '/' then
               fullPath := prefix & cmdPath;
            else
               fullPath := prefix & '/' & cmdPath;
            end if;
            if C_is_executable_file( to_string( fullPath ) & ASCII.NUL ) then
               exit;
            else
               fullPath := null_unbounded_string;
            end if;
         end loop;
      end if;
   end if;
end findCmdPath;

procedure run( cmd : unbounded_string;
   ap : argument_list_access;
   success : out boolean;
   background : boolean := false;
   cache : boolean := true ) is
   fullpath : unbounded_string;

   myPID    : aPID;
   newJob   : aJob;
begin

   findCmdPath( cmd, fullPath );

   if length( fullPath ) = 0 then
      cmdpos := cmdpos - 1;
      err( "'" & to_string( cmd ) & "' command not found" );
   else
      if Trace then
         put_trace( to_string( fullPath ) );
      end if;
      -- Spawn already does a fork, so this is wasteful.  But it will be OK
      -- once we get rid of GNAT.OS_Lib.
      if background then
         myPID := fork;
         if myPID < 0 then
            err( "unable to fork and run the command" );
            Success := false;
         elsif myPID = 0 then
            spawn( To_String( fullPath ), ap.all, Success );
            -- drop out of shell child process
            done := true;
            exit_block := true;
         else
	    newJob.pid := myPID;
	    newJob.cmd := cmd;
	    newJob.status := running;
	    newJob.quiet := false;
	    jobList.Insert( jobs, newJob );
            put_line( cmd & " (" & myPID'img & ") has started" );
            Success := true;
         end if;
      else
         spawn( To_String( fullPath ), ap.all, Success );
      end if;
   end if;
end run;

procedure run_inpipe( cmd : unbounded_string;
   ap : argument_list_access;
   success : out boolean;
   background : boolean := false;
   cache : boolean := true ) is
   fullpath : unbounded_string;

   myPID    : aPID;
   newJob   : aJob;
   result   : integer;
begin
   pipeStackTop := pipeStack'first;
   pipe( result, pipeStack( pipeStackTop ) );
   if result /= 0 then
      err( "unable to create pipe: errno " & errno'img );
      pipeStackTop := pipeStackTop-1;
      Success := false;
      return;
   end if;

   findCmdPath( cmd, fullPath );
   if length( fullPath ) = 0 then
      cmdpos := cmdpos - 1;
      err( "'" & to_string( cmd ) & "' command not found" );
   else
      -- Spawn already does a fork, so this is wasteful.  But it will be OK
      -- once we get rid of GNAT.OS_Lib.
      if background then
         myPID := fork;                      -- start a new process
         if myPID < 0 then                   -- failed?
            err( "unable to fork and run the command" );
            Success := false;
         elsif myPID = 0 then                -- in child process?
            close( stdout );                 -- redirect stdout to pipe
            result := integer( dup2( pipeStack( pipeStackTop )( intoPipe ), stdout ) );
            if result < 0 then               -- failed?
               err( "unable to dup2 the pipe: errno " & errno'img );
               Success := false;
               return;
            end if;
            close( pipeStack( pipeStackTop )( intoPipe ) ); -- copied this
            close( pipeStack( pipeStackTop )( outOfPipe ) ); -- not using pipe output

            if trace then
	       put_trace( to_string( fullPath ) &
	         " output attached to input end of pipe fd" & 
	         pipeStack( pipeStackTop )( intoPipe )'img );
            end if;

            spawn( To_String( fullPath ), ap.all, Success );
	    -- GNAT.OS_Lib.Spawn

            -- drop out of shell child process
            done := true;                    -- bail out of shell
	    error_found := true;             -- suppress errors
            exit_block := true;              -- exit any enclosing block
         else                                -- parent process?
	    newJob.pid := myPID;             -- record the job
	    newJob.cmd := cmd;               -- in the job table
	    newJob.status := running;
	    newJob.quiet := true;
	    jobList.Insert( jobs, newJob );
            --put_line( cmd & " (" & myPID'img & ") has started" );
            Success := true;
         end if;
      else
	 err( "internal error: pipelines must be run in background" );
      end if;
   end if;
end run_inpipe;

procedure run_frompipe( cmd : unbounded_string;
   ap : argument_list_access;
   success : out boolean;
   background : boolean := false;
   cache : boolean := true ) is

   fullpath : unbounded_string;
   oldStdin : aFileDescriptor;
   result   : integer;
begin
   findCmdPath( cmd, fullPath );

   if length( fullPath ) = 0 then
      cmdpos := cmdpos - 1;
      err( "'" & to_string( cmd ) & "' command not found" );
   else
      if background then                    -- should never be
	 err( "internal error: final pipeline command must be run in foreground" );
	 Success := false;
	 return;
      end if;

      oldStdin := dup( stdin );             -- save original stdin
      if oldStdin < 0 then                  -- not OK? abort
         err( "unable to save stdin: errno" & errno'img );
         return;
      end if;
      close( stdin );                       -- redirect standard input
      result := integer( dup2( pipeStack( pipeStackTop )( outOfPipe), stdin ) );
      if result < 0 then                    -- failed?
         err( "unable to dup2 the pipe for " & to_string( cmd ) &
	    "input: errno " & errno'img );
         Success := false;
      else
         -- close all the pipe file descriptors.  We've copied the ones
         -- we need with dup2.
         closePipeline; -- or close-on-exec
         if trace then
	    put_trace( to_string( fullPath ) &
	       " input attached to output end of pipe fd" & 
	       pipeStack( pipeStackTop )( outOfPipe )'img );
         end if;
         spawn( To_String( fullPath ), ap.all, Success );
         close( stdin );                    -- closing pipe
      end if;
      result := integer( dup2( oldStdin, stdin ) );
      if result < 0 then                    -- failed? show error
         err( "internal error: unable to restore old stdin: " &
	    " unable to dup2(" & oldStdin'img & "," &
	    stdin'img & "): errno " & errno'img );
      end if;
      close( oldStdin );                    -- discard original stdin copy

   end if;
end run_frompipe;

procedure run_bothpipe( cmd : unbounded_string;
   ap : argument_list_access;
   success : out boolean;
   background : boolean := false;
   cache : boolean := true ) is
   fullpath : unbounded_string;

   myPID    : aPID;
   newJob   : aJob;
   result   : integer;
begin
   pipeStackTop := pipeStackTop+1;             -- prepare for another pipe
   if pipeStackTop > aPipeStack'last then      -- no available pipes?
      err( "pipeline too long" );              -- print error
      pipeStackTop := pipeStackTop-1;          -- avoid future constraint
      Success := false;                        -- errors and flag failure
      return;                                  -- and quit
   end if;
   pipe( result, pipeStack( pipeStackTop ) );  -- create pipe
   if result /= 0 then                         -- failed?
      err( "unable to create pipe: errno " & errno'img );
      pipeStackTop := pipeStackTop-1;          -- top not valid pipe
      Success := false;                        -- flag failure
      return;                                  -- and quit
   end if;

   findCmdPath( cmd, fullPath );
   if length( fullPath ) = 0 then
      cmdpos := cmdpos - 1;
      err( "'" & to_string( cmd ) & "' command not found" );
   else
      -- Spawn already does a fork, so this is wasteful.  But it will be OK
      -- once we get rid of GNAT.OS_Lib.
      if background then
         myPID := fork;
         if myPID < 0 then
            err( "unable to fork to run " & to_string( cmd ) );
            Success := false;
         elsif myPID = 0 then
            close( stdin );
            result := integer( dup2( pipeStack( pipeStackTop-1 )( outOfPipe ), stdin ) );
            if result < 0 then
               err( "unable to dup2 the pipe for " & to_string( cmd ) &
		  " input: errno " & errno'img );
               Success := false;
               return;
            end if;
            close( stdout );
            result := integer( dup2( pipeStack( pipeStackTop )( intoPipe ), stdout ) );
            if result < 0 then
               err( "unable to dup2 the pipe for " & to_string( cmd ) &
		  " output: errno " & errno'img );
               Success := false;
               return;
            end if;
	    -- close all the pipe file descriptors.  We've copied the ones
	    -- we need with dup2.
	    closePipeline; -- or close-on-exec

            if trace then
	       put_trace( to_string( fullPath ) &
	         " input attached to output end of pipe fd" &
	         pipeStack( pipeStackTop-1 )( outOfPipe )'img );
	       put_trace( to_string( fullPath ) &
	         " output attached to input end of pipe fd" &
	         pipeStack( pipeStackTop )( intoPipe )'img );
            end if;

            spawn( To_String( fullPath ), ap.all, Success );

            -- drop out of shell child process
            done := true;
	    error_found := true;
            exit_block := true;
         else
	    newJob.pid := myPID;
	    newJob.cmd := cmd;
	    newJob.status := running;
	    newJob.quiet := true;
	    jobList.Insert( jobs, newJob );
            --put_line( cmd & " (" & myPID'img & ") has started" );
            Success := true;
         end if;
      else
	 err( "internal error: pipelines must be run in background" );
         --spawn( To_String( fullPath ), ap.all, Success );
      end if;
   end if;
end run_bothpipe;

procedure closePipeline is
  -- doesn't reset the top of stack
begin
  for i in 1..pipeStackTop loop
      close( pipeStack( i )( intoPipe ) );
      close( pipeStack( i )( outOfPipe ) );
  end loop;
end closePipeline;

procedure free_it is new ada.unchecked_deallocation( argument_list,
  argument_list_access );

procedure free( ap : in out argument_list_access ) is
  -- deallocate an argument list
begin
  for i in ap.all'range loop
      free( ap( i ) );
  end loop;
  free_it( ap );
end free;

procedure wait4children is
  -- wait for all jobs to finish
  oldJob : aJob;
  pid    : aPID;
  status : integer;
begin
  -- clean up any completed child processes
  pid := 0;
  loop
     oldJob.cmd := null_unbounded_string;
     wait( pid, status );
     exit when pid = -1 and errno /= EINTR;
     for i in 1..jobList.Length( jobs ) loop
	 jobList.Find( jobs, i, oldJob );
	 if oldJob.pid = pid then
	    jobList.Clear( jobs, i );
	    exit;
	 end if;
     end loop;
     if not oldJob.quiet then
        put_line( oldJob.cmd &" (" & pid'img & ") is finished" );
     end if;
  end loop;
end wait4children;

procedure updateJobStatus is
  -- update the status for all jobs, displaying changes
  status : integer;                               -- wait status
  result : aPID;                                  -- wait result code
  job    : aJob;                                  -- a BUSH job
begin
  for i in 1..jobList.Length( jobs ) loop         -- for all jobs
      jobList.Find( jobs, i, job );               -- next job
<<retry>> waitpid( result, job.pid, status, WNOHANG+WUNTRACED );
      if result = -1 and errno = EINTR then       -- interrupted syscall?
	 goto retry;                              -- retry it
      elsif result = -1 and errno /= ECHILD then  -- error if any error
	 err( "waitpid: error #" & errno'img );   -- execpt ECHILD
      elsif result = -1 then                      -- no such job? (ECHILD)
         jobList.Clear( jobs, i );                -- then it's done
	 if not job.quiet then
            put_line( job.cmd &" (" & job.pid'img & ") has finished" );
         end if;
      elsif result = 0 then                       -- unable to get status?
	if job.status /= running then             -- show status on a change
	   if not job.quiet then
              put_line( job.cmd &" (" & job.pid'img & ") is running" );
           end if;
        end if;
        job.status := running;                    -- still running
      elsif result /= 0 then                      -- able to get status?
	if job.status /= stopped then             -- show status on a change
	   if not job.quiet then
              put_line( job.cmd &" (" & job.pid'img & ") has stopped" );
           end if;
        end if;
        job.status := stopped;                    -- probably stopped
      end if;
  end loop;
end updateJobStatus;

procedure putJobList is
   -- display the job list
  oldJob : aJob;
begin
  updateJobStatus;
  for i in 1..jobList.Length( jobs ) loop
      jobList.Find( jobs, i, oldJob );
      put( oldJob.cmd &" (" & oldJob.pid'img & ") is " );
      case oldJob.status is
      when running =>
	put_line( "running" );
      when stopped =>
	put_line( "stopped" );
      when others =>
	put_line( "unknown status" );
      end case;
  end loop;
end putJobList;

end run_cmd;

