------------------------------------------------------------------------------
-- Lexical Scanner (the thing that reads your source code)                  --
-- Also, the semantic stuff.                                                --
--                                                                          --
-- Part of BUSH                                                             --
------------------------------------------------------------------------------
--                                                                          --
--                 Copyright (C) 2001 Ken O. Burtch & FSF                   --
--                                                                          --
-- This is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  This is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with this;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- This is maintained at http://www.vaxxine.com/pegasoft                    --
--                                                                          --
------------------------------------------------------------------------------

with system,
    ada.text_io,
    ada.command_line.environment,
    ada.strings.unbounded.text_io,
    ada.characters.handling,
    bush_os,
    script_io,
    user_io,
    string_util;
use ada.text_io,
    ada.command_line,
    ada.command_line.environment,
    ada.strings.unbounded,
    ada.strings.unbounded.text_io,
    ada.characters.handling,
    bush_os,
    script_io,
    user_io,
    string_util;

pragma Optimize( time );

package body scanner is

semicolon_string   : unbounded_string := to_unbounded_string( ";" );
verticalbar_string : unbounded_string := to_unbounded_string( "|" );

-- for converting numeric types with integer values to strings

type integerOutputType is delta 0.1 digits System.Max_Digits-2;
-- delta can't be 1, so we'll settle for 0.1.  This is the largest
-- long float number we can convert to an integer (and vice vera)
-- without rounding or resorting to scientific notation.  Why
-- -2? Anything higher resulted in rounding of Max_Int and Min_Int.

-- Identifer/Type Subprograms

procedure declareKeyword( id : out identifier; s : string ) is
-- initialize a new keyword / internal identifier
begin
  if identifiers_top = identifier'last then
     raise symbol_table_overflow;
  else
     identifiers( identifiers_top ).name := To_Unbounded_String( s );
     identifiers( identifiers_top ).kind := identifier'first;
     identifiers( identifiers_top ).value := Null_Unbounded_String;
     identifiers( identifiers_top ).class := otherClass;
     identifiers_top := identifiers_top + 1;
     id := identifiers_top-1;
  end if;
end declareKeyword;

function deleteIdent( id : identifier ) return boolean is
begin
  if id >= identifiers_top then
     return false;
  elsif id = identifiers_top-1 then
     identifiers_top := identifiers_top - 1;
     return true;
  elsif identifiers( id ).deleted then
     return false;
  end if;
  identifiers( id ).deleted := true;
  return true;
end deleteIdent;

procedure findIdent( name : unbounded_string; id : out identifier ) is
-- find an identifier, eof_t if failed
begin
  id := eof_t;
  for i in reverse 1..identifiers_top-1 loop
      if identifiers( i ).name = name and not identifiers( i ).deleted then
         id := i;
         exit;
      end if;
  end loop;
end findIdent;

procedure init_env_ident( s : string ) is
-- initialize an environment variable
  eqpos : natural := 0;
begin
  if identifiers_top = identifier'last then
     raise symbol_table_overflow;
  else
     for i in 1..s'length loop
         if s(i) = '=' then
            eqpos := i;
            exit;
         end if;
     end loop;
     identifiers( identifiers_top ).name := To_Unbounded_String( s(1..eqpos-1) );
     identifiers( identifiers_top ).kind := string_t;
     identifiers( identifiers_top ).value := To_Unbounded_String( s(eqpos+1..s'length ) );
     identifiers( identifiers_top ).import := true;
     identifiers_top := identifiers_top + 1;
  end if;
end init_env_ident;

procedure declareIdent( id : out identifier; name : unbounded_string;
  kind : identifier; class : anIdentifierClass := otherClass ) is
begin
  if identifiers_top = identifier'last then
     raise symbol_table_overflow;
  else
     id := identifiers_top;
     identifiers( id ).name := name;
     identifiers( id ).kind := kind;
     identifiers( id ).value := Null_Unbounded_String;
     identifiers( id ).class := class;
     identifiers_top := identifiers_top+1;
  end if;
end declareIdent;

procedure declareIdent( id : out identifier; name : string;
  kind : identifier; class : anIdentifierClass := otherClass ) is
begin
  declareIdent( id, to_unbounded_string( name ), kind, class );
end declareIdent;

procedure declareStandardConstant( name : string; kind : identifier;
   value : string ) is
-- declare a system constant and we don't care what token it is
     id : identifier;
begin
  if identifiers_top = identifier'last then
     raise symbol_table_overflow;
  else
     id := identifiers_top;
     identifiers( id ).name := to_unbounded_string( name );
     identifiers( id ).kind := kind;
     identifiers( id ).value := to_unbounded_string( value );
     identifiers( id ).class := constClass;
     identifiers_top := identifiers_top+1;
  end if;
end declareStandardConstant;

procedure declareStandardConstant( token : out identifier; name : string;
   kind : identifier; value : string ) is
-- ditto, but return the token
begin
   token := identifiers_top;
   declareStandardConstant( name, kind, value );
end declareStandardConstant;

procedure put_trace( msg : string ) is
begin
  put_line( standard_error, "=> (" & msg & ")" );
end put_trace;

procedure err_previous( msg : string ) is
-- same as err, but don't hilight the current token because
-- the error actually happened before it.  Just mark start
-- of current token.
  savepos : integer;
begin
  savepos := lastpos;
  lastpos := firstpos;
  err( msg );
  lastpos := savepos;
end err_previous;

procedure err( msg : string ) is
-- if this is the first error encountered, display the message
-- set the token to eof_t to abort the parsing and set the
-- error_found flag to indicate that an error was encountered
  cmdline  : unbounded_string;
  firstpos : natural;
  lastpos  : natural;
begin
  -- Only display if a new error
  if error_found then
     return;
  end if;

  if inputMode /= interactive then
     put( standard_error, "In line" );
     put_line( standard_error, lineno'img );
  end if;

  -- Block Traceback

  if blocks_top > blocks'first then
     put( standard_error, "While in " );
     for i in reverse blocks'first..blocks_top-1 loop
         if i /= blocks_top-1 then
            put( standard_error, " in " );
         end if;
         put( standard_error, To_String( blocks( i ).blockName ) );
     end loop;
     New_Line( standard_error );
  end if;

  -- Command line that errored (ie the current line)

  getCommandLine( cmdline, firstpos, lastpos );
  put_line( standard_error, cmdline );

  -- Error Pointer

  for i in 1..firstpos-1 loop
     put( standard_error, " " );
  end loop;
  put( standard_error, "^" );
  if lastpos > firstpos then
     for i in 1..lastpos-firstpos-1 loop
        put( standard_error, "-" );
     end loop;
     put( standard_error, "^" );
  end if;
  put( standard_error, " " );
  -- Error Message
  put_line( standard_error, msg );
  error_found := true;
  token := eof_t;
end err;

procedure err_tokenize( msg : string; cmdline : string ) is
-- if this is the first error encountered, display the message
-- set the token to eof_t to abort the parsing and set the
-- error_found flag to indicate that an error was encountered
begin
  -- Only display if a new error
  if error_found then
     return;
  end if;

  if inputMode /= interactive then
     put( standard_error, "In line" );
     put_line( standard_error, lineno'img );
  end if;

  -- Command line that errored (ie the current line)

  put_line( standard_error, cmdline );

  -- Error Pointer

  for i in 1..firstpos-1 loop
     put( standard_error, " " );
  end loop;
  put( standard_error, "^" );
  if lastpos-1 > firstpos then
     for i in 1..lastpos-firstpos-2 loop
        put( standard_error, "-" );
     end loop;
     put( standard_error, "^" );
  end if;
  put( standard_error, " " );
  -- Error Message
  put_line( standard_error, msg );
  error_found := true;
  token := eof_t;
end err_tokenize;

procedure pushBlock( newScope : boolean := false;
  newName : string := "" ) is
-- start a new identifier scope, remember where we "parked"
begin
  if blocks_top = block'last then
     raise block_table_overflow;
  else
     blocks( blocks_top ).startpos := scriptLineStart;
     blocks( blocks_top ).identifiers_top := identifiers_top;
     blocks( blocks_top ).newScope := newScope;
     blocks( blocks_top ).blockName := To_Unbounded_String( newName );
     markScanner( blocks( blocks_top ).state );
     blocks_top := blocks_top + 1;
  end if;
end pushBlock;

procedure pullBlock is
-- restore the previous identifier scope
begin
  if blocks_top = block'first then
     raise block_table_overflow; -- underflow, actually
  else
     blocks_top := blocks_top - 1;
     identifiers_top := blocks( blocks_top ).identifiers_top;
  end if;
end pullBlock;

procedure topOfBlock is
-- return to the top of the current block, read that line
  posn    : long_integer;
  discard : aliased unbounded_string;
  b  : boolean;
begin
   if blocks_top > blocks'first then
      if inputMode /= interactive then
         posn := lseek( scriptFile, blocks( blocks_top-1 ).startpos, 0 );
         b := lineRead( discard'access );
         if trace and not exit_block and not error_found then
            put( standard_error, "=> " & '"' );
            put( standard_error, getCommandLine );
            put_line( standard_error, "" & '"' );
         end if;
      end if;
      resumeScanning( blocks( blocks_top-1 ).state );
  end if;
end topOfBlock;

function isLocal( id : identifier ) return boolean is
  -- true if identifier is local to the current scope
  blockStart : identifier := 1; -- start from very first identifier
  nextBlock       : block;
  hasLocalBlock : boolean;
begin
  if blocks_top > blocks'first then                    -- in a nested block?
     nextBlock := blocks_top - 1;                      -- start with latest
     hasLocalBlock := true;                            -- assume we'll find one
     while not blocks( nextBlock ).newScope loop       -- if not new scope
        if nextBlock = 1 then                          -- hit bottom?
           hasLocalBlock := false;                     -- no local blocks
           exit;                                       -- and quit
        end if;
        nextBlock := nextBlock - 1;                    -- try previous block
     end loop;
     if hasLocalBlock then                              -- found local scope?
        blockStart := blocks( nextBlock ).identifiers_top; -- use it
     end if;
  end if;
--put_line( "isLocal: testing " & identifiers( id ).name );
--put_line( "isLocal: blocks top is " & Blocks_top'img );
--put_line( "isLocal: block is " & nextBlock'img );
--put_line( "isLocal: " & id'img & " >= " & blockStart'img );
   return id >= blockStart;
end isLocal;

function to_numeric( s : unbounded_string ) return long_float is
-- convert an unbounded string to a long float
begin
  if Element( s, 1 ) = '-' then
     return long_float'value( to_string( s ) );
  elsif Element( s, 1 ) = ' ' then
     return long_float'value( to_string( s ) );
  else
     return long_float'value( " " & to_string( s ) );
  end if;
end to_numeric;

function to_numeric( id : identifier ) return long_float is
-- convert an identifer's value to a long float
begin
   return to_numeric( identifiers( id ).value );
end to_numeric;

function to_unbounded_string( f : long_float ) return unbounded_string is
-- convert a long_float to an  unbounded string
  f_trunc : long_float := long_float'truncation( f );
begin

  -- integer value?

   if f - f_trunc = 0.0 then
      -- there's no guarantee that a long_long_integer will fit into
      -- a long_float's mantissa, so we'll use a decimal type.
      --if f <= long_float( long_long_integer'last ) and
      --   f >= long_float( long_long_integer'first ) then
      if f <= long_float( integerOutputType'last ) and
         f >= long_float( integerOutputType'first ) then
         return to_unbounded_string( long_long_integer( f )'img );
      end if;
   end if;

  -- else default to the image

   return to_unbounded_string( long_float'image( f ) );
end to_unbounded_string;

function is_reserved( id : identifier ) return boolean is
--is this still useful? true if an Ada reserved word
begin
  return id < reserved_top;
end is_reserved;

procedure startScanner is
  -- set up keywords, constants, and environment variables
  temp_id : identifier;
begin
  declareKeyword( keyword_t, "_keyword" );
  declareKeyword( abort_t, "abort" );
  declareKeyword( abs_t, "abs" );
  declareKeyword( abstract_t, "abstract" );
  declareKeyword( accept_t, "accept" );
  declareKeyword( access_t, "access" );
  declareKeyword( aliased_t, "aliased" );
  declareKeyword( all_t, "all" );
  declareKeyword( and_t, "and" );
  declareKeyword( array_t, "array" );
  declareKeyword( at_t, "at" );
  declareKeyword( begin_t, "begin" );
  declareKeyword( body_t, "body" );
  declareKeyword( case_t, "case" );
  declareKeyword( constant_t, "constant" );
  declareKeyword( declare_t, "declare" );
  declareKeyword( delay_t, "delay" );
  declareKeyword( delta_t, "delta" );
  declareKeyword( digits_t, "digits" );
  declareKeyword( do_t, "do" );
  declareKeyword( else_t, "else" );
  declareKeyword( elsif_t, "elsif" );
  declareKeyword( end_t, "end" );
  declareKeyword( entry_t, "entry" );
  declareKeyword( exception_t, "exception" );
  declareKeyword( exit_t, "exit" );
  declareKeyword( for_t, "for" );
  declareKeyword( function_t, "function" );
  declareKeyword( generic_t, "generic" );
  declareKeyword( goto_t, "goto" );
  declareKeyword( if_t, "if" );
  declareKeyword( in_t, "in" );
  declareKeyword( is_t, "is" );
  declareKeyword( limited_t, "limited" );
  declareKeyword( loop_t, "loop" );
  declareKeyword( mod_t, "mod" );
  declareKeyword( new_t, "new" );
  declareKeyword( not_t, "not" );
  declareKeyword( null_t, "null" );
  declareKeyword( of_t, "of" );
  declareKeyword( or_t, "or" );
  declareKeyword( others_t, "others" );
  declareKeyword( out_t, "out" );
  declareKeyword( package_t, "package" );
  declareKeyword( pragma_t, "pragma" );
  declareKeyword( private_t, "private" );
  declareKeyword( procedure_t, "procedure" );
  declareKeyword( protected_t, "protected" );
  declareKeyword( raise_t, "raise" );
  declareKeyword( range_t, "range" );
  declareKeyword( record_t, "record" );
  declareKeyword( rem_t, "rem" );
  declareKeyword( renames_t, "renames" );
  declareKeyword( requeue_t, "requeue" );
  declareKeyword( return_t, "return" );
  declareKeyword( reverse_t, "reverse" );
  declareKeyword( select_t, "select" );
  declareKeyword( separate_t, "separate" );
  declareKeyword( subtype_t, "subtype" );
  declareKeyword( tagged_t, "tagged" );
  declareKeyword( task_t, "task" );
  declareKeyword( terminate_t, "terminate" );
  declareKeyword( then_t, "then" );
  declareKeyword( type_t, "type" );
  declareKeyword( until_t, "until" );
  declareKeyword( use_t, "use" );
  declareKeyword( when_t, "when" );
  declareKeyword( while_t, "while" );
  declareKeyword( with_t, "with" );
  declareKeyword( xor_t, "xor" );

  declareKeyword( env_t, "env" );
  declareKeyword( typeset_t, "typeset" );
  declareKeyword( unset_t, "unset" );
  declareKeyword( trace_t, "trace" );
  declareKeyword( help_t, "help" );
  declareKeyword( clear_t, "clear" );
  declareKeyword( jobs_t, "jobs" );
  declareKeyword( logout_t, "logout" );
  reserved_top := identifiers_top;

  -- From now on, not reserved

  declareKeyword( symbol_t, "_symbol" );
  declareKeyword( eof_t, "[End of File]" );

  declareIdent( variable_t, "root variable type", keyword_t );
  declareIdent( uni_numeric_t, "universal_numeric", variable_t, typeClass );
  declareIdent( uni_string_t, "universal_string", variable_t, typeClass );
  declareIdent( root_enumerated_t, "root enumerated", variable_t, typeClass );
  declareIdent( command_t, "command", variable_t, typeClass );
  declareIdent( file_type_t, "file_type", variable_t, typeClass );
  identifiers( identifiers_top-1 ).limit := true; -- limited type
  declareIdent( socket_type_t, "socket_type", variable_t, typeClass );
  identifiers( identifiers_top-1 ).limit := true; -- limited type
  declareIdent( universal_t, "universal_typeless", variable_t, typeClass );
  declareIdent( integer_t, "integer", uni_numeric_t, typeClass );
  declareIdent( natural_t, "natural", uni_numeric_t, typeClass );
  declareIdent( positive_t, "positive", uni_numeric_t, typeClass );
  declareIdent( short_short_integer_t, "short_short_integer", uni_numeric_t, typeClass );
  declareIdent( short_integer_t, "short_integer", uni_numeric_t, typeClass );
  declareIdent( long_integer_t, "long_integer", uni_numeric_t, typeClass );
  declareIdent( long_long_integer_t, "long_long_integer", uni_numeric_t, typeClass );
  declareIdent( character_t, "character", uni_string_t, typeClass );
  declareIdent( float_t, "float", uni_numeric_t, typeClass );
  declareIdent( short_float_t, "short_float", uni_numeric_t, typeClass );
  declareIdent( long_float_t, "long_float", uni_numeric_t, typeClass );
  declareIdent( boolean_t, "boolean", root_enumerated_t, typeClass );
  declareIdent( string_t, "string", uni_string_t, typeClass );
  declareIdent( duration_t, "duration", uni_numeric_t, typeClass );
  declareIdent( file_mode_t, "file_mode", root_enumerated_t, typeClass );

  declareIdent( strlit_t, "string literal", uni_string_t );
  declareIdent( charlit_t, "string literal", uni_string_t );
  declareIdent( number_t, "numeric literal", uni_numeric_t );

  declareStandardConstant( false_t, "false", boolean_t, "0" );
  declareStandardConstant( true_t, "true", boolean_t, "1" );
  declareStandardConstant( in_file_t, "in_file", file_mode_t, "0" );
  declareStandardConstant( out_file_t, "out_file", file_mode_t, "1" );
  declareStandardConstant( append_file_t, "append_file", file_mode_t, "2" );

  declareKeyword( create_t, "create" );
  declareKeyword( open_t, "open" );
  declareKeyword( close_t, "close" );
  declareKeyword( get_t, "get" );
  declareKeyword( get_line_t, "get_line" );
  declareKeyword( inkey_t, "inkey" );
  declareKeyword( put_t, "put" );
  declareKeyword( put_line_t, "put_line" );
  declareKeyword( new_line_t, "new_line" );
  declareKeyword( is_open_t, "is_open" );
  declarekeyword( end_of_file_t, "end_of_file" );
  declareKeyword( end_of_line_t, "end_of_line" );
  declareKeyword( name_t, "name" );
  declareKeyword( mode_t, "mode" );
  declareKeyword( skip_line_t, "skip_line" );
  declareKeyword( line_t, "line" );
  declareKeyword( random_t, "numerics.random" );
  declareKeyword( shift_left_t, "numerics.shift_left" );
  declareKeyword( shift_right_t, "numerics.shift_right" );
  declareKeyword( rotate_left_t, "numerics.rotate_left" );
  declareKeyword( rotate_right_t, "numerics.rotate_right" );
  declareKeyword( shift_right_arith_t, "numerics.shift_right_arithmetic" );
  declareKeyword( sqrt_t, "numerics.sqrt" );
  declareKeyword( log_t, "numerics.log" );
  declareKeyword( exp_t, "numerics.exp" );
  declareKeyword( sin_t, "numerics.sin" );
  declareKeyword( cos_t, "numerics.cos" );
  declareKeyword( tan_t, "numerics.tan" );
  declareKeyword( cot_t, "numerics.cot" );
  declareKeyword( arcsin_t, "numerics.arcsin" );
  declareKeyword( arccos_t, "numerics.arccos" );
  declareKeyword( arctan_t, "numerics.arctan" );
  declareKeyword( arccot_t, "numerics.arccot" );
  declareKeyword( sinh_t, "numerics.sinh" );
  declareKeyword( cosh_t, "numerics.cosh" );
  declareKeyword( tanh_t, "numerics.tanh" );
  declareKeyword( coth_t, "numerics.coth" );
  declareKeyword( arcsinh_t, "numerics.arcsinh" );
  declareKeyword( arccosh_t, "numerics.arccosh" );
  declareKeyword( arctanh_t, "numerics.arctanh" );
  declareKeyword( arccoth_t, "numerics.arccoth" );
  declareKeyword( floor_t, "numerics.floor" );
  declareKeyword( ceiling_t, "numerics.ceiling" );
  declareKeyword( rounding_t, "numerics.rounding" );
  declareKeyword( unbiased_rounding_t, "numerics.unbiased_rounding" );
  declareKeyword( truncation_t, "numerics.truncation" );
  declareKeyword( remainder_t, "numerics.remainder" );
  declareKeyword( exponent_t, "numerics.exponent" );
  declareKeyword( fraction_t, "numerics.fraction" );
  declareKeyword( leading_part_t, "numerics.leading_part" );
  declareKeyword( copy_sign_t, "numerics.copy_sign" );
  declareKeyword( sturges_t, "numerics.sturges" );
  declareKeyword( max_t, "numerics.max" );
  declareKeyword( min_t, "numerics.min" );
  declareKeyword( machine_t, "numerics.machine" );
  declareKeyword( scaling_t, "numerics.scaling" );
  declareKeyword( value_t, "numerics.value" );
  declareKeyword( pos_t, "numerics.pos" );

  declareKeyword( glob_t, "strings.glob" );
  declareKeyword( match_t, "strings.match" );
  declareKeyword( element_t, "strings.element" );
  declareKeyword( slice_t, "strings.slice" );
  declareKeyword( index_t, "strings.index" );
  declareKeyword( index_non_blank_t, "strings.index_non_blank" );
  declareKeyword( count_t, "strings.count" );
  declareKeyword( replace_slice_t, "strings.replace_slice" );
  declareKeyword( insert_t, "strings.insert" );
  declareKeyword( overwrite_t, "strings.overwrite" );
  declareKeyword( sdelete_t, "strings.delete" );
  declareKeyword( trim_t, "strings.trim" );
  declareKeyword( length_t, "strings.length" );
  declareKeyword( head_t, "strings.head" );
  declareKeyword( tail_t, "strings.tail" );
  declareKeyword( val_t, "strings.val" );
  declareKeyword( image_t, "strings.image" );
  declareKeyword( field_t, "strings.field" );
  declareKeyword( mktemp_t, "strings.mktemp" );

  declareKeyword( cmd_argument_t, "command_line.argument" );
  declareKeyword( cmd_argcount_t, "command_line.argument_count" );
  declareKeyword( cmd_commandname_t, "command_line.command_name" );
  declareKeyword( cmd_setexit_t, "command_line.set_exit_status" );

  declareKeyword( system_t, "system" );
  declareKeyword( set_input_t, "set_input" );
  declareKeyword( set_output_t, "set_output" );
  declareKeyword( set_error_t, "set_error" );
  declareKeyword( reset_t, "reset" );
  declareKeyword( delete_t, "delete" );
  declareStandardConstant( standard_input_t, "standard_input", file_type_t,
    "<defined by parser>" );
  declareStandardConstant( standard_output_t, "standard_output", file_type_t,
    "<defined by parser>" );
  declareStandardConstant( standard_error_t, "standard_error", file_type_t,
    "<defined by parser>" );
  -- these should not be constants, but we don't have a function type yet
  declareStandardConstant( current_input_t, "current_input", file_type_t,
    "<defined by parser>" );
  declareStandardConstant( current_output_t, "current_output", file_type_t,
    "<defined by parser>" );
  declareStandardConstant( current_error_t, "current_error", file_type_t,
    "<defined by parser>" );

  -- built-in shell

  declareKeyword( pwd_t, "pwd" );
  declareKeyword( cd_t, "cd" );
  declareKeyword( wait_t, "wait" );
  --declareKeyword( command_t, "command" );

  -- package standard constants

  declareStandardConstant( "ASCII.NUL", character_t, "" & ASCII.NUL );
  declareStandardConstant( "ASCII.SOH", character_t, "" & ASCII.soh );
  declareStandardConstant( "ASCII.STX", character_t, "" & ASCII.stx );
  declareStandardConstant( "ASCII.ETX", character_t, "" & ASCII.etx );
  declareStandardConstant( "ASCII.EOT", character_t, "" & ASCII.eot );
  declareStandardConstant( "ASCII.ENQ", character_t, "" & ASCII.enq );
  declareStandardConstant( "ASCII.ACK", character_t, "" & ASCII.ack );
  declareStandardConstant( "ASCII.BEL", character_t, "" & ASCII.bel );
  declareStandardConstant( "ASCII.BS",  character_t, "" & ASCII.bs );
  declareStandardConstant( "ASCII.HT",  character_t, "" & ASCII.ht );
  declareStandardConstant( "ASCII.LF",  character_t, "" & ASCII.lf );
  declareStandardConstant( "ASCII.VT",  character_t, "" & ASCII.vt );
  declareStandardConstant( "ASCII.FF",  character_t, "" & ASCII.ff );
  declareStandardConstant( "ASCII.CR",  character_t, "" & ASCII.cr );
  declareStandardConstant( "ASCII.SO",  character_t, "" & ASCII.so );
  declareStandardConstant( "ASCII.SI",  character_t, "" & ASCII.si );
  declareStandardConstant( "ASCII.DLE", character_t, "" & ASCII.dle );
  declareStandardConstant( "ASCII.DC1", character_t, "" & ASCII.dc1 );
  declareStandardConstant( "ASCII.DC2", character_t, "" & ASCII.dc2 );
  declareStandardConstant( "ASCII.DC3", character_t, "" & ASCII.dc3 );
  declareStandardConstant( "ASCII.DC4", character_t, "" & ASCII.dc4 );
  declareStandardConstant( "ASCII.NAK", character_t, "" & ASCII.nak );
  declareStandardConstant( "ASCII.SYN", character_t, "" & ASCII.syn );
  declareStandardConstant( "ASCII.ETB", character_t, "" & ASCII.etb );
  declareStandardConstant( "ASCII.CAN", character_t, "" & ASCII.can );
  declareStandardConstant( "ASCII.EM",  character_t, "" & ASCII.em );
  declareStandardConstant( "ASCII.SUB", character_t, "" & ASCII.sub );
  declareStandardConstant( "ASCII.ESC", character_t, "" & ASCII.esc );
  declareStandardConstant( "ASCII.FS",  character_t, "" & ASCII.fs );
  declareStandardConstant( "ASCII.GS",  character_t, "" & ASCII.gs );
  declareStandardConstant( "ASCII.RS",  character_t, "" & ASCII.rs );
  declareStandardConstant( "ASCII.US",  character_t, "" & ASCII.us );
  declareStandardConstant( "ASCII.DEL", character_t, "" & ASCII.del );
  declareStandardConstant( "ASCII.Exclam",    character_t, "!" );
  declareStandardConstant( "ASCII.Quotation", character_t, """" );
  declareStandardConstant( "ASCII.Sharp",     character_t, "#" );
  declareStandardConstant( "ASCII.Dollar",    character_t, "$" );
  declareStandardConstant( "ASCII.Percent",   character_t, "%" );
  declareStandardConstant( "ASCII.Ampersand", character_t, "&" );
  declareStandardConstant( "ASCII.Colon",     character_t, ":" );
  declareStandardConstant( "ASCII.Semicolon", character_t, ";" );
  declareStandardConstant( "ASCII.Query",     character_t, "?" );
  declareStandardConstant( "ASCII.At_Sign",   character_t, "@" );
  declareStandardConstant( "ASCII.L_Bracket", character_t, "[" );
  declareStandardConstant( "ASCII.Back_Slash",character_t, "\" );
  declareStandardConstant( "ASCII.R_Bracket", character_t, "]" );
  declareStandardConstant( "ASCII.Circumflex",character_t, "^" );
  declareStandardConstant( "ASCII.Underline", character_t, "_" );
  declareStandardConstant( "ASCII.Grave",     character_t, "`" );
  declareStandardConstant( "ASCII.L_Brace",   character_t, "{" );
  declareStandardConstant( "ASCII.Bar",       character_t, "|" );
  declareStandardConstant( "ASCII.R_Brace",   character_t, "}" );
  declareStandardConstant( "ASCII.Tilde",     character_t, "~" );
-- technically, ASCII should also define...
--declareStandard( "ASCII.LC_A", Character_t:= 'a';
--declareStandard( "ASCII.LC_Z: constant Character:= 'z';

  -- package system constants

  declareStandardConstant( "System.System_Name", uni_string_t, "SYSTEM_NAME_BUSH" );
  declareStandardConstant( "System.Min_Int", uni_numeric_t, to_string( to_unbounded_string( long_float( integerOutputType'first+0.9 ) ) ) );
  -- out minimum integer is the limit of a long_float's mantissa.  should
  -- probably check that system.min_int isn't smaller, but Gnat gives bogus
  -- result on long_float( max_int ) if mantissa isn't big enough.
  declareStandardConstant( "System.Max_Int", uni_numeric_t, to_string( to_unbounded_string( long_float( integerOutputType'last-0.9 ) ) ) );
  -- out maximum integer is the limit of a long_float's mantissa
  -- probably check that system.min_max isn't smaller
  declareStandardConstant( "System.Max_Binary_Modulus", uni_numeric_t,
    long_long_float'image( long_long_float( system.max_binary_modulus ) ) );
  declareStandardConstant( "System.Max_Nonbinary_Modulus", uni_numeric_t,
    long_long_float'image( long_long_float( system.max_nonbinary_modulus ) ) );
  declareStandardConstant( "System.Max_Base_Digits", uni_numeric_t, system.max_base_digits'img );
  declareStandardConstant( "System.Max_Digits", uni_numeric_t, system.max_digits'img ); 
  declareStandardConstant( "System.Max_Mantissa", uni_numeric_t, system.max_mantissa'img );
  declareStandardConstant( "System.Fine_Delta", uni_numeric_t, system.fine_delta'img );
  declareStandardConstant( "System.Tick", uni_numeric_t, system.tick'img );
  declareStandardConstant( "System.Storage_Unit", uni_numeric_t, system.storage_unit'img );
  declareStandardConstant( "System.Word_Size", uni_numeric_t, system.word_size'img );
  declareStandardConstant( "System.Memory_Size", uni_numeric_t, system.memory_size'img );
  declareStandardConstant( "System.Default_Bit_Order", uni_string_t, system.default_bit_order'img );

  -- numerics package constants

  declareStandardConstant( numerics_e_t, "numerics.e", long_float_t,
      "2.7182818284590452354" );
  declareStandardConstant( numerics_log2_e_t, "numerics.log2_e", long_float_t,
      "1.4426950408889634074" );
  declareStandardConstant( numerics_log10_e_t, "numerics.log10_e", long_float_t,
      "0.43429448190325182765" );
  declareStandardConstant( numerics_ln10_t, "numerics.ln2", long_float_t,
      "0.69314718055994530942" );
  declareStandardConstant( numerics_ln2_t, "numerics.ln10", long_float_t,
      "2.30258509299404568402" );
  declareStandardConstant( numerics_pi_t, "numerics.pi", long_float_t,
      "3.14159265358979323846" );
  declareStandardConstant( numerics_pi_by_2_t, "numerics.pi_by_2", long_float_t,
      "1.57079632679489661923" );
  declareStandardConstant( numerics_pi_by_4_t, "numerics.pi_by_4", long_float_t,
      "0.78539816339744830962" );
  declareStandardConstant( numerics_pi_under_1_t, "numerics.pi_under_1", long_float_t,
      "0.31830988618379067154" );
  declareStandardConstant( numerics_pi_under_2_t, "numerics.pi_under_2", long_float_t,
      "0.63661977236758134308" );
  declareStandardConstant( numerics_sqrt_pi_under_2_t, "numerics.sqrt_pi_under_2",
      long_float_t, "1.12837916709551257390" );
  declareStandardConstant( numerics_sqrt_2_t, "numerics.sqrt_2", long_float_t,
      "1.4142135623730950488" );
  declareStandardConstant( numerics_sqrt_2_under_1_t, "numerics.sqrt_2_under_1", long_float_t,
      "0.70710678118654752440" );

  for i in 1..Environment_Count loop
      init_env_ident( Environment_Value( i ) );
  end loop;

  -- OLDPWD/PWD should be declared here if they don't already exist

  findIdent( to_unbounded_string( "PATH" ), temp_id );
  if temp_id = eof_t then
     declareIdent( temp_id, "PATH", uni_string_t );
  end if;
  if rshOpt then
     identifiers( temp_id).class := constClass;
  end if;
  findIdent( to_unbounded_string( "PWD" ), temp_id );
  if temp_id = eof_t then
     declareIdent( temp_id, "PWD", uni_string_t );
  end if;
  findIdent( to_unbounded_string( "OLDPWD" ), temp_id );
  if temp_id = eof_t then
     declareIdent( temp_id, "OLDPWD", uni_string_t );
  end if;
  findIdent( to_unbounded_string( "HOME" ), temp_id );
  if temp_id = eof_t then
     declareIdent( temp_id, "HOME", uni_string_t );
  end if;

  -- clear history

  historyNext := 1;
  for i in historyArray'range loop
      history(i) := null_unbounded_string;
  end loop;
end startScanner;

procedure resetScanner is
-- restart the scanner, discarding all declarations
begin
  identifiers_top := identifier'first;
  reserved_top := identifier'last;
  blocks_top := block'first;
  error_found := false;
  onlyAda95 := false;
  exit_block := false;
  restriction_no_auto_declarations := false;
  syntax_check := false;
  last_status := 0;
  done := false;
  trace := false;
  cmdpos := 3;
  startScanner; -- not very elegant
end resetScanner;

procedure shutdownScanner is
-- not much to shut down
begin
  null;
end shutdownScanner;

function Escape_String( s : unbounded_string ) return unbounded_string is
-- create a printable string my marking unprintable characters with
-- "[# ascii-code]"
  newstr : unbounded_string;
begin
  newstr := Null_Unbounded_String;
  for i in 1..length( s ) loop
      if Element( s, i ) < ' ' or Element( s, i ) > '~' then
         newstr := newstr & "[#" & character'pos( Element( s, i ) )'img & "]";
      else
         newstr := newstr & Element( s, i );
      end if;
  end loop;
  return newstr;
end Escape_String;

procedure Put_Identifier( id : identifier ) is
-- show an identifier's name, value and attributes on standard output
begin
  if not identifiers( id ).deleted then
     put( identifiers( id ).name );
     if id <= reserved_top-1 then
	put_line( " ( AdaScript reserved word )" );
	return;
     end if;
     if identifiers( id ).kind /= keyword_t then
        put( " = " );
     end if;
   -- should really used root type to determine quoting
     if identifiers( id ).kind = string_t then
         put( '"' );
         put( Escape_String( identifiers( id ).value ) );
         put( '"' );
     elsif identifiers( id ).kind = character_t then
         put( "'" );
         put( Escape_String( identifiers( id ).value ) );
         put( "'" );
     else
         put( Escape_String( identifiers( id ).value ) );
     end if;
     put( " ( " );
     if identifiers( id ).import then
        put( "imported " );
     end if;
     if identifiers( id ).export then
        put( "exported " );
     end if;
     if identifiers( id ).volatile then
        put( "volatile " );
     end if;
     case identifiers( id ).class is
     when subClass =>
        put( "subtype of " );
     when typeClass =>
        put( "new type of " );
     when constClass =>
        put( "constant " );
     when others =>
        put( "identfier of type " );
     end case;
     if identifiers( identifiers( id ).kind ).limit then
        put( "limited " );
     end if;
     if identifiers( id ).kind = keyword_t then
	put( "keyword" );
     else
        put( identifiers( identifiers( id ).kind ).name );
     end if;
     put_line( " )" );
  end if;
end Put_Identifier;

procedure RefreshVolatile( id : identifier ) is
-- look up a volatile variable and refresh it
-- (volatile variables are assumed to be environment variables)
  key : unbounded_string;
  ev  : unbounded_string;
begin
  key := identifiers( token ).name & "=";
  for i in 1..Environment_Count loop
      ev := to_unbounded_string( Environment_Value( i ) );
      if Head( ev, length( key ) ) = key then
         identifiers( token ).value :=
             Tail( ev, length( ev ) - length( key ) );
        end if;
  end loop;
end RefreshVolatile;

function get_uni_type( original : identifier ) return identifier is
-- dereference identifier until we find the universal type that this type
-- is based on.  quit if a circular relationship is suspected
-- errors result in type universal
  temp_id : identifier;
  count   : natural := 0;
begin
  -- safety check
  if identifiers( original ).kind = eof_t then
        err( "type expected" );
        return universal_t;
  elsif identifiers( original ).kind = keyword_t then
        err( "type expected, not a keyword" );
        return universal_t;
  end if;
  temp_id := original;
  while identifiers( temp_id ).kind /= variable_t loop
     temp_id := identifiers( temp_id ).kind;
     count := count + 1;
     if count >= 100 then
        err( "circular type relationship" );
        exit;
     end if;
  end loop;
  return temp_id;
end get_uni_type;

function get_base_type( original : identifier ) return identifier is
-- dereference identifier until we find the original, parent root type
-- (ie for types declared with "subtype", roll them back to their
-- parent types)
-- quit if a circular relationship is suspected
-- errors result in type universal
-- Original should be a type
  temp_id : identifier;
  count   : natural := 0;
begin
  -- safety check
  if identifiers( original ).kind = eof_t then
        err( "type expected" );
        return universal_t;
  elsif identifiers( original ).kind = keyword_t then
        err( "type expected, not a keyword" );
        return universal_t;
  end if;
  temp_id := original;
  while identifiers( temp_id ).class = subClass loop
     temp_id := identifiers( temp_id ).kind;
     count := count + 1;
     if count >= 100 then
        err( "circular type relationship" );
        exit;
     end if;
  end loop;
  return temp_id;
end get_base_type;

function get_class_name( class : anIdentifierClass ) return string is
begin
  if class = constClass then
     return "constant";
  elsif class = typeClass then
     return "type";
  elsif class = subClass then
     return "subtype";
  else
     return "variable";
  end if;
end get_class_name;

function class_ok( id : identifier; class : anIdentifierClass ) return boolean is
-- check identifier is of a certain class
begin
  if identifiers( id ).class /= class then
     if id = eof_t then
	err( "internal error: eof given to class_ok(1)" );
     elsif id < reserved_top then
        err_previous( "a reserved word is not a " &
           get_class_name( class ) );
     else
        err_previous( '"' & to_string( identifiers( id ).name ) &
           """ is a " &
	   get_class_name( identifiers( id ).class ) &
	   ", not a " &
           get_class_name( class ) );
     end if;
     return false;
  end if;
  return true;
end class_ok;

function class_ok( id : identifier; c1,c2 : anIdentifierClass ) return boolean is
-- check identifier is of a certain class
begin
  if identifiers( id ).class /= c1 and identifiers( id ).class /= c2 then
     if id = eof_t then
	err( "internal error: eof given to class_ok(2)" );
     elsif id < reserved_top then
        err_previous( "a reserved word is not a " &
           get_class_name( c1 ) &
           " or a " &
           get_class_name( c2 ) );
     else
        err_previous( '"' & to_string( identifiers( id ).name ) &
           """ is a " &
	   get_class_name( identifiers( id ).class ) &
	   ", not a " &
           get_class_name( c1 ) &
           " or a " &
           get_class_name( c2 ) );
     end if;
     return false;
  end if;
  return true;
end class_ok;

function uni_types_ok( left_type, right_type : identifier ) return boolean is
-- check that the two types are extended from a command universal type
  effective_left_type : identifier;
  effective_right_type : identifier;
  msg : unbounded_string;
  ch : character;
begin
  if not class_ok( left_type, typeClass, subClass ) then
     return false;
  elsif not class_ok( right_type, typeClass, subClass ) then
     return false;
  elsif left_type = variable_t or right_type = variable_t then
     return true;
  end if;
  effective_left_type := get_uni_type( left_type );
  effective_right_type := get_uni_type( right_type );
  if effective_left_type = universal_t or effective_right_type = universal_t then
     return true;
  end if;
  if effective_left_type /= effective_right_type then
    msg := "type " & identifiers( left_type ).name;
    if effective_left_type = root_enumerated_t then
       msg := msg & " (an enumerated type)";
    elsif identifiers( left_type ).kind /= variable_t then
       msg := msg & " (a";
       ch := Element( identifiers( effective_left_type ).name, 1 );
       if ch='a' or ch='e' or ch='i' or ch='o' or
          ch='A' or ch='E' or ch='I' or ch='O' then
          msg := msg & "n";
       end if;
       msg := msg & " " & identifiers( effective_left_type ).name & ")";
    end if;
    msg := msg & " is inherently different from a";
    ch := Element( identifiers( right_type ).name, 1 );
    if ch='a' or ch='e' or ch='i' or ch='o' or
       ch='A' or ch='E' or ch='I' or ch='O' then
       msg := msg & "n";
    end if;
    msg := msg & " " & identifiers( right_type ).name;
    if effective_right_type = root_enumerated_t then
       msg := msg & " (an enumerated type)";
    elsif identifiers( right_type ).kind /= variable_t then
       msg := msg & " (a";
       ch := Element( identifiers( effective_right_type ).name, 1 );
       if ch='a' or ch='e' or ch='i' or ch='o' or
          ch='A' or ch='E' or ch='I' or ch='O' then
          msg := msg & "n";
       end if;
       msg := msg & " " & identifiers( effective_right_type ).name & ")";
    end if;
    err_previous( to_string( msg ) );
    return false;
  end if;
  return true;
end uni_types_ok;

function base_types_ok( left_type, right_type : identifier ) return boolean is
-- check that the two types are logically compatible
  effective_left_type : identifier;
  effective_right_type : identifier;
begin
  if not uni_types_ok( left_type, right_type ) then
     return false;
  end if;
  if left_type = universal_t or right_type = universal_t then
     return true;
  end if;
  effective_left_type := get_base_type( left_type );
  effective_right_type := get_base_type( right_type );

--put( identifiers( left_type ).name );
--put( " base is => " );
--put( identifiers( effective_left_type ).name ); 
--put( " vs. " );
--put( identifiers( right_type ).name );
--put( "  base is=> " );
--put_line( identifiers( effective_right_type ).name ); 

  if effective_left_type = uni_numeric_t and get_uni_type( right_type ) = uni_numeric_t then
     return true;
  end if;
  if effective_left_type = uni_string_t and get_uni_type( right_type ) = uni_string_t then
     return true;
  end if;
  if effective_right_type = uni_numeric_t and get_uni_type( left_type ) = uni_numeric_t then
     return true;
  end if;
  if effective_right_type = uni_string_t and get_uni_type( left_type ) = uni_string_t then
     return true;
  end if;
  if effective_left_type /= effective_right_type then
     err_previous( "type '" & to_string( identifiers( left_type ).name ) &
          "' is not compatible with type '" &
          to_string( identifiers( right_type ).name ) & "'" );
     return false;
  end if;
  return true;
end base_types_ok;

procedure getCommandLine ( cmdline : out unbounded_string;
  token_firstpos, token_lastpos : out natural ) is
  -- like getCommandLine function, but return token posn
  line_firstpos, line_lastpos : natural;
  indent  : natural;
  len     : natural;
begin
  line_firstpos := cmdpos;                           -- start at current pos
  line_lastpos := cmdpos;                            -- start at current pos

  -- find beginning and end of command line
  -- (as it appears in the byte code)

  if line_firstpos > 1 then                          -- sane value?
     line_firstpos := line_firstpos - 1;             -- search for previous
     while script( line_firstpos ) /= ASCII.NUL loop -- ASCII.NUL
	line_firstpos := line_firstpos - 1;
     end loop;
  end if;
  if line_lastpos <= script'length then              -- sane value?
     while script( line_lastpos ) /= ASCII.NUL loop  -- look for next
       line_lastpos := line_lastpos + 1;             -- ASCII.NUL
     end loop;                                       -- or this one if
  end if;                                            -- on one
  if line_lastpos - line_firstpos <= 2 then          -- a blank line?
     cmdLine := null_unbounded_string;               -- return null string
     token_firstpos := 1;
     token_lastpos := 1;
     return;
  end if;

  -- skip ASCII.NUL's and record indentation

  line_firstpos := line_firstpos + 1;
  line_lastpos := line_lastpos - 1;
  indent := natural( integer( character'pos( script( line_firstpos ) ) - 1 ) );
  line_firstpos := line_firstpos+1;

  -- find token in command line

  if firstpos > line_firstpos then                  -- token on line?
     token_firstpos := firstpos-line_firstpos+1;    -- position in
     token_lastpos := lastpos-line_firstpos+1;      -- returned string
     cmdline := null_unbounded_string;              -- begin decompression
     for i in line_firstpos..line_lastpos loop      -- for bytes in script
         if script( i ) > '~' then                  -- a byte code? expand
	    cmdline := cmdline & identifiers( character'pos( script(i) )-128 ).name;
	    len := length( identifiers( character'pos( script(i) ) - 128 ).name );
            if firstpos = lastpos and firstpos = i then -- tokenized keyword?
	       token_lastpos := token_lastpos + len-1; -- adjust end position
            elsif lastpos > i then                     -- token shifted?
	       token_lastpos := token_lastpos + len-1; -- adjust
               if firstpos > i then
	          token_firstpos := token_firstpos + len-1;
               end if;
            end if;
         else                                       -- not a code?
	    cmdline := cmdline & script( i );       -- just add
         end if;
     end loop;                                      -- for all codes
     token_firstpos := token_firstpos + indent;     -- adjust token position
     token_lastpos := token_lastpos + indent;       -- for the size of ident
  else                                              -- not processed yet?
     token_firstpos := 1;                           -- position at
     token_lastpos := 1;                            -- first character
     cmdline := null_unbounded_string;              -- same, without
     for i in line_firstpos..line_lastpos loop      -- token stuff...
         if script( i ) > '~' then
	    cmdline := cmdline & identifiers( character'pos( script(i) )-128 ).name;
	    len := length( identifiers( character'pos( script(i) ) - 128 ).name );
         else
	    cmdline := cmdline & script( i );
         end if;
     end loop;
  end if;
  insert( cmdline, 1, to_string( indent * " " ) ); -- expand indentation
  if token_firstpos > length( cmdline ) then       -- read past end of cmd?
     token_firstpos := line_lastpos+1-line_firstpos; -- treat token as
     token_lastpos := token_firstpos;              -- one char past the end
  end if;
end getCommandLine;

function getCommandLine return unbounded_string is
  -- return current command line, fully indented, but
  -- not including the LF separating commands
  -- Tokens aren't expanded yet.
  firstpos, lastpos : natural;
  cmdline : unbounded_string;
begin
  getCommandLine( cmdline, firstpos, lastpos );
  return cmdline;
end getCommandLine;

procedure get_next_token is
-- the main scanner procedure, interpret the input text and
-- summarize it as a token.  Declare new, unknown identifiers
-- on the identifier stack.
  id   : identifier;
  word : unbounded_string;
  ch   : character;
  -- ch is a character buffer to reduce array accesses.  Really,
  -- these should be optimized away by the compiler, but you'd
  -- be surprised by what a compiler won't optimize away...
begin

  -- Out of data?  Never any more data.

  if token = eof_t then
     return;
  end if;

  -- End of line?  What we do depends on where we're reading from

<<redo>>

   ch := script( cmdpos );
   while ch = ASCII.NUL loop
      lineno := lineno + 1;
      if trace then
         if syntax_check or (not exit_block and not error_found) then
	    cmdpos := cmdpos + 2; -- first character of next command
            put( standard_error, "=> " & '"' );
            put( standard_error, getCommandLine );
            put( standard_error, """ [" );
            put( standard_error, lineno'img );
            put_line( standard_error, "]" );
	    cmdpos := cmdpos - 2;
         end if;
      end if;
      cmdpos := cmdpos+2; -- skip LF and indent marker
      ch := script( cmdpos );
   end loop;

  -- skip any white space before token

  if ch = ' ' or ch = ASCII.HT then
     cmdpos := cmdpos + 1;
     while script( cmdpos ) = ' ' or script( cmdpos ) = ASCII.HT loop
        cmdpos := cmdpos + 1;
     end loop;
     ch := script( cmdpos );
  end if;

  -- token begins here

  firstpos := cmdpos;
  lastpos := cmdpos;

  -- check for compressed token

  if ch > '~' then
     token := identifier( character'pos( script( cmdpos ) ) - 128 );
     cmdpos := cmdpos + 1;
     return;
  end if;

  -- Control characters and leading underscores filtered out in tokenize stage

  if is_letter( ch ) then
     lastpos := lastpos+1;
     ch := script( lastpos );
     while is_alphanumeric( ch ) or ch = '_' or ch = '.' loop
           if script( lastpos..lastpos+1 ) = ".." then -- ".."
              exit;
           end if;
           lastpos := lastpos+1;
           exit when ch = ASCII.NUL;
	   ch := script( lastpos );
     end loop;
     id := eof_t;
     -- case insensitive by lower-casing word
     -- word := toLower( To_Unbounded_string( slice( command, cmdpos, lastpos-1 ) ) );
     -- the problem is that commands and environment variables are case-sensitve
     -- so we're case-sensitive
     lastpos := lastpos - 1;
     word := To_Unbounded_string( script( cmdpos..lastpos ) );
     for i in reverse 1..identifiers_top-1 loop
         if identifiers( i ).name = word then
            id := i;
            exit;
         end if;
     end loop;
     if id = eof_t then
        declareIdent( token, word, new_t );
     else
        if identifiers( id ).deleted then
           identifiers( id ).deleted := false;
           identifiers( id ).kind := new_t;
        end if;
        token := id;
     end if;
     cmdpos := lastpos+1;
     return;
  elsif is_digit( ch ) then
     -- number
     lastpos := lastpos+1;
     ch := script( lastpos );
     while is_digit( ch ) or ch = '_' or ch = '.' loop
         if script( lastpos..lastpos+1 ) = ".." then -- ".."
            exit;
         end if;
         lastpos := lastpos+1;
	 ch := script( lastpos );
         exit when ch = ASCII.NUL;
     end loop;
     lastpos := lastpos - 1;
     if script( lastpos ) = '.' then
	err( "no digits after decimal" );
	return;
     end if;
     identifiers( number_t ).value := To_Unbounded_String( script( cmdpos..lastpos ) );
     token := number_t;
     cmdpos := lastpos+1;
     return;
  elsif ch = ''' then
     -- character literal
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while script( lastpos ) /= ''' loop
         lastpos := lastpos+1;
     end loop;
     identifiers( charlit_t ).value := To_Unbounded_String( script( cmdpos..lastpos-1 ) );
     cmdpos := lastpos+1;
     token := charlit_t;
     return;
  elsif ch = '"' then
     -- string literal
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while script( lastpos ) /= '"' loop
         lastpos := lastpos+1;
     end loop;
     identifiers( strlit_t ).value := To_Unbounded_String( script( cmdpos..lastpos-1 ) );
     token := strlit_t;
     cmdpos := lastpos+1; -- skip last "
     return;
  elsif ch = '`' then
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while script( lastpos ) /= '`' loop
         lastpos := lastpos+1;
     end loop;
     identifiers( string_t ).value := To_Unbounded_String( script( cmdpos..lastpos-1 ) );
     token := strlit_t;
     cmdpos := lastpos+1; -- skip last `
     return;
  else
     -- other punctuation symbols
     case ch is
     when '$' =>
          cmdpos := cmdpos + 1;
	  if script( cmdpos ) = '$' then
	     cmdpos := cmdpos + 1;
	  elsif script( cmdpos ) = '#' then
	     cmdpos := cmdpos + 1;
	  elsif script( cmdpos ) = '?' then
	     cmdpos := cmdpos + 1;
	  elsif script( cmdpos ) >= '0' and script( cmdpos ) <= '9' then
             cmdpos := cmdpos + 1;
          end if;
     when '#' =>
	         cmdpos := cmdpos + 1;
    	     while script( cmdpos ) /= ASCII.NUL loop
	            cmdpos := cmdpos + 1;
          end loop;
          goto redo;
     when '-' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '-' then
	            while script( cmdpos ) /= ASCII.NUL loop
	               cmdpos := cmdpos + 1;
             end loop;
             goto redo;
          end if;
     when '=' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '>' then
             cmdpos := cmdpos + 1;
          end if;
     when ':' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '=' then
             cmdpos := cmdpos + 1;
          end if;
     when '*' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '*' then
             cmdpos := cmdpos + 1;
          end if;
     when '>' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '=' then
             cmdpos := cmdpos + 1;
          end if;
     when '<' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '=' then
             cmdpos := cmdpos + 1;
          end if;
     when '/' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '=' then
             cmdpos := cmdpos + 1;
          end if;
     when '.' =>
          cmdpos := cmdpos + 1;
          if script( cmdpos ) = '.' then
             cmdpos := cmdpos + 1;
          end if;
     when '_' =>
          err( "Leading underscores not allowed in identifiers" );
          return;
     when '!' =>
	  if script( cmdpos+1 ) = '=' then
             err( "/= expected" );
          end if;
          return;
     when others =>
	         cmdpos := cmdpos + 1;
     end case;
     lastpos := cmdpos-1;
     identifiers( symbol_t ).value := To_Unbounded_String(
        script( firstpos..lastpos ) );
     token := symbol_t;
     return;
  end if;
end get_next_token;

procedure expect( expected_token : identifier ) is
-- check for the specified identifier.  If the current token matches,
-- get the next token otherwise show an error
begin
  if token /= expected_token then
     if expected_token = keyword_t then
        err( "keyword expected" );
     elsif expected_token = number_t then
        err( "number expected" );
     elsif expected_token = strlit_t then
        err( "string literal expected" );
     elsif expected_token = symbol_t then
        err( "symbol expected" );
     else
        err( to_string( identifiers( expected_token ).name ) & " expected" );
     end if;
  end if;
  get_next_token;
end expect;

procedure expect( expected_token : identifier; value : string ) is
-- check for the specified identifier and value.  If the current token
-- and its value matches, get the next token otherwise show an error
begin
  if value /= to_string( identifiers( token ).value ) then
      err( "'" & value & "' expected" );
      get_next_token;
  else
      get_next_token;
  end if;
end expect;

procedure Put_Token is
-- for debugging, show the current token, its value, type and properities
begin
  Put_Line( "'" & identifiers( token ).name & "'" );
  Put_Line( "    Token type = " & identifiers( identifiers( token ).kind ).name );
  Put_Line( "    Token value = '" & escape_string( identifiers( token ).value ) & "'" );
  Put( "    Token properties = " );
  if identifiers( token ).import then
     Put( "import " );
  end if;
  if identifiers( token ).export then
     Put( "export " );
  end if;
  if is_reserved( token ) then
     Put( "(reserved keyword) " );
  end if;
  New_Line;
end Put_Token;

procedure skipWhiteSpace is
-- move command pos to the first non-white space character
begin
  firstPos := cmdpos;
  while script( firstPos ) /= ASCII.NUL loop
    exit when script( firstPos ) /= ' ';
    firstPos := firstPos + 1;
  end loop;
  cmdPos := firstPos;
end skipWhiteSpace;

procedure getShellWord( wordType : out aShellWordType;
   word, pattern : out unbounded_string;  First : boolean := false ) is
-- return a shell "word" with quotations removed, delimited by a trailing
-- space, the end of the line, or an open paranthesis (if the first word--
-- that is, the command to run)
  ch          : character;
  pos         : integer;
  firstPos    : integer;
  inSQuote    : boolean := false; -- in single quoted part
  inDQuote    : boolean := false; -- in double quoted part
  inBackslash : boolean := false; -- in backslash part
  escaping    : boolean := false; -- escaping glob chars

begin

  -- skip leading spaces

  skipWhiteSpace;
  firstPos := cmdpos;
  wordType := normalWord;

  -- return word

  word := null_unbounded_string;
  pattern := null_unbounded_string;
  if script( firstPos ) = ASCII.NUL then
     err( "; expected" );
     return;
  end if;
  pos := firstPos;

  -- leading semi-colon always stands alone as a word

  if script( pos ) = ';' then
     cmdPos := pos+1;
     word := semicolon_string;
     pattern := semicolon_string;
     wordType := semicolonWord;
     return;
  elsif script( pos ) = '|' then
     cmdPos := pos+1;
     word := verticalbar_string;
     pattern := verticalbar_string;
     wordType := pipeWord;
     return;
  end if;

  while script( pos ) /= ASCII.NUL loop
    ch := script( pos );
    if ch = '"' and not inSQuote and not inBackslash then
       inDQuote := not inDQuote;
    elsif ch = ''' and not inDQuote and not inBackslash then
       inSQuote := not inSQuote;
       escaping := inSQuote;
    elsif ch = '\' and not inSQuote and not inBackslash then
       inBackslash := true;
       pattern := pattern & "\";
    elsif ch > '~' then
       word := word & identifiers( character'pos( ch )-128 ).name;
    else
       -- terminating characters (whitespace or semi-colon)
       exit when (ch = ' ' or ch = ASCII.HT or ch = ';' or ch = '|' )
          and not inDQuote and not inSQuote and not inBackslash;
       -- and terminate if looks like AdaScript-style parameters
       exit when First and script( pos ) = '(';
       -- backslash => user already escaped it
       if escaping and not inBackslash then
          case ch is
          when '*' => pattern := pattern & "\";
	  when '[' => pattern := pattern & "\";
	  when '\' => pattern := pattern & "\";
	  when '?' => pattern := pattern & "\";
          when others => null;
          end case;
       end if;
       pattern := pattern & ch;
       word := word & ch;
       if inBackslash then
          inBackslash := false;
       end if;
    end if;
    pos := pos + 1;
  end loop;
  cmdPos := pos;
  if inSQuote then
     err( "missing single quote mark" );
  elsif inDQuote then
     err( "missing double quote mark" );
  end if;
end getShellWord;


------------------------------------------------------
-- Saving/Restoring Position
--
------------------------------------------------------

procedure markScanner( scannerState : out aScannerState ) is
-- mark the current position of the scanner, including the token
begin
  scannerState.token   := token;
  scannerState.first   := firstpos;
  scannerState.last    := lastpos;
  scannerState.cmdpos  := cmdpos;
  --scannerState.command := command;
  scannerState.lineno  := lineno;
  if token = symbol_t or token = strlit_t or token = charlit_t or token = number_t then
     scannerState.value := identifiers( token ).value;
  end if;
end markScanner;

procedure resumeScanning( scannerState : aScannerState ) is
-- resume at a previously marked position, restoring the token
begin
  token    := scannerState.token;
  firstpos := scannerState.first;
  lastpos  := scannerState.last;
  cmdpos   := scannerState.cmdpos;
  --command  := scannerState.command;
  lineno   := scannerState.lineno;
  if token = symbol_t or token = strlit_t or token = charlit_t or token = number_t then
     identifiers( token ).value := scannerState.value;
  end if;
end resumeScanning;


-- "BYTE CODE" GENERATION

type compressionInfo is record
     compressedScript     : unbounded_string; -- the compressed script being built
     --commandNotCompressed : boolean;          -- is this command being compressed?
     --checkForCompression  : boolean;          -- check this command for compressing?
end record;

procedure line2ByteCode( ci : in out compressionInfo;
  command : unbounded_string ) is
-- convert a source line into byte code, compressing tokens
-- and indent space

  id   : identifier;
  word : unbounded_string;
  decimalCount : natural;
  uncompressed_script : unbounded_string;
  newstr : unbounded_string;

begin

  -- get rid of white space

  if length( command ) = 0 then
     ci.compressedScript := ci.compressedScript & ASCII.SOH & ASCII.NUL;
     -- blank line? no indentation
     return;
  end if;

  cmdpos := 1;
  if Element( command, cmdpos ) = ' ' or Element( command, 
    cmdpos ) = ASCII.HT then
    while Element( command, cmdpos ) = ' ' or Element( 
       command, cmdpos ) = ASCII.HT loop
       cmdpos := cmdpos + 1;
       exit when cmdpos > length( command );  -- ignore if at end of line
    end loop;
    lastpos := cmdpos-1;  -- not really lastpos
	   if lastpos > 254 then -- hopefully, never, but harmless
	      lastpos := 254;    -- to truncate leading indentation
    end if;
    ci.compressedScript := ci.compressedScript & character'val( lastpos+1 );
    firstpos := cmdpos;
	   lastpos := cmdpos;
  else
    ci.compressedScript := ci.compressedScript & ASCII.SOH;
  end if;

  -- tokenize contents of line

<<next>> if cmdpos > length( command ) then
     --if ci.commandNotCompressed then
     --   ci.compressedScript := ci.compressedScript & "[UNCOMPRESS]";
     --   ci.commandNotCompressed := false;
     --end if;
     ci.compressedScript := ci.compressedScript & ASCII.NUL;
     return;
  end if;

  -- white space

  firstpos := cmdpos;
  lastpos := cmdpos;
  if Element( command, cmdpos ) = ' ' or Element( command, cmdpos ) = ASCII.HT then
     while Element( command, cmdpos ) = ' ' or Element( command, cmdpos ) = ASCII.HT loop
        cmdpos := cmdpos + 1;
        if cmdpos > length( command ) then  -- ignore if at end of line
           goto next;
        end if;
     end loop;
     lastpos := cmdpos-1;
     ci.compressedScript := ci.compressedScript & slice( command, firstpos, lastpos );
     goto next;

  -- illegal characters

  elsif is_control( Element( command, cmdpos ) ) or Element( command, cmdpos ) > '~' then
     err_tokenize( "Unexpected character ASCII" & character'pos( Element( command, cmdpos ) )'img, to_string( command ) );
     return;

  -- leading underscores

  elsif Element( command, cmdpos ) = '_' or
     -- identifiers with underscores not allowed, but we can't check for
     -- that here since we don't know yet if the identifier is part of a
     -- shell command or an AdaScript command (underscores in a shell
     -- command's parameters are OK).

  -- identifiers

  is_letter( Element( command, cmdpos ) ) then
     lastpos := cmdpos+1;
     if lastpos <= length( command ) then
        while is_alphanumeric( Element( command, lastpos ) ) or
              Element( command, lastpos ) = '_' or
              Element( command, lastpos ) = '.' loop
              if Element( command, lastpos ) = '.' and then lastpos < length( command ) then
                 if Element( command, lastpos+1 ) = '.' then -- ".."
                    exit;
                 end if;
              end if;
            lastpos := lastpos+1;
            exit when lastpos > length( command );
        end loop;
     end if;
     id := eof_t;
     -- case insensitive by lower-casing word
     -- word := toLower( To_Unbounded_string( slice( command, cmdpos, lastpos-1 ) ) );
     -- the problem is that commands and environment variables are case-sensitve
     -- so we're case-sensitive
     lastpos := lastpos - 1;
     word := To_Unbounded_string( slice( command, cmdpos, lastpos ) );
     cmdpos := lastpos+1;
     for i in reverse 1..identifiers_top-1 loop
         if identifiers( i ).name = word then
            id := i;
            exit;
         end if;
     end loop;
     if id < reserved_top then
	ci.compressedScript := ci.compressedScript & character'val( 128 + integer(id) );
	       goto next;
     end if;
     -- not a reserved word?  may be a shell command, but only if first word
     --ci.checkForCompression := true;
     --ci.commandNotCompressed := true;
  elsif is_digit( Element( command, cmdpos ) ) then
     -- number
     lastpos := cmdpos;
     decimalCount := 0;
     while is_digit( Element( command, lastpos)) or Element( command, lastpos) = '_' or
         Element( command, lastpos) = '.' loop
         if Element( command, lastpos ) = '.' and then lastpos < length( command ) then
            if Element( command, lastpos+1 ) = '.' then -- ".."
               exit;
            end if;
            decimalCount := decimalCount+1;
	    if decimalCount > 1 then
	       err_tokenize( "too many decimal points in floating point literal",  to_string( command ) );
	       return;
	    end if;
         end if;
         lastpos := lastpos+1;
         exit when lastpos > length( command );
     end loop;
     --identifiers( number_t ).value := To_Unbounded_String( Slice( command, cmdpos, lastpos-1 ) );
     cmdpos := lastpos;
     lastpos := lastpos-1;
  elsif Element( command, cmdpos ) = ''' then
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while Element( command, lastpos ) /= ''' loop
         lastpos := lastpos+1;
         exit when lastpos > length( command );
     end loop;
     if lastpos > length( command ) then
	err_tokenize( "missing single quote", to_string( command ) );
	return;
     else
        identifiers( charlit_t ).value := To_Unbounded_String( Slice( command, cmdpos, lastpos-1 ) );
        lastpos := lastpos+1; -- skip last '
        cmdpos := lastpos+1;
        token := charlit_t;
     end if;
  elsif Element( command, cmdpos ) = '"' then
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while Element( command, lastpos ) /= '"' loop
         lastpos := lastpos+1;
         exit when lastpos > length( command );
     end loop;
     if lastpos > length( command ) then
	err_tokenize( "missing double quote", to_string( command ) );
	return;
     else
        identifiers( strlit_t ).value := To_Unbounded_String( Slice( command, cmdpos, lastpos-1 ) );
        token := strlit_t;
        cmdpos := lastpos+1; -- skip last "
     end if;
  elsif Element( command, cmdpos ) = '`' then
     cmdpos := cmdpos+1;
     lastpos := cmdpos;
     while Element( command, lastpos ) /= '`' loop
         lastpos := lastpos+1;
         exit when lastpos > length( command );
     end loop;
     if lastpos > length( command ) then
	err_tokenize( "missing back quote", to_string( command ) );
	return;
     else
        identifiers( string_t ).value := To_Unbounded_String( Slice( command, cmdpos, lastpos-1 ) );
        token := strlit_t;
        cmdpos := lastpos+1; -- skip last `
     end if;
  else
     -- other punctuation symbols
     firstpos := cmdpos;
     case Element( command, cmdpos ) is
     when '$' =>
          if cmdpos < length( command ) then
             cmdpos := cmdpos + 1;
          end if;
     when '#' =>
           cmdpos := length( command ); -- for read of next line (or EOF)
     when '-' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '-' then
                cmdpos := length( command ); -- for read of next line (or EOF)
             end if;
          end if;
     when '=' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '>' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when ':' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '=' then
                cmdpos := cmdpos + 1;
		--if ci.checkForCompression  then
		--   ci.checkForCompression := false;
                --   ci.commandNotCompressed := false;
		--end if;
             end if;
          end if;
     when '*' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '*' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when '>' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '=' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when '<' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '=' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when '\' =>
          if cmdpos < length( command ) then
             cmdpos := cmdpos + 1;
          end if;
     when '/' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '=' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when '.' =>
          if cmdpos < length( command ) then
             if Element( command, cmdpos+1 ) = '.' then
                cmdpos := cmdpos + 1;
             end if;
          end if;
     when others => null; -- just that character
     end case;
     --identifiers( symbol_t ).value := To_Unbounded_String( Slice( command, firstpos, cmdpos ) );
     --token := symbol_t;
     lastpos := cmdpos;
     cmdpos := cmdpos + 1;
  end if;
  ci.compressedScript := ci.compressedScript &
    slice( command, firstpos, lastpos );
  goto next;
end line2ByteCode;

procedure compileCommand( command : unbounded_string ) is
  ci : compressionInfo;
begin
  if script /= null then
     free( script );
  end if;
  ci.compressedScript := to_unbounded_string( ASCII.SOH &
     character'val( 128 + eof_t ) & ASCII.NUL );
  --ci.commandNotCompressed := false;
  --ci.checkForCompression := true;
  cmdpos := 1;
  line2ByteCode( ci, command );
  ci.compressedScript := ci.compressedScript & ASCII.SOH &
     character'val( 128 + eof_t ) & ASCII.NUL;
  script := new string( 1..length( ci.compressedScript ) );
  script.all := to_string( ci.compressedScript );
end compileCommand;

procedure compileScript( firstLine : unbounded_string ) is
  ci : compressionInfo;
  newstr : unbounded_string;
  command : aliased unbounded_string := firstLine;
begin
  if script /= null then
     free( script );
  end if;
  ci.compressedScript := to_unbounded_string( ASCII.SOH &
     character'val( 128 + eof_t ) & ASCII.NUL );
  --ci.commandNotCompressed := false;
  --ci.checkForCompression := true;
  cmdpos := 1;
  -- parser loads first line...
  loop
     line2ByteCode( ci, command );
     exit when error_found;
     exit when not LineRead( command'access );
  end loop;
  ci.compressedScript := ci.compressedScript & ASCII.SOH & character'val(
    128 + eof_t ) & ASCII.NUL;

  --for i in 1..length( ci.compressedScript ) loop
  --    if Element( ci.compressedScript, i )  = ASCII.NUL then
  --       newstr := to_unbounded_string(
--	     "[#" &
--	     character'pos( Element( ci.compressedScript, i ) )'img &
--	     " ]" );
 --        newstr := newstr & to_unbounded_string( "" &
  --         ASCII.LF );
   --      --newstr := to_unbounded_string( "" & Element( script, i ) );
    --  elsif Element( ci.compressedScript, i ) < ' ' then
    --     newstr := to_unbounded_string(
--	     "[#" &
--	     character'pos( Element( ci.compressedScript, i ) )'img &
--	     " ]" );
 --     elsif Element( ci.compressedScript, i ) > '~' then
  --       newstr :=
--	     "[#" &
--	     character'pos( Element( ci.compressedScript, i ) )'img &
--	     " = " &
--	     identifiers(
--	        identifier( character'pos( Element( ci.compressedScript, i ) ) - 128 )
 --            ).name &
--	     " ]";
 --     else
  --       newstr := to_unbounded_string( "" &
   --         Element( ci.compressedScript, i ) );
    --  end if;
     -- put(  newstr  );
  --end loop;
 -- new_line;
 -- put_line( "Size =" & length( ci.compressedScript )'img );
  script := new string( 1..length( ci.compressedScript ) );
  script.all := to_string( ci.compressedScript );
end compileScript;


end scanner;
