
------------------------------------------------------------------------------
-- Lexical Scanner (the thing that reads your source code)                  --
-- Also, the semantic stuff.                                                --
--                                                                          --
--                                                                          --
-- Part of BUSH                                                             --
------------------------------------------------------------------------------
--                                                                          --
--                 Copyright (C) 2001 Ken O. Burtch & FSF                   --
--                                                                          --
-- This is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  This is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with this;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- This is maintained at http://www.vaxxine.com/pegasoft                    --
--                                                                          --
------------------------------------------------------------------------------

with system, ada.unchecked_deallocation, ada.strings.unbounded, script_io;
use ada.strings.unbounded, script_io;


package scanner is

------------------------------------------------------
-- Scanner Global Variables / Flags
--
------------------------------------------------------

isLoginShell : boolean := false; -- TRUE if a login shell

type anInputMode is ( interactive, fromScriptFile );
-- interactive - user is typing commands from the prompt
-- fromScriptFile - commands are executing from a script file

type identifier is new integer range 1..1200;
-- identifiers are identified by a unique number

--command    : aliased unbounded_string; -- current line
cmdpos     : natural := 0;     -- next char to read
firstpos   : natural := 0;     -- deliniates the last token
lastpos    : natural := 0;     -- deliniates the last token
done       : boolean := false; -- true if exiting shell
trace      : boolean := false; -- true if trace is on
onlyAda95  : boolean := false; -- pragma ada_95
restriction_no_auto_declarations : boolean := false; -- pragma restriction
inputMode  : anInputMode := interactive;
last_status: integer := 0;     -- status of last shell command
lineno     : natural := 0;

token       : identifier;       -- item we're currently looking at
error_found : boolean := false; -- true if an error occurred
exit_block  : boolean := false; -- true if "exit"ing
itself      : unbounded_string; -- value of @
itself_type : identifier;       -- type of @
itself_defined : boolean := false;  -- true if itself is meaningful
last_output : unbounded_string;
last_output_type : identifier;
syntax_check: boolean := false; -- true if syntax checking

-- Command Line Options

syntaxOpt   : boolean := false; -- true if -c
nosyntaxOpt : boolean := false; -- true if -n
rshOpt      : boolean := false; -- true if -r
verboseOpt  : boolean := false; -- true if -v
optionOffset: natural := 0;     -- offset to script parameters

type scriptptr is access all string;
script : scriptptr := null;
procedure free is new ada.unchecked_deallocation( string, scriptPtr );
firstScriptCommandOffset : constant natural := 5;

------------------------------------------------------
-- Command History
--
-- Command history is a circular queue
------------------------------------------------------

--type historyArray is array( 1..20 ) of unbounded_string;

--history    : historyArray;      -- the previous 20 commands
--historyNext : integer;          -- next to be overwritten


------------------------------------------------------
-- Identifiers (The Symbol Table)
--
------------------------------------------------------

type anIdentifierClass is ( constClass, subClass, typeClass, otherClass );
 
type declaration is record
        name     : unbounded_string := Null_Unbounded_String;
        kind     : identifier;
        value    : unbounded_string := Null_Unbounded_String;
        class    : anIdentifierClass := otherClass;
        import   : boolean := false;
        export   : boolean := false;
        volatile : boolean := false;
        limit    : boolean := false;
        deleted: boolean := false;
end record;

type identifiersArray is array ( identifier ) of declaration;
identifiers : identifiersArray;
identifiers_top : identifier := identifier'first;
reserved_top : identifier := identifier'last;
-- this arrangement means the last array element is never accessed

-- "Captain, the ship can't take any more!"

symbol_table_overflow : exception;


------------------------------------------------------
-- Predefined identifiers    
--

keyword_t : identifier;  -- .kind for an Ada keyword

------------------------------------------------------
-- other internal identifiers
--
-- users should never see these
------------------------------------------------------

abort_t    : identifier;
abs_t      : identifier;
abstract_t : identifier;
accept_t   : identifier;
access_t   : identifier;
aliased_t  : identifier;
all_t      : identifier;
and_t      : identifier;
array_t    : identifier;
at_t       : identifier;
begin_t    : identifier;
body_t     : identifier;
case_t     : identifier;
constant_t : identifier;
declare_t  : identifier;
delay_t    : identifier;
delta_t    : identifier;
digits_t   : identifier;
do_t       : identifier;
else_t     : identifier;
elsif_t    : identifier;
end_t      : identifier;
entry_t    : identifier;
exception_t : identifier;
exit_t     : identifier;
for_t      : identifier;
function_t : identifier;
generic_t  : identifier;
goto_t     : identifier;
if_t       : identifier;
in_t       : identifier;
is_t       : identifier;
limited_t  : identifier;
loop_t     : identifier;
mod_t      : identifier;
new_t      : identifier;
not_t      : identifier;
null_t     : identifier;
of_t       : identifier;
or_t       : identifier;
others_t   : identifier;
out_t      : identifier;
package_t  : identifier;
pragma_t   : identifier;
private_t  : identifier;
procedure_t : identifier;
protected_t : identifier;
raise_t    : identifier;
range_t    : identifier;
record_t   : identifier;
rem_t      : identifier;
renames_t  : identifier;
requeue_t  : identifier;
return_t   : identifier;
reverse_t  : identifier;
select_t   : identifier;
separate_t : identifier;
subtype_t  : identifier;
tagged_t   : identifier;
task_t     : identifier;
terminate_t : identifier;
then_t     : identifier;
type_t     : identifier;
until_t    : identifier;
use_t      : identifier;
when_t     : identifier;
while_t    : identifier;
with_t     : identifier;
xor_t      : identifier;

------------------------------------------------------
-- shell-specific commands
--
------------------------------------------------------

env_t      : identifier;
typeset_t  : identifier;
unset_t    : identifier;
trace_t    : identifier;
help_t     : identifier;
clear_t    : identifier;
jobs_t     : identifier;
logout_t   : identifier;

------------------------------------------------------
-- other internal identifiers
--
-- users should never see these
------------------------------------------------------

symbol_t      : identifier;  -- punctuation/etc., value = string of punctuation
--exitblock_t   : identifier;  -- generated by 'exit'
--exitprogram_t : identifier;  -- generated by 'return'
eof_t         : identifier;  -- end of statement / abort statement

------------------------------------------------------
-- predefined types
--
-- variable_t is the root type of all variables and
-- is only used to mark the fundamental types.
--
-- the fundamental types are univeral number, universal
-- string and universal (typeless).  All have .kind =
-- variable_t.
--
-- the basic Ada types are derived from the universal
-- types and have a .kind = some univeral type.
--
-- user types are, of course, derived from the basic
-- Ada types and have a .kind = some basic Ada type
------------------------------------------------------

variable_t            : identifier;

uni_numeric_t         : identifier;
uni_string_t          : identifier;
universal_t           : identifier;
root_enumerated_t     : identifier;
command_t             : identifier;

file_type_t           : identifier;
socket_type_t         : identifier;
integer_t             : identifier;
natural_t             : identifier;
positive_t            : identifier;
short_short_integer_t : identifier;
short_integer_t       : identifier;
long_integer_t        : identifier;
long_long_integer_t   : identifier;
character_t           : identifier;
float_t               : identifier;
short_float_t         : identifier;
long_float_t          : identifier;
boolean_t             : identifier;
string_t              : identifier;
duration_t            : identifier;
file_mode_t           : identifier;

------------------------------------------------------
-- predefined constants / built-in functions
--

strlit_t   : identifier;  -- string literal, value = the literal
charlit_t  : identifier;  -- character literal, value = the literal
number_t   : identifier;  -- numeric literal, value = the literal
false_t    : identifier;  -- boolean value 'false'
true_t     : identifier;  -- boolean value 'true'
in_file_t  : identifier;  -- file_mode value 'in_file'
out_file_t : identifier;  -- file_mode value 'out_file'
append_file_t : identifier; -- file_mode value 'append_file'
open_t     : identifier;  -- built-in text_io.open
create_t   : identifier;  -- built-in text_io.create
close_t    : identifier;  -- built-in text_io.close
get_t      : identifier;  -- built-in text_io.get
get_line_t : identifier;  -- built-in text_io.get_line
inkey_t    : identifier;  -- built-in inkey
put_t      : identifier;  -- built-in text_io.put
put_line_t : identifier;  -- built-in text_io.put_line
new_line_t : identifier;  -- built-in text_io.new_line
skip_line_t: identifier;  -- built-in text_io.skip_line
is_open_t  : identifier;  -- built-in text_io.is_open
end_of_file_t : identifier;  -- built-in text_io.end_of_file
end_of_line_t : identifier;  -- built-in text_io.end_of_line
line_t     : identifier;  -- built-in text_io.line
name_t     : identifier;  -- built-in text_io.name
mode_t     : identifier;  -- built-in text_io.mode
set_input_t : identifier;  -- built-in text_io.set_input
set_output_t : identifier;  -- built-in text_io.set_output
reset_t    : identifier;  -- built-in text_io.reset
delete_t   : identifier;  -- built-in text_io.delete
set_error_t : identifier;  -- built-in text_io.set_error
standard_input_t  : identifier; -- built-in text_io.standard_input
standard_output_t : identifier; -- built-in text_io.standard_output
standard_error_t  : identifier; -- built-in text_io.standard_error
current_input_t   : identifier; -- built-in text_io.current_input
current_output_t  : identifier; -- built-in text_io.current_output
current_error_t   : identifier; -- built-in text_io.current_error

-- Ada.Numerics.Float_Random

random_t    : identifier;  -- built-in ada.numerics.float_random.random

-- Interfaces

shift_left_t   : identifier;
shift_right_t  : identifier;
rotate_left_t  : identifier;
rotate_right_t : identifier;
shift_right_arith_t : identifier;

-- Ada.Numerics.Long_Elementary_Functions

sqrt_t       : identifier;
log_t        : identifier;
exp_t        : identifier;
sin_t        : identifier;
cos_t        : identifier;
tan_t        : identifier;
cot_t        : identifier;
arcsin_t     : identifier;
arccos_t     : identifier;
arctan_t     : identifier;
arccot_t     : identifier;
sinh_t       : identifier;
cosh_t       : identifier;
tanh_t       : identifier;
coth_t       : identifier;
arcsinh_t    : identifier;
arccosh_t    : identifier;
arctanh_t    : identifier;
arccoth_t    : identifier;

floor_t      : identifier;
ceiling_t    : identifier;
rounding_t : identifier;
unbiased_rounding_t : identifier;
truncation_t : identifier;
remainder_t  : identifier;
exponent_t   : identifier;
fraction_t   : identifier;
leading_part_t : identifier;
copy_sign_t  : identifier;
sturges_t    : identifier;
max_t        : identifier;
min_t        : identifier;
machine_t    : identifier;
scaling_t    : identifier;
value_t      : identifier;
pos_t        : identifier;
val_t        : identifier;
image_t      : identifier;

field_t      : identifier;
mktemp_t     : identifier;

numerics_e_t : identifier;
numerics_log2_e_t : identifier;
numerics_log10_e_t : identifier;
numerics_ln10_t : identifier;
numerics_ln2_t : identifier;
numerics_pi_t : identifier;
numerics_pi_by_2_t : identifier;
numerics_pi_by_4_t : identifier;
numerics_pi_under_1_t : identifier;
numerics_pi_under_2_t : identifier;
numerics_sqrt_pi_under_2_t : identifier;
numerics_sqrt_2_t : identifier;
numerics_sqrt_2_under_1_t : identifier;

-- Gnat.RegExp/Pat

glob_t      : identifier;  -- built-in gnat.regexp.match
match_t     : identifier;  -- built-in gnat.regpat.match

-- Ada.Strings.Unbounded

element_t    : identifier;
slice_t      : identifier;
index_t      : identifier;
index_non_blank_t : identifier;
count_t      : identifier;
replace_slice_t : identifier;
insert_t     : identifier;
overwrite_t  : identifier;
sdelete_t    : identifier;
trim_t       : identifier;
length_t     : identifier;
head_t       : identifier;
tail_t       : identifier;


-- Ada.Command_Line

cmd_argument_t    : identifier;
cmd_argcount_t    : identifier;
cmd_commandname_t : identifier;
cmd_setexit_t     : identifier;

-- Built-in

system_t   : identifier;  -- built-in system
pwd_t      : identifier;  -- built-in pwd
cd_t       : identifier;  -- built-in cd
wait_t     : identifier;  -- built-in wait
--command_t  : identifier;  -- built-in command

------------------------------------------------------
-- Subprograms
--
------------------------------------------------------


------------------------------------------------------
-- Declarations
--
------------------------------------------------------

procedure declareKeyword( id : out identifier; s : string );
-- initialize a new keyword / internal identifier

function deleteIdent( id : identifier ) return boolean;
-- delete an identifier, true if successful

procedure findIdent( name : unbounded_string; id : out identifier );
-- find an identifier, eof_t if failed

procedure init_env_ident( s : string );
-- initialize an environment variable

procedure declareIdent( id : out identifier; name : unbounded_string;
  kind : identifier; class : anIdentifierClass := otherClass );
-- declare an identifier

procedure declareStandardConstant( name : string;
  kind : identifier; value : string );
-- declare a system constant and we don't care what token it is

procedure err( msg : string );
-- if this is the first error encountered, display the message
-- set the token to eof_t to abort the parsing and set the
-- error_found flag to indicate that an error was encountered

------------------------------------------------------
-- Saving/Restoring Position  
--
------------------------------------------------------

type aScannerState is record
  token   : identifier;       -- the current token
  value   : unbounded_string; -- value (if symbol token)
  first   : natural;          -- firstpos value
  last    : natural;          -- lastpos value
  cmdpos  : natural;          -- cmdpos value
  --command : unbounded_string; -- command line
  lineno  : natural;
end record;

procedure markScanner( scannerState : out aScannerState );
-- mark the current position of the scanner, including the token

procedure resumeScanning( scannerState : aScannerState );
-- resume at a previously marked position, restoring the token

pragma inline( markScanner );
pragma inline( resumeScanning );


------------------------------------------------------
-- Scope / Block Statement Control
--
------------------------------------------------------

type blockDeclaration is record
  startpos        : long_integer;     -- where in script file block starts
  identifiers_top : identifier;       -- where block declarations begin
  newScope        : boolean := false; -- true if new identifier scope
  blockName       : unbounded_string := Null_Unbounded_String;
  state           : aScannerState;    -- the position on the line
end record;
type block is new integer range 1..32;
type blocksArray is array( block ) of blockDeclaration;
-- this arrangement means the last array element is never accessed

blocks : blocksArray; -- stack of scoping blocks / compound statements
blocks_top : block := block'first; -- next free stack position

block_table_overflow : exception;

procedure pushBlock( newScope : boolean := false;
  newName : string := "" );
-- start a new identifier scope, remember where we "parked"
-- if newScope is false, the new block is the start of a
-- multi-line statement (eg. an "if"); if true, the new
-- block is a nested declaration scope (eg. a "declare")

procedure pullBlock;
-- restore the previous identifier scope

procedure topOfBlock;
-- return to the top of the current block, read that line

function isLocal( id : identifier ) return boolean;
-- true if identifier is local to the current scope


------------------------------------------------------
-- Type Conversion              
--
------------------------------------------------------

function to_numeric( s : unbounded_string ) return long_float;
function to_numeric( id : identifier ) return long_float;
-- convert a string (or an identifier's string value) into a
-- floating-point number

function to_unbounded_string( f : long_float ) return unbounded_string;
-- convert a floating-point number back into a string

function Escape_String( s : unbounded_string ) return unbounded_string;
-- create a printable string my marking unprintable characters with
-- "[# ascii-code]"


------------------------------------------------------
-- Type Checking / Verification
--
------------------------------------------------------

function is_reserved( id : identifier ) return boolean;
-- true if an Ada reserved word

function get_uni_type( original : identifier ) return identifier;
-- dereference identifier until we find the universal type that this type
-- is based on.  quit if a circular relationship is suspected
-- errors result in type universal

function get_base_type( original : identifier ) return identifier;
-- dereference identifier until we find the original, parent root type
-- (ie for types declared with "subtype", roll them back to their
-- parent types)
-- quit if a circular relationship is suspected
-- errors result in type universal

function class_ok( id : identifier; class : anIdentifierClass ) return boolean;
function class_ok( id : identifier; c1,c2 : anIdentifierClass ) return boolean;
-- check identifier is of a certain class

function uni_types_ok( left_type, right_type : identifier ) return boolean;
-- check that the two types are extended from a command universal type

function base_types_ok( left_type, right_type : identifier ) return boolean;
-- check that the two types are logically compatible

pragma inline( is_reserved );

------------------------------------------------------
-- Scanning Tokens            
--
------------------------------------------------------

procedure get_next_token;
-- the main scanner procedure, interpret the input text and
-- summarize it as a token.  Declare new, unknown identifiers
-- on the identifier stack.

procedure expect( expected_token : identifier );
-- check for the specified identifier.  If the current token matches,
-- get the next token otherwise show an error

procedure expect( expected_token : identifier; value : string );
-- check for the specified identifier and value.  If the current token
-- and its value matches, get the next token otherwise show an error

pragma inline( expect );

------------------------------------------------------
-- Bourne Shell "Word" Processing    
--
------------------------------------------------------

procedure skipWhiteSpace;
-- advance to first non-white space token

type aShellWordType is ( normalWord, semicolonWord, pipeWord );

procedure getShellWord( wordType : out aShellWordType;
   word, pattern : out unbounded_string;  First : boolean := false );
-- get the next "word".  if first is true, treated as a command
-- word

------------------------------------------------------
-- Housekeeping               
--
------------------------------------------------------

procedure startScanner;
-- set up keywords, constants, and environment variables

procedure resetScanner;
-- restart the scanner, discarding all declarations

procedure shutdownScanner;
-- stop the scanner


------------------------------------------------------
-- Other Subprograms          
--
------------------------------------------------------

procedure Put_Token;
-- for debugging, show the current token, its value, type and properities

procedure Put_Identifier( id : identifier );
-- show an identifier's name, value and attributes on standard output

procedure Put_Trace( msg : string );
-- put this string if trace is true (on)

procedure RefreshVolatile( id : identifier );
-- reload this identifier's value from the environment

function getCommandLine return unbounded_string;
procedure getCommandLine ( cmdline : out unbounded_string;
  token_firstpos, token_lastpos : out natural );
-- de-tokenize and return the original command string

procedure compileCommand( command : unbounded_string );
procedure compileScript( firstLine : unbounded_string );

end scanner;
